#include "CollectionBase/CollectionFactory.h"
#include "CollectionBase/CollectionDescription.h"
#include "CollectionBase/ICollectionCursor.h"
#include "CollectionBase/CatalogCollectionCursor.h"

#include "FileCatalog/FCEntry.h"
#include "FileCatalog/IFCContainer.h"
#include "FileCatalog/IFCAction.h"
#include "FileCatalog/FCException.h"
#include "FileCatalog/IFileCatalog.h"

#include "Reflex/PluginService.h"

#include "PersistencySvc/ISession.h"
#include "POOLCore/Exception.h"

#include "CoralBase/MessageStream.h"

#include "boost/tokenizer.hpp"

#include "RVersion.h"
#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
using namespace ROOT;
#endif

pool::CollectionFactory pool::CollectionFactory::s_instance;


const std::string pool::CollectionFactory::c_fileType = "PoolCollection";


pool::CollectionFactory::CollectionFactory()
      : m_myOwnCatalog( false ),
        m_catalog( 0 ) 
{ }


pool::CollectionFactory::~CollectionFactory()
{
   if( m_myOwnCatalog ) delete m_catalog; 
}


pool::CollectionFactory* pool::CollectionFactory::get()
{
  return &s_instance;
}


pool::ICollection*
pool::CollectionFactory::create( const std::string& type,
                                 const pool::ICollectionDescription& description,
                                 pool::ICollection::OpenMode openMode,
                                 pool::ISession* session ) const
{
   const char* out_levl = getenv("POOL_OUTMSG_LEVEL");
   if( out_levl ) {
      coral::MsgLevel       mLvL = coral::Info;
      std::string outLvl( out_levl );
      for( size_t i = 0; i < outLvl.size(); i++ )
         outLvl[i] = tolower( outLvl[i] );
      if( outLvl == "verbose" ) {
         mLvL = coral::Verbose;
      } else
         if( outLvl == "debug" ) {
            mLvL = coral::Debug;
         } else
            if( outLvl == "info" ) {
               mLvL = coral::Info;
            } else
               if( outLvl == "warning" ) {
                  mLvL = coral::Warning;
               } else
                  if( outLvl == "error" ) {
                     mLvL = coral::Error;
                  } else
                     if( outLvl == "fatal" ) {
                        mLvL = coral::Fatal;
                     } else
                        if( outLvl == "always" ) {
                           mLvL = coral::Always;
                        }
      coral::MessageStream::setMsgVerbosity( mLvL );
   }
   
   ICollection *coll = Reflex::PluginService::CreateWithId<ICollection*>( type, &description, openMode, session );
   if( !coll ) {
      std::string errorMsg = "Reflex::PluginService::CreateWithId(" + type + "," + description.name() + ") FAILED";
      if( type == "MemoryCollection" ) {
         errorMsg = "Collection type <MemoryCollection> is not supported";
      }
      throw pool::Exception( errorMsg,
                             "CollectionFactory::create",
                             "CollectionBase" );
   }
   return coll;
}





pool::ICollection*
pool::CollectionFactory::createAndRegister( pool::ICollectionDescription& description,
                                            pool::IFileCatalog* collectionCatalog,
                                            bool overwrite,
                                            std::string logicalName,
                                            pool::MetaDataEntry* metadata,
                                            pool::ISession* session ) const
{
  if( !description.hasEventReferenceColumn() )  {
    std::string errorMsg = "Cannot register a collection fragment in a collection catalog. Must add an event reference columnto collection fragment `" + description.name() + "' to make it a collection.";
    throw pool::Exception( errorMsg,
                           "CollectionFactory::createAndRegister",
                           "CollectionBase" );
  }

  std::string physicalName;
  if( description.type() == "PHYSICAL_NAME" ) {
     physicalName = description.name(); // should be already a type|connection|name
     std::string type, connection, name;     
     extract( physicalName, type, connection, name );
     dynamic_cast<CollectionDescription&>(description).setName(name);
     dynamic_cast<CollectionDescription&>(description).setType(type); 
  } else {
     physicalName = description.type() + "|" + description.connection() + "|" + description.name();
  }
  
  if( !collectionCatalog) collectionCatalog = getDefaultCatalog();
  pool::FileCatalog::FileID guid;
  std::string fileType;
  pool::IFCAction fcAction;
  collectionCatalog->setAction( fcAction );
  collectionCatalog->start();
  fcAction.lookupFileByPFN( physicalName, guid, fileType );
  collectionCatalog->commit();
  
  if( guid.length() )  {
    if( overwrite )    {
      if( ! isUnique( guid, *collectionCatalog ) )      {
	std::string errorMsg = "Cannot overwrite collection `" + description.name() + "' because it has been replicated.";
        throw pool::Exception( errorMsg,
                               "CollectionFactory::createAndRegister", 
                               "CollectionBase" );
      }
    } 
    else {
      std::string errorMsg = "User did not authorize overwrite of registered collection `" + description.name() + "'.";
      throw pool::Exception( errorMsg,
                             "CollectionFactory::createAndRegister", 
                             "CollectionBase" );
    }
  }  

  ICollection::OpenMode openMode = overwrite ? ICollection::CREATE_AND_OVERWRITE : ICollection::CREATE;

  pool::ICollection* collection = create( description.type(), description, openMode, session );
  
  if( collection )  {
    pool::FCregister fcRegister;
    collectionCatalog->setAction( fcRegister );
    collectionCatalog->start();
    if( ! guid.length() )    {
      fcRegister.registerPFN( physicalName, 
                              c_fileType, 
                              guid );
    }
    if( logicalName.length() )    {
      try {
        fcRegister.registerLFN( physicalName, logicalName );
      } 
      catch( pool::FCduplicateLFNException )
      {};
    }
    if( metadata ) {
      fcRegister.registerMetaData( guid, *metadata );
    }
    collectionCatalog->commit();
  }

  return collection;
}


bool 
pool::CollectionFactory::registerExisting( pool::ICollectionDescription& description,
                                           pool::IFileCatalog* collectionCatalog,
                                           std::string logicalName,
                                           pool::MetaDataEntry* metadata,
                                           pool::ISession* session ) const
{
   if( !description.hasEventReferenceColumn() )  {
      std::string errorMsg = "Cannot register a collection fragment in a collection catalog.";
      throw pool::Exception( errorMsg,
                             "CollectionFactory::registerExisting",
                             "CollectionBase" );
   }

   std::string physicalName = 
      description.type() + "|" + description.connection() + "|" + description.name();

   pool::ICollection* collection = openWithPhysicalName( physicalName,
                                                         collectionCatalog,
                                                         true, 
                                                         session );

   if( !collection || !collection->exists( description.name(), false, true ) )  {
      std::string errorMsg = "Collection with physical name `" + physicalName + 
         "' does not exist. Cannot register it in collection catalog.";
      throw pool::Exception( errorMsg,
                             "CollectionFactory::registerExisting", 
                             "CollectionBase" );
   }
  
   if( !collectionCatalog) collectionCatalog = getDefaultCatalog();
   pool::FileCatalog::FileID guid;
   std::string fileType;
   pool::IFCAction fcAction;
   collectionCatalog->setAction( fcAction );
   collectionCatalog->start();
   fcAction.lookupFileByPFN( physicalName, guid, fileType );
   collectionCatalog->commit();

   if( guid.length() )  {
      std::string errorMsg = "Collection with physical name `" + physicalName + "' is already registered in collection catalog.";
      throw pool::Exception( errorMsg,
                             "CollectionFactory::registerExisting",
                             "CollectionBase" );
   }

  pool::FCregister fcRegister;
  collectionCatalog->setAction( fcRegister );
  collectionCatalog->start();
  if( ! guid.length() )
  {
    fcRegister.registerPFN( physicalName, c_fileType, guid );
  }
  if( logicalName.length() )
  {
    try
    {
      fcRegister.registerLFN( physicalName, logicalName );
    } 
    catch( pool::FCduplicateLFNException ){};
  }
  if( metadata )
  {
    fcRegister.registerMetaData( guid, *metadata );
  }
  collectionCatalog->commit();

  return true;
}


pool::ICollection* 
pool::CollectionFactory::openWithPhysicalName( const std::string& physicalName,
                                               pool::IFileCatalog* collectionCatalog,
                                               bool readOnly,
                                               pool::ISession* session ) const
{
   if( readOnly ) {
      std::string type, connection, name;
      extract( physicalName, type, connection, name );
      CollectionDescription	description(name, type, connection);
      return create( type, description, ICollection::READ, session);
  }else{
      if( !collectionCatalog) collectionCatalog = getDefaultCatalog();
      pool::FileCatalog::FileID guid;
      std::string fileType;
      pool::IFCAction fcAction;
      collectionCatalog->setAction( fcAction );
      collectionCatalog->start();
      fcAction.lookupFileByPFN( physicalName, guid, fileType );
      collectionCatalog->commit();
    
      if( guid.length() )  {
	 return openWithGuid( guid, collectionCatalog, readOnly, session );
      }

      /*  MN - old Kristo's code, 
      else  {
	 std::string errorMsg = "No collection registered with physical name `" + physicalName + "' in collection catalog.";
	 throw pool::Exception( errorMsg,
				"CollectionFactory::openWithPhysicalName",
				"CollectionBase" );
      }
      */
   }
   // MN  added instead of Kristo's code
   std::string type, connection, name;
   extract( physicalName, type, connection, name );  
   CollectionDescription	description(name, type, connection);
   return create( type, description, ICollection::UPDATE, session);
}


pool::ICollection* 
pool::CollectionFactory::openWithLogicalName( const std::string& logicalName,
                                              pool::IFileCatalog* collectionCatalog,
                                              bool readOnly,
                                              pool::ISession* session ) const
{
  if( !collectionCatalog) collectionCatalog = getDefaultCatalog();
  pool::FileCatalog::FileID guid;
  pool::IFCAction fcAction;
  collectionCatalog->setAction( fcAction );
  collectionCatalog->start();
  fcAction.lookupFileByLFN( logicalName, guid );
  collectionCatalog->commit();
  if( guid.length() )
  {
    return this->openWithGuid( guid, collectionCatalog, readOnly, session );
  }
  else
  {
    std::string errorMsg = "No collection registered with logical name `" + logicalName + "' in  collection catalog.";
    throw pool::Exception( errorMsg,
                           "CollectionFactory::openWithLogicalName",
                           "CollectionBase" );
  }
}


pool::ICollection* 
pool::CollectionFactory::openWithGuid( const pool::FileCatalog::FileID& guid,
                                       pool::IFileCatalog* collectionCatalog,
                                       bool readOnly,
                                       pool::ISession* session ) const
{
  if( !collectionCatalog) collectionCatalog = getDefaultCatalog();

  if( !readOnly && !isUnique( guid, *collectionCatalog ) ) {
     throw pool::Exception( "Cannot open collection with GUID `" + guid + "' for updates because it has been replicated.",
			    "CollectionFactory::openWithGuid", 
			    "CollectionBase" );
  }

  pool::PFNContainer pfns( collectionCatalog );
  pool::FClookup fcLookup;
  collectionCatalog->setAction( fcLookup );
  collectionCatalog->start();
  fcLookup.lookupPFN( guid, pfns );
  bool found_PFN = pfns.hasNext();
  collectionCatalog->commit();
  if( !found_PFN )  {
     throw pool::Exception( "No collection registered with GUID `" + guid + "' in collection catalog.",
                           "CollectionFactory::openWithGuid", 
                           "CollectionBase" );
  }


  std::string physicalName;
  std::string fileType;
  pool::IFCAction fcAction;
  collectionCatalog->setAction( fcAction );
  collectionCatalog->start();
  fcAction.lookupBestPFN( guid,
                          readOnly ? pool::FileCatalog::READ : pool::FileCatalog::UPDATE,
                          pool::FileCatalog::SEQUENTIAL, // ...?
                          physicalName,
                          fileType );
  collectionCatalog->commit();
  
  if ( fileType != c_fileType )  {
     std::string errorMsg = "Entries in collection catalog must be of type `" + c_fileType + "'.";
     throw pool::Exception( errorMsg,
			    "CollectionFactory::openWithGuid",
			    "CollectionBase" );
  }
  
  std::string name, type, connection;
  extract( physicalName, type, connection, name );  
  ICollection::OpenMode openMode = readOnly ? ICollection::READ : ICollection::UPDATE;
  pool::CollectionDescription description( name, type, connection ); 
  return create( description.type(), description, openMode, session );
}


pool::ICollectionCursor* 
pool::CollectionFactory::getCatalogCollectionCursor( pool::IFileCatalog* collectionCatalog,
                                                     std::string collectionLevelQuery,
                                                     std::string rowLevelQuery,
                                                     std::vector<std::string>* attribOutputList,
                                                     std::vector<std::string>* tokenOutputList,
                                                     int rowCacheSize,
                                                     pool::ISession* session ) const
{
  if( !collectionCatalog) collectionCatalog = getDefaultCatalog();
  return new pool::CatalogCollectionCursor( *collectionCatalog,
                                            collectionLevelQuery,
                                            rowLevelQuery,
                                            attribOutputList,
                                            tokenOutputList,
                                            rowCacheSize,
                                            session );
}


bool 
pool::CollectionFactory::isUnique( const pool::FileCatalog::FileID& guid,
                                   pool::IFileCatalog& collectionCatalog ) const
{  
  pool::PFNContainer pfns( &collectionCatalog );
  pool::FClookup fcLookup;
  collectionCatalog.setAction( fcLookup );
  collectionCatalog.start();
  fcLookup.lookupPFN( guid, pfns );
  if( ! pfns.hasNext() )
  {
    collectionCatalog.commit();
    std::string errorMsg = "A Collection with GUID `" + guid + "' is not registered in the collection catalog.";
    throw pool::Exception( errorMsg,
                           "CollectionFactory::isUnique", 
                           "CollectionBase" );
  }
  
  pfns.Next();
  bool result = !pfns.hasNext();
  collectionCatalog.commit();

  return result;
}


void 
pool::CollectionFactory::extract( const std::string& physicalName,
                                  std::string& type,
                                  std::string& connection,
                                  std::string& name ) const
{
  typedef boost::tokenizer<boost::char_separator<char> > Tizer;
  boost::char_separator<char> sep( "|", "", boost::keep_empty_tokens );
  Tizer tizer( physicalName, sep);
  Tizer::iterator token=tizer.begin(); 
  type = *token;
  connection = *(++token);
  name = *(++token);
}




void 
pool::CollectionFactory::setCatalog( pool::IFileCatalog* catalog)
{
   if( m_catalog && m_myOwnCatalog ) {
      m_catalog->disconnect();
      delete m_catalog;
   }
   m_catalog = catalog; 
   m_myOwnCatalog=false;  
}



pool::IFileCatalog* 
pool::CollectionFactory::getDefaultCatalog() const
{  
   if( m_catalog )
      return m_catalog;
  
   m_catalog = new IFileCatalog();
  
   bool ok = false;  
   const char* chr = getenv("POOL_COLLECTION_WRITE_CATALOG");
   if(chr && strlen(chr)){
      m_catalog->setWriteCatalog( chr );
      ok = true;
   }
   chr = getenv("POOL_COLLECTION_READ_CATALOGS");
   if(chr && strlen(chr)){
      std::string str = chr;
      typedef boost::tokenizer<boost::char_separator<char> > Tizer;
      boost::char_separator<char> sep(" ,");
      Tizer tizer( str, sep);
      for( Tizer::iterator token=tizer.begin();
	   token!=tizer.end();
	   ++token){
	 m_catalog->addReadCatalog( *token );
      }
      ok = true;
   }
   if(!ok)
      m_catalog->setWriteCatalog("xmlcatalog_file:CollectionCatalog.xml");
  
   m_catalog->connect();
   m_myOwnCatalog = true;
  
   return m_catalog;
}







