#ifndef OBJECTRELATIONALACCESS_OBJECTRELATIONALMAPPINGELEMENT_H
#define OBJECTRELATIONALACCESS_OBJECTRELATIONALMAPPINGELEMENT_H

#include <map>
#include <set>
#include <vector>
#include <string>

namespace pool {

  /**
   * @class ObjectRelationalMappingElement ObjectRelationalMappingElement.h ObjectRelationalAccess/ObjectRelationalMappingElement.h
   *
   * Class describing an element of the object-relational mapping structure.
   */

  class ObjectRelationalMappingElement
  {
  public:
    typedef enum { Undefined = -1,
                   Object,
                   Primitive,
                   Array,
                   CArray,
                   InlineCArray,
                   Pointer,
                   Reference,
                   PoolReference,
                   PoolPointer,
                   Blob } ElementType;

    /// Converts the enumeration type to a string
    static std::string elementTypeAsString( ElementType elementType );

    /// Converts a string into an element type
    static ElementType elementTypeFromString( const std::string& elementType );

    /// Forms the scoped variable names for all of the fields
    static std::string scopedVariableName( const std::string& variableName, const std::string& scope);

    /// Forms the variable names for the c-array elements
    static std::string variableNameForArray( const std::string& arrayVariable, unsigned int index );

    /// Empty Constructor
    ObjectRelationalMappingElement();

    /// Constructor
    ObjectRelationalMappingElement( const std::string& elementType,
                                    const std::string& variableName,
                                    const std::string& variableType,
                                    const std::string& tableName );

    // Copy constructor
    ObjectRelationalMappingElement( const ObjectRelationalMappingElement& element);

    // Assignment operator
    ObjectRelationalMappingElement& operator=(const ObjectRelationalMappingElement& element);

    /// Destructor
    ~ObjectRelationalMappingElement() {}

    /**
     * Returns the element type
     */
    ElementType elementType() const;

    /**
     * Returns the scope name
     */
    const std::string& scopeName() const;

    /**
     * Returns the variable name
     */
    const std::string& variableName() const;

    /**
     * Returns the variable C++ type
     */
    const std::string& variableType() const;

    /**
     * Returns the associated table name
     */
    const std::string& tableName() const;

    /**
     * Returns the associated columns
     */
    const std::vector< std::string >& columnNames() const;

    /**
     * Returns the table names involved in the elememt tree  
     */
    std::set<std::string> involvedTables() const;
     
    /**
     * Returns the full element scope  
     */
    const std::string& fullScope() const;
    
    /**
     * Changes the type of the element and propagates the changes to the sub-elements accordingly
     * Note: It should be called before a call to alterTableName and setColumnNames.
     * @param elementType The new type
     */
    void alterType( const std::string& elementType );
    
    /**
     * Changes the name of the associated table and propagates the new information to the sub-elements.
     * Note: It should be called before a call to setColumnNames.
     * @param tableName The new table name
     */
    void alterTableName( const std::string& tableName );

    /**
     * Sets the column names. It is propagated to objects having the same associated table.
     * @param columns The names of the associated columns.
     */
    void setColumnNames( const std::vector< std::string >& columns );

    /// Iterator definition
    typedef std::map< std::string, ObjectRelationalMappingElement >::iterator iterator;
    typedef std::map< std::string, ObjectRelationalMappingElement >::const_iterator const_iterator;

    /// Returns an iterator in the beginning of the sequence
    iterator begin();
    const_iterator begin() const;

    /// Returns an iterator in the end of the sequence
    iterator end();
    const_iterator end() const;

    /// Retrieves a sub-element
    iterator find( const std::string& key );
    const_iterator find( const std::string& key ) const;

    /// Remove a sub-element
    bool removeSubElement( const std::string& key );

    /**
     * Appends a new sub-element.
     * In case the current element is not an object or an array, an ObjectRelationalException is thrown.
     *
     * @param elementType The element type
     * @param variableName The variable name
     * @param variableType The variable type
     * @param tableName The name of the associated table
     */
    ObjectRelationalMappingElement& appendSubElement( const std::string& elementType,
                                                      const std::string& variableName,
                                                      const std::string& variableType,
                                                      const std::string& tableName );

    /**
     * Appends a the specified sub-element.
     * In case the current element is not an object or an array, an ObjectRelationalException is thrown.
     *
     * @param element The element to be appended
     */
    void appendSubElementTree( const ObjectRelationalMappingElement& element );

    /// Returns the scope for column name generation
    const std::string& scopeForColumnName() const;

    void override(const ObjectRelationalMappingElement& source);
    
  private:
    /**
     * The type of this mapping element. It can have one of the following values.
     */
    ElementType m_elementType;
    /**
     * The scope (parent of the given element). In case of top level classes it is an
     * empty string. In case of a field under a top level class it is the fully
     * qualified class name. If the element of the field defines a sub-element then
     * their scope name is the "scope name of the parent" + "::" + "the field name"
     * (m_variableName).
     */
    std::string m_scopeName;
    /**
     * For a top level class this is the same as m_variableType. For a field in a class it is the name of the
     * field itself. For an element in an array, it is the same as the m_variableType.
     */
    std::string m_variableName;
    /**
     * The C++ type of the element.
     */
    std::string m_variableType;
    /**
     * The name of the associated table. For an object it is where it's identity columns
     * are defined. For a primitive and a POOL reference, a reference or a pointer it is the table of
     * the parent object. For an array it is the table where the array elements are stored.
     */
    std::string m_tableName;
    /**
     * The array of the names of the associated columns. For an object they are the ones defining
     * its identity. A primitive is associated to a single column. For a POOL reference they are the ones defining the object
     * identity, plus two columns corresponding to the two fields of the OID.
     * For a pointer or a reference the first column is the referenced
     * table and the rest refer to the identity columns of the target object. For an array the first column is the "position"
     * column and the rest are the columns forming the foreign key constraint w.r.t. the parent table.
     */
    std::vector<std::string> m_columnNames;
    /**
     * The map of sub-elements.
     */
    std::map< std::string, ObjectRelationalMappingElement > m_subElements;
    /**
     * The scope to be used for column name generation
     */
    std::string m_scopeForColumnName;
    /**
     * The scope of the parent mapping element
     */
    std::string m_fullScope;

  };
}

#include "ObjectRelationalMappingElement.inl"

#endif // OBJECTRELATIONALACCESS_OBJECTRELATIONALMAPPINGELEMENT_H
