#include "ObjectRelationalAccess/ObjectRelationalMappingPersistency.h"
#include "ObjectRelationalAccess/ObjectRelationalMapping.h"
#include "ObjectRelationalAccess/ObjectRelationalException.h"
#include "ObjectRelationalAccess/ObjectRelationalNames.h"
#include "MappingMaterialization.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/IColumn.h"
#include "RelationalAccess/ITablePrivilegeManager.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/ITableDescription.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "RelationalAccess/SchemaException.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/MessageStream.h"

#include <sstream>
#include <memory>

pool::ObjectRelationalMappingPersistency::ObjectRelationalMappingPersistency( coral::ISchema& schema ):
  m_schema( schema ),
  m_storedMappings(),
  m_versionsLoaded( false ),
  m_tempMappingTable()
{}


pool::ObjectRelationalMappingPersistency::~ObjectRelationalMappingPersistency()
{
  for ( std::map< std::string, pool::ObjectRelationalMapping* >::iterator iMapping = m_storedMappings.begin();
        iMapping != m_storedMappings.end(); ++iMapping ) {
    if ( iMapping->second ) delete iMapping->second;
  }
}

void
pool::ObjectRelationalMappingPersistency::initialize() const {
  const_cast<pool::ObjectRelationalMappingPersistency*>(this)->loadMappingVersions();  
}

bool
pool::ObjectRelationalMappingPersistency::upgradeSchema() {
  bool ret = false;
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  try {
     // if missing, create all of the required tables
     this->createIfNecessaryTheMappingTables();
     // check if the existing mapping version table has got all of the expected columns
     coral::ITable& versionTable = m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() );
     const coral::ITableDescription& versionTableDescr = versionTable.description();
     try {
       versionTableDescr.columnDescription( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable() );
       log << coral::Warning << " Table \""<< pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() <<"\" is up to date."<< coral::MessageStream::endmsg;
     } catch (const coral::InvalidColumnNameException&){
       ret = true;
       versionTable.schemaEditor().insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable(),
                                              coral::AttributeSpecification::typeNameForType<std::string>());
       log << coral::Warning << " Added column \""<< pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable() << "\" to table \""<< pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() <<"\"."<< coral::MessageStream::endmsg;
       std::ostringstream setClause;
       setClause << pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable() << "= 'POOL_CONTAINER'";
       versionTable.dataEditor().updateRows( setClause.str(),std::string(""),coral::AttributeList());
       versionTable.schemaEditor().setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable() );
     }
  } catch (const coral::Exception& e){
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::upgradeSchema" );    
  }
  return ret;
}

bool
pool::ObjectRelationalMappingPersistency::existsMappingDatabase(){
  bool ret = false;
  try {
    ret = m_schema.existsTable( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() );
  } catch (const coral::Exception& e){
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::existsMappingDatabase" );
  }
  return ret;
}

bool 
pool::ObjectRelationalMappingPersistency::removeMaterializedTables(const std::string& version, bool dryRun ){

  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  try {
    // Check the existence of the versions table.
    if ( ! m_schema.existsTable( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() ) ) {
      log << coral::Warning << " The mapping version table has not been found." << coral::MessageStream::endmsg;
      return false;
    }

    const pool::ObjectRelationalMapping& fullMapping = retrieveMapping( version );
    std::set<std::string> tableList;
    for(ObjectRelationalMapping::const_iterator iChild=fullMapping.begin();iChild!=fullMapping.end();iChild++)
    {
      makeTableList(iChild->second,tableList);
    }
    if(tableList.size()>0){
      log << coral::Info << " Dropping tables referenced by mapping version \"" << version << "\"."<<coral::MessageStream::endmsg;
    } else {
      log << coral::Info << " No table found referenced by mapping version \"" << version << "\"."<<coral::MessageStream::endmsg;
      return false;
    }
    
    for(std::set<std::string>::reverse_iterator iTable=tableList.rbegin();iTable!=tableList.rend();iTable++){

      // check if some other mapping version is referencing the same table
      std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementTable() ).newQuery() );
      query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable() );
      std::ostringstream condition;
      condition << pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable() << " =:" <<
        pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable();
      coral::AttributeList condData;
      condData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable());
      coral::AttributeList::iterator iAttribute = condData.begin();
      iAttribute->data< std::string >() = *iTable;
      query->setCondition( condition.str(), condData );
      std::set<std::string> versions;
      coral::ICursor& cursor = query->execute();
      while ( cursor.next() ) {
        const coral::AttributeList& currentRow = cursor.currentRow();
        versions.insert(currentRow[pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable()].data<std::string>());
      }
      versions.erase( version );
      coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
      if(m_schema.existsTable(*iTable)){
        if(versions.empty()){
          log << coral::Warning << "Dropping table \"" << *iTable <<"\"" << coral::MessageStream::endmsg;
          if(!dryRun) m_schema.dropTable(*iTable);
        } else{
          log << coral::Info << "Table \"" << *iTable <<"\" is currenlty used by other mapping versions." << coral::MessageStream::endmsg;
        }
      }
    }
  } catch (const coral::Exception& e){
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::removeMaterializedTables" );
  }
  return true;
}

bool
pool::ObjectRelationalMappingPersistency::removeMapping( const std::string& version )
{
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  if ( ! m_versionsLoaded) loadMappingVersions();
  if ( m_storedMappings.find( version ) == m_storedMappings.end() ) {
    // Print an error here
    log << coral::Warning << "A mapping with version name \"" << version
        << "\" does not exist in the database. Cannot remove it." << coral::MessageStream::endmsg;
    return false;
  }
  
  // Remove the transient representation of the mapping
  std::map< std::string, pool::ObjectRelationalMapping* >::iterator iMapping = m_storedMappings.find( version );
  if ( iMapping->second ) delete iMapping->second;
  m_storedMappings.erase( iMapping );

  // Remove all rows in the tables with the version.
  coral::AttributeList whereData;
  whereData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable());
  whereData.begin()->data<std::string>() = version;

  try {
    std::string conditionForClassVersionTable = pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable() + " = :" + pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable();
    m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionTable() ).dataEditor().deleteRows( conditionForClassVersionTable, whereData );
    std::string conditionForVersionTable = pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable() + " = :" + pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable();
    m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() ).dataEditor().deleteRows( conditionForVersionTable, whereData );
    std::string conditionForElementTable = pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable() + "= :" + pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable() ;
    m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementTable() ).dataEditor().deleteRows( conditionForElementTable, whereData );
    std::string conditionForColumnsTable = pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementColumnsTable() + "= :" + pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable();
    m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementColumnsTable() ).dataEditor().deleteRows( conditionForColumnsTable, whereData );
  } catch ( const coral::Exception& e){
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::removeMapping" );    
  }
  return true;
}


bool
pool::ObjectRelationalMappingPersistency::storeMapping( const pool::ObjectRelationalMapping& mapping,
                                                        const std::string& classVersion,
                                                        std::string containerName  )
{
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );

  // Check whether the mapping version already exists.
  if ( ! m_versionsLoaded) loadMappingVersions();
  if ( m_storedMappings.find( mapping.version() ) != m_storedMappings.end() ) {
    log << coral::Error << "A mapping with version name \"" << mapping.version()
        << "\" is already stored in the database. No new mapping is stored." << coral::MessageStream::endmsg;
    return false;
  }

  try {
    
    // Store an entry into the version table
    coral::ITable& versionTable = m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() );
    coral::AttributeList dataForVersionTable;
    dataForVersionTable.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable());
    dataForVersionTable.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable());
    coral::AttributeList::iterator iAttribute = dataForVersionTable.begin();
    iAttribute->data< std::string >() = mapping.version();
    ++iAttribute;
    iAttribute->data< std::string >() = containerName;
    versionTable.dataEditor().insertRow( dataForVersionTable );

    // Store an entry into the class version table
    coral::ITable& classVersionTable = m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionTable() );
    coral::AttributeList dataForClassVersionTable;
    dataForClassVersionTable.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable());
    dataForClassVersionTable.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable());
    dataForClassVersionTable.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable());
    coral::AttributeList::iterator iAttr = dataForClassVersionTable.begin();
    iAttr->data< std::string >() = mapping.version();
    ++iAttr;
    iAttr->data< std::string >() = classVersion;
    ++iAttr;
    iAttr->data< std::string >() = containerName;

    std::string existingVersion = this->selectMappingVersion( classVersion, containerName );
    if(existingVersion.empty()){
      classVersionTable.dataEditor().insertRow( dataForClassVersionTable );
    } else {
      std::string setClause = pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable()+" =: "+
        pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable();
      std::string whereClause = pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable()+" =: "+
        pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable()+" AND "+
        pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable()+" =: "+
        pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable();
      classVersionTable.dataEditor().updateRows(setClause,whereClause,dataForClassVersionTable);
    }

    // Store the mapping elements.
    coral::ITable& mappingTable = m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementTable() );
    coral::AttributeList  dataBuffer;
    dataBuffer.extend< std::string >( pool::ObjectRelationalAccess::ObjectRelationalNames::elementTypeFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalAccess::ObjectRelationalNames::variableTypeFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable() );
    dataBuffer.extend< std::string >( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable() );

    dataBuffer[ pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable() ].data<std::string>()= mapping.version();
    coral::ITable& columnsTable = m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementColumnsTable() );
    coral::AttributeList columnDataBuffer;
    columnDataBuffer.extend< std::string >( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementColumnsTable() );
    columnDataBuffer.extend< std::string >( pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementColumnsTable() );
    columnDataBuffer.extend< std::string >( pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementColumnsTable() );
    columnDataBuffer.extend< std::string >( pool::ObjectRelationalAccess::ObjectRelationalNames::columnFieldInMappingElementColumnsTable() );
    columnDataBuffer.extend< int >( pool::ObjectRelationalAccess::ObjectRelationalNames::positionFieldInMappingElementColumnsTable() );

    columnDataBuffer[ pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementColumnsTable() ].data<std::string>()= mapping.version();
    for ( pool::ObjectRelationalMapping::const_iterator iElement = mapping.begin();
          iElement != mapping.end(); ++iElement ) {
      pool::ObjectRelationalMappingPersistency::storeMappingElement( mappingTable.dataEditor(),
                                                                     dataBuffer,
                                                                     columnsTable.dataEditor(),
                                                                     columnDataBuffer,
                                                                     iElement->second );
    }
    m_storedMappings[ mapping.version() ] = 0;
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::storeMapping" );    
  }
  return true;
}

std::map<std::string,std::string>
pool::ObjectRelationalMappingPersistency::retrieveMappingVersions() const
{
  std::map<std::string,std::string> mappingList;
  try {
    std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() ).newQuery() );
    query->setDistinct();
    query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable() );
    query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable() );
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string containerName = currentRow[pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable()].data<std::string>();
      std::string mappingVersion = currentRow[pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable()].data<std::string>();
      mappingList.insert(std::make_pair(mappingVersion,containerName));
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::containerForMappingVersion" );
  }
  return mappingList;
}

std::map<std::pair<std::string,std::string>,std::string>
pool::ObjectRelationalMappingPersistency::retrieveMappingForClassVersions() const{
  std::map<std::pair<std::string,std::string>,std::string> classVersionMap;
  try {
    std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionTable() ).newQuery() );
    query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable() );
    query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable() );
    query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable() );
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string classVersion = currentRow[pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable()].data<std::string>();
      std::string containerName = currentRow[pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable()].data<std::string>();
      std::string mappingVersion = currentRow[pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable()].data<std::string>();
      classVersionMap.insert(std::make_pair(std::make_pair(classVersion,containerName),mappingVersion));
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::selectMappingVersion" );
  }
  return classVersionMap;
}

std::string
pool::ObjectRelationalMappingPersistency::selectMappingVersion( const std::string& classVersion, const std::string& containerName ) const
{
  std::string result("");
  try {
    std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionTable() ).newQuery() );
    query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable() );
    std::ostringstream condition;
    condition << pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable() << " =:" <<
      pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable() << " AND " <<
      pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable();
    condition << " =:" << pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable();

    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable());
    condData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = classVersion;
    ++iAttribute;
    iAttribute->data< std::string >() = containerName;
    query->setCondition( condition.str(), condData );
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      result = currentRow[pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable()].data<std::string>();
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::selectMappingVersion" );
  }
  return result;
}

void
pool::ObjectRelationalMappingPersistency::selectMapping(const std::string& version,
                                                        std::map<std::pair<std::string,std::string>,std::pair<std::string,std::string> >& mappingStruct) const
{
  try {
    std::auto_ptr<coral::IQuery> query(m_schema.newQuery());
    query->addToTableList( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementTable(),"T0");
    query->addToTableList( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementColumnsTable(), "T1");
    query->addToOutputList( "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable());
    query->addToOutputList( "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable());
    query->addToOutputList( "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable() );
    query->addToOutputList( "T1."+pool::ObjectRelationalAccess::ObjectRelationalNames::columnFieldInMappingElementColumnsTable());
    std::ostringstream condition;
    condition << "T0."<<pool::ObjectRelationalAccess::ObjectRelationalNames::elementTypeFieldInMappingElementTable()<<"= '"<<
      ObjectRelationalAccess::ObjectRelationalNames::primitiveMappingElementType()<<"' AND "<<
      "T0."<<pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable()<<"="
      "T1."<<pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable()<<" AND "<<
      "T0."<<pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable()<<"="<<
      "T1."<<pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable()<<" AND "<<
      "T0."<<pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable()<<"= :"<<
      pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = version;
    query->setCondition( condition.str(), condData );
    query->addToOrderList("T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable());
    query->addToOrderList("T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable());
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string scopeName = currentRow[ "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable()].data<std::string>();
      std::string variableName = currentRow[ "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable() ].data<std::string>();
      std::string tableName = currentRow[ "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable() ].data<std::string>();
      std::string columnName = currentRow[ "T1."+pool::ObjectRelationalAccess::ObjectRelationalNames::columnFieldInMappingElementColumnsTable()].data<std::string>();
      mappingStruct.insert(std::make_pair(std::make_pair(scopeName,variableName),std::make_pair(tableName,columnName)));
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::selectMapping" );
  }
}

std::string
pool::ObjectRelationalMappingPersistency::containerForMappingVersion( const std::string& mappingVersion) const {
  std::string currentContainerName("");
  try {
    std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() ).newQuery() );
    query->setDistinct();
    query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable() );
    std::ostringstream condition;
    condition << pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable() << " =:"<<
      pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = mappingVersion;
    query->setCondition( condition.str(), condData );
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      currentContainerName = currentRow[pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable()].data<std::string>();
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::containerForMappingVersion" );
  }
  return currentContainerName;
}

std::vector<std::string>
pool::ObjectRelationalMappingPersistency::mappingsForContainer( const std::string& containerName ) const {
  std::vector<std::string> mappings;
  try {
    std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() ).newQuery() );
    query->setDistinct();
    query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable());
    std::ostringstream condition;
    condition << pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable() << " =:"<<
      pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = containerName;
    query->setCondition( condition.str(), condData );
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      mappings.push_back(currentRow[pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable()].data<std::string>());
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::mappingsForContainer" );
  }
  return mappings;
}

std::map<std::string, std::string>
pool::ObjectRelationalMappingPersistency::containerTableMap()
{
  std::map<std::string,std::string> containerTableMap;
  try {
    std::auto_ptr<coral::IQuery> query(m_schema.newQuery());
    query->addToTableList( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementTable(),"T0");
    query->addToTableList( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable(), "T1");
    query->setDistinct();
    query->addToOutputList( "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable() );
    query->addToOutputList( "T1."+pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable());
    std::ostringstream condition;
    condition << "T0."<<pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable()<<"="<<
      "T1."<<pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable();
    coral::AttributeList condData;
    query->setCondition(condition.str(),condData);
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string tableName = currentRow[ "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable()].data<std::string>();
      std::string containerName = currentRow[ "T1."+pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable()].data<std::string>();
      containerTableMap.insert(std::make_pair(tableName,containerName));
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::containerTableMap" );
  }
  return containerTableMap;
}

const pool::ObjectRelationalMapping&
pool::ObjectRelationalMappingPersistency::retrieveMapping( const std::string& version ) const
{
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  pool::ObjectRelationalMapping* mappingPointer = 0;
  try {
    if ( ! m_versionsLoaded) {
      const_cast<pool::ObjectRelationalMappingPersistency*>(this)->loadMappingVersions();
    }

    std::map< std::string, pool::ObjectRelationalMapping* >::const_iterator iMapping = m_storedMappings.find( version );
    if ( iMapping == m_storedMappings.end() ) {
      log << coral::Error << "No mapping with version name \"" << version << "\" is stored in the database." << coral::MessageStream::endmsg;
    } else {
      mappingPointer = iMapping->second;
      if ( ! mappingPointer ) {
        const_cast<pool::ObjectRelationalMappingPersistency*>(this)->loadMapping( version );
        mappingPointer = m_storedMappings.find( version )->second;
      }
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::retrieveMapping" );
  }
  if(! mappingPointer ){
    throw pool::ObjectRelationalException( "Non-existing mapping version name",
                                           "ObjectRelationalMappingPersistency::retrieveMapping" );
  }
  
  return *mappingPointer;
}

pool::ObjectRelationalMapping*
pool::ObjectRelationalMappingPersistency::exportMapping( const std::string& version ) const
{

  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  pool::ObjectRelationalMapping* mapping = 0;
  try {
    if ( ! m_versionsLoaded) {
      const_cast<pool::ObjectRelationalMappingPersistency*>(this)->loadMappingVersions();
    }

    std::map< std::string, pool::ObjectRelationalMapping* >::const_iterator iMapping = m_storedMappings.find( version );
    if ( iMapping == m_storedMappings.end() ) {
      log << coral::Error << "No mapping with version name \"" << version << "\" is stored in the database." << coral::MessageStream::endmsg;
    } else {
      mapping = new pool::ObjectRelationalMapping( version );
      const_cast<pool::ObjectRelationalMappingPersistency*>(this)->loadMappingInOneGo( version, mapping);
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::retrieveMapping" );
  }
  if(!mapping){
    throw pool::ObjectRelationalException( "Non-existing mapping version name",
                                             "ObjectRelationalMappingPersistency::exportMapping" );
  }
  return mapping;
}

bool
pool::ObjectRelationalMappingPersistency::materializeMapping( const std::string& version,
                                                              bool evolve,
                                                              bool dryRun )
{
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  // Checking whether the mapping exists.
  bool ret = false;
  try {
    if ( ! m_versionsLoaded) {
      const_cast<pool::ObjectRelationalMappingPersistency*>(this)->loadMappingVersions();
    }
    std::map< std::string, pool::ObjectRelationalMapping* >::iterator iMapping = m_storedMappings.find( version );
    if ( iMapping == m_storedMappings.end() ) {
      log << coral::Error << "Mapping version \"" << version << "\" does not exist. Not materializing the mapping..." << coral::MessageStream::endmsg;
      return false;
    }
    if ( ! iMapping->second ) {
      this->loadMapping( version );
      iMapping = m_storedMappings.find( version );
    }
    pool::ObjectRelationalAccess::MappingMaterialization materialization( m_schema );
    ret = materialization.materialize( *( iMapping->second ), evolve, dryRun );
    if(dryRun){
      const std::set<std::string>& logs = materialization.operationList();
      if(logs.empty()) log << coral::Warning << "# No changes on the schema are needed." << coral::MessageStream::endmsg;
      for( std::set<std::string>::const_iterator iLog = logs.begin();
           iLog!=logs.end();++iLog){
        log << coral::Warning << "# Schema change: "<<*iLog<<coral::MessageStream::endmsg;
      }
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::materializeMapping" );
  }
  return ret;
}

bool
pool::ObjectRelationalMappingPersistency::materializeMapping( const pool::ObjectRelationalMapping& mapping,
                                                              bool evolve,
                                                              bool dryRun )
{
  // Checking whether the mapping exists.
  bool ret = false;
  try {
    pool::ObjectRelationalAccess::MappingMaterialization materialization( m_schema );
    ret = materialization.materialize( mapping, evolve, dryRun );
    if(dryRun){
      coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
      const std::set<std::string>& logs = materialization.operationList();
      if(logs.empty()) log << coral::Warning << "# No changes on the schema are needed." <<coral::MessageStream::endmsg;
      for( std::set<std::string>::const_iterator iLog = logs.begin();
           iLog!=logs.end();++iLog){
        log << coral::Warning << "# Schema change: "<<*iLog<<coral::MessageStream::endmsg;
      }
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::materializeMapping" );
  }
  return ret;
}

bool
pool::ObjectRelationalMappingPersistency::assignMapping( const std::string& classVersion, const std::string& mappingVersion, const std::string& toContainerName)
{
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
  // first check if the specified mapping is used by an other container
  std::string currentCont = containerForMappingVersion(mappingVersion);
  if(currentCont.empty()){
    log << coral::Error << "Mapping \""<<mappingVersion<<"\" for container \""<<toContainerName<<"\" does not exist in the database."<<coral::MessageStream::endmsg;
    return false;
  } else {
    if(currentCont != toContainerName){
      log << coral::Error << "Mapping \""<<mappingVersion<<"\" is already assigned to container \""<<currentCont<<"\"."<<coral::MessageStream::endmsg;
      return false;
    }
  }
  try {
    std::auto_ptr<coral::IQuery> query0( m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() ).newQuery() );
    query0->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable() );
    std::ostringstream cond;
    cond << pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable()<< " =:"<<
      pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable());
    coral::AttributeList::iterator iAttr = condData.begin();
    iAttr->data< std::string >() = mappingVersion;
    query0->setCondition( cond.str(), condData);
    std::string currentContainerName("");
    coral::ICursor& cursor0 = query0->execute();
    while ( cursor0.next() ) {
      const coral::AttributeList& currentRow = cursor0.currentRow();      
      currentContainerName = currentRow[pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable()].data<std::string>();
    }
    if(currentContainerName.empty()){
      log << coral::Error << "Mapping version \""+mappingVersion+"\" has not been found in the database." << coral::MessageStream::endmsg;
      return false;
    } else {
      if(toContainerName != currentContainerName){
        log << coral::Error << "Mapping version=\""<<mappingVersion<<"\" is already assigned to container \""<<currentContainerName<<"\"." << coral::MessageStream::endmsg;
        return false;
      }
    }
    
    // first check if the classVersion and the container are already represented
    std::auto_ptr<coral::IQuery> query1( m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionTable() ).newQuery() );
    query1->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable() );
    std::ostringstream condition;
    condition << pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable() << " =:"<<
      pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable();
    condition << " AND ";
    condition << pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable() << " =:"<<
      pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable();
    coral::AttributeList conditionData;
    conditionData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable());
    conditionData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable());
    coral::AttributeList::iterator iAttribute = conditionData.begin();
    iAttribute->data< std::string >() = classVersion;
    ++iAttribute;
    iAttribute->data< std::string >() = toContainerName;    
    query1->setCondition( condition.str(), conditionData );
    std::set<std::string> classVersions;
    coral::ICursor& cursor1 = query1->execute();
    std::string currentVersion("");
    while ( cursor1.next() ) {
      const coral::AttributeList& currentRow = cursor1.currentRow();
      currentVersion = currentRow[pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable()].data<std::string>();
    }
    bool update = false;
    if(!currentVersion.empty()){
      if(currentVersion!=mappingVersion){
        update = true;
      } else {
        log << coral::Debug << "Mapping version=\""<<mappingVersion<<"\" is already assigned to container \""<<toContainerName<<"\" for class version \""<<
          classVersion<<"\"." << coral::MessageStream::endmsg;
        return true;
      }
    }
    
    coral::ITable& classVersionTable = m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionTable() );
    coral::AttributeList inputData;
    inputData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable());
    inputData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable());
    inputData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable());
    coral::AttributeList::iterator iInAttr = inputData.begin();
    iInAttr->data< std::string >() = mappingVersion;
    ++iInAttr;
    iInAttr->data< std::string >() = classVersion;
    ++iInAttr;
    iInAttr->data< std::string >() = toContainerName;
    if(update){
      std::string setClause = pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable()+" =:"+
        pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable();
      std::string whereClause = pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable()+" =:"+
        pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable()+" AND "+
        pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable()+" =:"+
        pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable();
      classVersionTable.dataEditor().updateRows( setClause,whereClause, inputData  );
    } else {
      classVersionTable.dataEditor().insertRow( inputData );
    }
    log << coral::Debug << "Mapping version \""<< mappingVersion << "\" for class version \""<<classVersion<<
      "\" has been assigned to pool container \"" << toContainerName << "\""<<coral::MessageStream::endmsg;
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::assignMapping" );
  }
  return true;
}

void pool::ObjectRelationalMappingPersistency::makeTableList(const ObjectRelationalMappingElement& element,std::set<std::string>& tableList)
{
  tableList.insert(element.tableName());
  for(ObjectRelationalMappingElement::const_iterator iChild=element.begin();iChild!=element.end();iChild++)
  {
    makeTableList(iChild->second,tableList);
  }
}

void
pool::ObjectRelationalMappingPersistency::createIfNecessaryTheMappingTables()
{
  coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );
 // The versions table.
  if ( ! m_schema.existsTable( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() ) ) {
    log << coral::Warning << "Creating table \"" << pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() << "\"" << coral::MessageStream::endmsg;
    std::auto_ptr< coral::TableDescription > description( new coral::TableDescription( "ObjectRelationalAccess" ) );
    description->setName(pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable());
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable(),
                               coral::AttributeSpecification::typeNameForType<std::string>());
    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable() );
    description->setUniqueConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable() );
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable(),
                               coral::AttributeSpecification::typeNameForType<std::string>());
    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInMappingVersionTable() );
    m_schema.createTable( *description ).privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );
  }

 // The class versions table.
  if ( ! m_schema.existsTable( pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionTable() ) ) {
    log << coral::Warning << "Creating table \"" << pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionTable() << "\"" << coral::MessageStream::endmsg;
    std::auto_ptr< coral::TableDescription > description( new coral::TableDescription( "ObjectRelationalAccess" ) );
    description->setName(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionTable());
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable(),
                               coral::AttributeSpecification::typeNameForType<std::string>(), 1000, false);
    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable() );
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable(),
                               coral::AttributeSpecification::typeNameForType<std::string>(), 1000, false);
    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable() );
    std::vector<std::string> cols;
    cols.push_back(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionFieldInClassVersionTable());
    cols.push_back(pool::ObjectRelationalAccess::ObjectRelationalNames::containerNameFieldInClassVersionTable());
    description->setUniqueConstraint( cols );
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable(),
                               coral::AttributeSpecification::typeNameForType<std::string>());
    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable() );  
    description->createForeignKey(
      pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable()+"_FK",
      pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionFieldInClassVersionTable(),
      pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable(),
      pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable());
    m_schema.createTable( *description ).privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );
  }
  
  // The table holding the mapping elements.
  if ( ! m_schema.existsTable( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementTable() ) ) {
    log << coral::Warning << "Creating table \"" << pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementTable() << "\"" << coral::MessageStream::endmsg;
    std::auto_ptr< coral::TableDescription > description( new coral::TableDescription( "ObjectRelationalAccess" ) );

    description->setName(pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementTable());
    
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::elementTypeFieldInMappingElementTable(),
                               coral::AttributeSpecification::typeNameForType<std::string>() );
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable(),
                               coral::AttributeSpecification::typeNameForType<std::string>() );
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable(),
                               coral::AttributeSpecification::typeNameForType<std::string>() );
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::variableTypeFieldInMappingElementTable(),
                               coral::AttributeSpecification::typeNameForType<std::string>() );
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable(),
                               coral::AttributeSpecification::typeNameForType<std::string>() );
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable(),
                               coral::AttributeSpecification::typeNameForType<std::string>() );
    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::elementTypeFieldInMappingElementTable() );
    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable() );
    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable() );
    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::variableTypeFieldInMappingElementTable() );
    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable() );
    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable() );
    m_schema.createTable( *description ).privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );
  }

  // The table holding the columns in the mapping elements.
  if ( ! m_schema.existsTable( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementColumnsTable() ) ) {
    log << coral::Warning << "Creating table \"" << pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementColumnsTable() << "\"" << coral::MessageStream::endmsg;
    std::auto_ptr< coral::TableDescription > description( new coral::TableDescription( "ObjectRelationalAccess" ) );

    description->setName(pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementColumnsTable());
    
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementColumnsTable(),
                               coral::AttributeSpecification::typeNameForType<std::string>() );
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementColumnsTable(),
                               coral::AttributeSpecification::typeNameForType<std::string>() );
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementColumnsTable(),
                               coral::AttributeSpecification::typeNameForType<std::string>() );
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::columnFieldInMappingElementColumnsTable(),
                               coral::AttributeSpecification::typeNameForType<std::string>() );
    description->insertColumn( pool::ObjectRelationalAccess::ObjectRelationalNames::positionFieldInMappingElementColumnsTable(),
                               coral::AttributeSpecification::typeNameForType<int>() );

    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementColumnsTable() );
    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementColumnsTable() );
    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementColumnsTable() );
    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::columnFieldInMappingElementColumnsTable() );
    description->setNotNullConstraint( pool::ObjectRelationalAccess::ObjectRelationalNames::positionFieldInMappingElementColumnsTable() );

    m_schema.createTable( *description ).privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );
  }

}


void
pool::ObjectRelationalMappingPersistency::loadMappingVersions()
{
  if ( m_versionsLoaded ) return;
  this->createIfNecessaryTheMappingTables();
  std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingVersionTable() ).newQuery() );
  query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable() );
  coral::ICursor& cursor = query->execute();
  while ( cursor.next() ) {
    const coral::AttributeList& currentRow = cursor.currentRow();
    m_storedMappings[ currentRow[pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingVersionTable()].data<std::string>()] = 0;
  }
  m_versionsLoaded = true;
}

void
pool::ObjectRelationalMappingPersistency::loadMappingInOneGo( const std::string& version, pool::ObjectRelationalMapping* mapping )
{
  std::vector< std::pair<std::string, MappingDescription> > topLevelObjects;

  std::auto_ptr<coral::IQuery> query(m_schema.newQuery());
  try {
    query->addToTableList( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementTable(),"T0");
    query->addToTableList( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementColumnsTable(), "T1");
    query->addToOutputList( "T1."+pool::ObjectRelationalAccess::ObjectRelationalNames::columnFieldInMappingElementColumnsTable());
    query->addToOutputList( "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::elementTypeFieldInMappingElementTable());
    query->addToOutputList( "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::variableTypeFieldInMappingElementTable());
    query->addToOutputList( "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable());
    query->addToOutputList( "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable());
    query->addToOutputList( "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable() );
    std::ostringstream condition;
    condition << "T1."<<pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable()<<"="<<
      "T0."<<pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable()<<" AND "<<
      "T1."<<pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable()<<"="
      "T0."<<pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable()<<" AND "<<
      "T1."<<pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable()<<"="<<
      "T0."<<pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable()<<" AND "<<
      "T0."<<pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable()<<"= :"<<
      pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable();
    coral::AttributeList condData;
    condData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable());
    coral::AttributeList::iterator iAttribute = condData.begin();
    iAttribute->data< std::string >() = version;
    query->setCondition( condition.str(), condData );
    query->addToOrderList("T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable());
    query->addToOrderList("T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable());
    query->addToOrderList("T1."+pool::ObjectRelationalAccess::ObjectRelationalNames::positionFieldInMappingElementColumnsTable());
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      const coral::AttributeList& currentRow = cursor.currentRow();
      std::string columnName = currentRow[ "T1."+pool::ObjectRelationalAccess::ObjectRelationalNames::columnFieldInMappingElementColumnsTable()].data<std::string>();
      std::string elementType = currentRow[ "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::elementTypeFieldInMappingElementTable() ].data<std::string>();
      std::string variableType = currentRow[ "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::variableTypeFieldInMappingElementTable() ].data<std::string>();
      std::string variableName = currentRow[ "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable() ].data<std::string>();
      std::string scopeName = currentRow[ "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable()].data<std::string>();
      std::string tableName = currentRow[ "T0."+pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable() ].data<std::string>();
      MappingDescription mapDescr;
      mapDescr.variableType = variableType;
      mapDescr.elementType = elementType;
      mapDescr.tableName = tableName;
      mapDescr.columns = std::vector<std::string>(1,columnName);
      if(scopeName == " "){  
        topLevelObjects.push_back(std::make_pair(variableName, mapDescr) );
      } else {
        std::map<std::string,std::map<std::string,MappingDescription> >::iterator iMap = m_tempMappingTable.find(scopeName);
        if(iMap==m_tempMappingTable.end()) {
          std::map<std::string,MappingDescription> entry;
          entry.insert(std::make_pair(variableName,mapDescr));
          m_tempMappingTable.insert(std::make_pair(scopeName,entry));
        } else {                    
          std::map<std::string,MappingDescription>::iterator iInnerMap = iMap->second.find(variableName);
          if(iInnerMap==iMap->second.end()){
            iMap->second.insert(std::make_pair(variableName,mapDescr));
          } else {
            iInnerMap->second.columns.push_back(columnName);
          }
        }
      
      }
    
    }
  } catch(const coral::Exception& e) {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::loadMappingInOneGo" );
  }
  
  for(std::vector<std::pair<std::string, MappingDescription> >::const_iterator iObj=topLevelObjects.begin();iObj!=topLevelObjects.end();iObj++){
    ObjectRelationalMappingElement& element = mapping->appendElement(iObj->first,iObj->second.tableName);    
    element.setColumnNames(iObj->second.columns);
    recursiveStoreMappingElement(element,iObj->first);
  }
  m_tempMappingTable.clear();
}

void
pool::ObjectRelationalMappingPersistency::loadMapping( const std::string& version )
{
  
  pool::ObjectRelationalMapping* mapping = m_storedMappings[ version ] = new pool::ObjectRelationalMapping( version );

  loadMappingInOneGo( version, mapping);

  //loadMappingLoop( version, mapping);
  
}
 
void
pool::ObjectRelationalMappingPersistency::loadMappingLoop( const std::string& version, pool::ObjectRelationalMapping* mapping  )
{
  std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementTable() ).newQuery() );
  std::ostringstream condition;
  condition << pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable() << "= :" <<
               pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable() << " AND " <<
               pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable() << "= :" <<
               pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable();
  coral::AttributeList condData;
  condData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable());
  condData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable());
  coral::AttributeList::iterator iAttribute = condData.begin();
  iAttribute->data< std::string >() = version;
  ++iAttribute;
  iAttribute->data< std::string >() = ' ';
  query->setCondition( condition.str(), condData );
  query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable() );
  query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable() );
  coral::ICursor& cursor = query->execute();
  while ( cursor.next() ) {
    const coral::AttributeList& currentRow = cursor.currentRow();
    std::string className = currentRow[ pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable() ].data<std::string>();
    std::string tableName = currentRow[ pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable() ].data<std::string>();
    ObjectRelationalMappingElement& parentElement = mapping->appendElement( className, tableName );
    // Fill the sub element
     this->loadMappingElement( version, parentElement );
  }
}


void
pool::ObjectRelationalMappingPersistency::loadMappingElement( const std::string& version,
                                                              pool::ObjectRelationalMappingElement& parentElement )
{

  // Load the columns
  this->loadMappingElementColumns( version, parentElement );

  // Fetch and append the sub-elements
  std::auto_ptr<coral::IQuery> query( m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementTable() ).newQuery() );
  std::string scope = parentElement.scopeName();
  if ( ! scope.empty() ) {
    scope += "::" + parentElement.variableName();
  }
  else {
    scope = parentElement.variableName();
  }
  std::ostringstream condition;
  condition << pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable() << "= :" <<
               pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable()<< " AND " <<
               pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable() << "= :" <<
               pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable();
  coral::AttributeList condData;
  condData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable());
  condData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable());
  coral::AttributeList::iterator iAttribute = condData.begin();
  iAttribute->data< std::string >() = version;
  ++iAttribute;
  iAttribute->data< std::string >() = scope;
  query->setCondition( condition.str(), condData );
  query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::elementTypeFieldInMappingElementTable() );
  query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable() );
  query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::variableTypeFieldInMappingElementTable() );
  query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable() );
  coral::ICursor& cursor = query->execute();
  while ( cursor.next() ) {
    const coral::AttributeList& currentRow = cursor.currentRow();
    coral::AttributeList::const_iterator iAttribute = currentRow.begin();
    std::string elementType = iAttribute->data<std::string>();
    ++iAttribute;
    std::string variableName = iAttribute->data<std::string>();
    ++iAttribute;
    std::string variableType = iAttribute->data<std::string>();
    ++iAttribute;
    std::string tableName = iAttribute->data<std::string>();
    ObjectRelationalMappingElement& element = parentElement.appendSubElement( elementType,
                                                                              variableName,
                                                                              variableType,
                                                                              tableName );
    this->loadMappingElement( version, element );
  }
}

void
pool::ObjectRelationalMappingPersistency::loadMappingElementColumns( const std::string& version,
                                                                     pool::ObjectRelationalMappingElement& element )
{
  coral::ITable& columnsTable = m_schema.tableHandle( pool::ObjectRelationalAccess::ObjectRelationalNames::mappingElementColumnsTable() );
  std::auto_ptr<coral::IQuery> query( columnsTable.newQuery() );
  const std::string scope = ( element.scopeName().empty() ) ? " " : element.scopeName();
  std::ostringstream condition;
  condition << pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementColumnsTable() << "= :" <<
               pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementColumnsTable() << " AND " <<
               pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable() << "= :" <<
               pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable() << " AND " <<
               pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable() << "= :" <<
               pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable();
  coral::AttributeList condData;
  condData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::versionFieldInMappingElementTable());
  condData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable());
  condData.extend<std::string>(pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable());
  coral::AttributeList::iterator iAttribute = condData.begin();
  iAttribute->data< std::string >() = version;
  ++iAttribute;
  iAttribute->data< std::string >() = scope;
  ++iAttribute;
  iAttribute->data< std::string >() = element.variableName();  
  query->setCondition( condition.str(), condData );
  query->addToOrderList( pool::ObjectRelationalAccess::ObjectRelationalNames::positionFieldInMappingElementColumnsTable() );
  query->addToOutputList( pool::ObjectRelationalAccess::ObjectRelationalNames::columnFieldInMappingElementColumnsTable() );
  coral::ICursor& cursor = query->execute();
  std::vector< std::string > columns;
  while ( cursor.next() ) {
    columns.push_back( cursor.currentRow().begin()->data<std::string>() );
  }
  element.setColumnNames( columns );
}


void
pool::ObjectRelationalMappingPersistency::storeMappingElement( coral::ITableDataEditor& dataEditor,
                                                               coral::AttributeList& dataBuffer,
                                                               coral::ITableDataEditor& columnsDataEditor,
                                                               coral::AttributeList& columnsDataBuffer,
                                                               const pool::ObjectRelationalMappingElement& element )
{
  dataBuffer[ pool::ObjectRelationalAccess::ObjectRelationalNames::elementTypeFieldInMappingElementTable() ].data<std::string>()= pool::ObjectRelationalMappingElement::elementTypeAsString( element.elementType() );
  if ( element.scopeName().empty() ) {
    dataBuffer[ pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable() ].data<std::string>()= " " ;
    columnsDataBuffer[ pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementColumnsTable() ].data<std::string>()= " ";
  }
  else {
    dataBuffer[ pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementTable() ].data<std::string>()= element.scopeName();
    columnsDataBuffer[ pool::ObjectRelationalAccess::ObjectRelationalNames::scopeNameFieldInMappingElementColumnsTable() ].data<std::string>()=element.scopeName();
  }
  dataBuffer[ pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementTable() ].data<std::string>()= element.variableName();
  dataBuffer[ pool::ObjectRelationalAccess::ObjectRelationalNames::variableTypeFieldInMappingElementTable() ].data<std::string>()= element.variableType();
  dataBuffer[ pool::ObjectRelationalAccess::ObjectRelationalNames::tableNameFieldInMappingElementTable() ].data<std::string>() = element.tableName();
  try 
  {
    dataEditor.insertRow( dataBuffer );
  }
  catch (const coral::SchemaException& e)
  {
    throw pool::ObjectRelationalException( e.what(),
                                           "ObjectRelationalMappingPersistency::storeMappingElement" );
  }

  columnsDataBuffer[ pool::ObjectRelationalAccess::ObjectRelationalNames::variableNameFieldInMappingElementColumnsTable() ].data<std::string>() = element.variableName();
  int iColumnId = 0;
  for ( std::vector< std::string >::const_iterator iColumn = element.columnNames().begin();
        iColumn != element.columnNames().end(); ++iColumn, ++iColumnId ) {
    columnsDataBuffer[ pool::ObjectRelationalAccess::ObjectRelationalNames::columnFieldInMappingElementColumnsTable() ].data<std::string>() = *iColumn;
    columnsDataBuffer[ pool::ObjectRelationalAccess::ObjectRelationalNames::positionFieldInMappingElementColumnsTable() ].data<int>() = iColumnId;
    try 
    {
      columnsDataEditor.insertRow( columnsDataBuffer );
    }
    catch (const coral::SchemaException& e)
    {
      throw pool::ObjectRelationalException( e.what(),
                                             "ObjectRelationalMappingPersistency::storeMappingElement" );
    }
  }

  for ( pool::ObjectRelationalMappingElement::const_iterator iElement = element.begin();
        iElement != element.end(); ++iElement ) {
    pool::ObjectRelationalMappingPersistency::storeMappingElement( dataEditor, dataBuffer, columnsDataEditor, columnsDataBuffer, iElement->second );
  }
}

void
pool::ObjectRelationalMappingPersistency::recursiveStoreMappingElement( ObjectRelationalMappingElement& parentElement, const std::string& scopeName ){
  std::map<std::string,std::map<std::string,MappingDescription> >::iterator iMap = m_tempMappingTable.find(scopeName);
  if(iMap!=m_tempMappingTable.end()){
    for(std::map<std::string,MappingDescription>::const_iterator iMapEntry=iMap->second.begin();iMapEntry!=iMap->second.end();iMapEntry++){
      std::string variableName = iMapEntry->first;
      ObjectRelationalMappingElement& element = parentElement.appendSubElement( iMapEntry->second.elementType,
                                                                                variableName,
                                                                                iMapEntry->second.variableType,
                                                                                iMapEntry->second.tableName );
      element.setColumnNames(iMapEntry->second.columns);
      std::string nextScope(scopeName);
      nextScope.append("::").append(variableName);
      recursiveStoreMappingElement( element, nextScope);
    }
    m_tempMappingTable.erase(scopeName);
  }
}

