#include "ObjectRelationalAccess/ObjectRelationalUtilities.h"
#include "ObjectRelationalAccess/ObjectRelationalNames.h"
#include "ObjectRelationalAccess/ObjectRelationalException.h"
#include "StringTokenize.h"
#include "Reflex/SharedLibrary.h"
#include "RelationalAccess/ISessionProxy.h"
#include "RelationalAccess/ConnectionService.h"
#include "RelationalAccess/IConnectionServiceConfiguration.h"
#include "CoralCommon/AuthenticationCredentials.h"
#include <stdexcept>

#include "RVersion.h"
#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
using namespace ROOT;
#endif

pool::ObjectRelationalUtilities::LocalAuthenticationService::LocalAuthenticationService()
  : m_credentials( 0 )
{
  m_credentials = new coral::AuthenticationCredentials( "POOL/Services/LocalAuthenticationService" );
}

pool::ObjectRelationalUtilities::LocalAuthenticationService::~LocalAuthenticationService()
{
  if ( m_credentials ) delete m_credentials;
}


const coral::IAuthenticationCredentials&
pool::ObjectRelationalUtilities::LocalAuthenticationService::credentials( const std::string& ) const
{
  return *m_credentials;
}

const coral::IAuthenticationCredentials&
pool::ObjectRelationalUtilities::LocalAuthenticationService::credentials( const std::string&, const std::string& ) const
{
  return *m_credentials;
}

void pool::ObjectRelationalUtilities::LocalAuthenticationService::setCredentials(const std::string& userName, const std::string& password)
{
  m_credentials->registerItem( "user", std::string( userName ) );
  m_credentials->registerItem( "password", std::string( password ) );
}


pool::ObjectRelationalUtilities::ObjectRelationalUtilities() : m_properties(),
                                                               m_flags(),
                                                               m_session(0),
                                                               m_sessionOwned(false),
                                                               m_verbosityLevel(coral::Warning),
                                                               m_authService(0){ 
}

pool::ObjectRelationalUtilities::ObjectRelationalUtilities(coral::ISessionProxy* session): m_properties(),
                                                                                           m_flags(),
                                                                                           m_session(session),
                                                                                           m_sessionOwned(false),
                                                                                           m_verbosityLevel(coral::Warning),
                                                                                           m_authService(0){
}

pool::ObjectRelationalUtilities::~ObjectRelationalUtilities(){
  if(m_session && m_sessionOwned) {
    delete m_session;
    m_session = 0;
  }
  if(m_authService) delete m_authService;
}

void pool::ObjectRelationalUtilities::connect(const std::string& connectionString, coral::AccessMode accessMode){
  coral::ConnectionService connService;
  if(m_session && m_sessionOwned) {
    delete m_session;
    m_session = 0;
  }
  m_sessionOwned = true;
  m_session = connService.connect( connectionString, accessMode);
  if(!m_session) throw pool::ObjectRelationalException("Could not connect to "+connectionString,"ObjectRelationalUtilities::connect");
}

coral::ISessionProxy* pool::ObjectRelationalUtilities::session(){
  if(!m_session) throw pool::ObjectRelationalException("Connection to database not available.","ObjectRelationalUtilities::session");
  return m_session;
}

void pool::ObjectRelationalUtilities::setProperty(const std::string& propertyName,const std::string& propertyValue){
  if(m_properties.find(propertyName)!=m_properties.end())
  {
    m_properties.erase(propertyName);
  }
  m_properties.insert(std::make_pair(propertyName,propertyValue));
}

void pool::ObjectRelationalUtilities::setFlag(const std::string& flagName,bool flag){
  if(m_flags.find(flagName)!=m_flags.end())
  {
    m_flags.erase(flagName);
  }
  m_flags.insert(std::make_pair(flagName,flag));  
}

std::string pool::ObjectRelationalUtilities::getProperty(const std::string& propertyName) const
{
  std::string ret = "";
  std::map<std::string,std::string>::const_iterator iProp =m_properties.find(propertyName);
  if(iProp!=m_properties.end())
  {
    ret = iProp->second;
  }
  return ret;
}

bool pool::ObjectRelationalUtilities::getFlag(const std::string& flagName) const
{
  bool ret = false;
  std::map<std::string,bool>::const_iterator iFlag =m_flags.find(flagName);
  if(iFlag!=m_flags.end())
  {
    ret = iFlag->second;
  }
  return ret;
}

void pool::ObjectRelationalUtilities::loadDictionaries(const std::string& dictionaries){
  if(!dictionaries.empty())
  { 
     std::vector<std::string> dictionaryList;
     stringTokenize(dictionaries,dictionaryList);
     loadLibraries(dictionaryList);
  }
}

void pool::ObjectRelationalUtilities::loadLibraries(std::vector<std::string> libraries){ 
 for ( std::vector<std::string>::const_iterator iLibrary = libraries.begin(); iLibrary != libraries.end(); ++iLibrary ) {
    const std::string& library = *iLibrary;
    coral::MessageStream log( pool::ObjectRelationalAccess::serviceName() );

    Reflex::SharedLibrary libloader( "lib" + library + ".so" );

    log << coral::Debug << "Loading library "<<library<<coral::MessageStream::endmsg;

    if(!libloader.Load()){
      log << coral::Error  <<  "Could not load library " << library << " - " << libloader.Error() << coral::MessageStream::endmsg;
    }
 }
}

void pool::ObjectRelationalUtilities::setMessageVerbosityLevel(coral::MsgLevel level )
{
  coral::MessageStream::setMsgVerbosity(level);
  m_verbosityLevel = level;
}

coral::MsgLevel pool::ObjectRelationalUtilities::messageVerbosityLevel() const {
  return m_verbosityLevel;
}

void pool::ObjectRelationalUtilities::initialize()
{
  std::string verbosityLevel = getProperty(verbosityProperty());
  if(!verbosityLevel.empty())
  {
    if(verbosityLevel==debugVerbosityPropertyValue()) setMessageVerbosityLevel(coral::Debug);
    else if(verbosityLevel==infoVerbosityPropertyValue()) setMessageVerbosityLevel(coral::Info);
  }
  std::string user = getProperty(userNameProperty());
  std::string password = getProperty(passwordProperty());
  if(!user.empty() && !password.empty()){
    coral::ConnectionService connServ;
    if(m_authService) delete m_authService;
    m_authService = new LocalAuthenticationService;
    m_authService->setCredentials(user,password);
    connServ.configuration().setAuthenticationService(*m_authService);
  } 
}

const std::string& pool::ObjectRelationalUtilities::connectionStringProperty(){
  static std::string s_connectionStringProperty = "ConnectionString";
  return s_connectionStringProperty;
}

const std::string& pool::ObjectRelationalUtilities::userNameProperty(){
  static std::string s_userNameProperty = "UserName";
  return s_userNameProperty;
}

const std::string& pool::ObjectRelationalUtilities::passwordProperty(){
  static std::string s_passwordProperty = "Password";
  return s_passwordProperty;
}

const std::string& pool::ObjectRelationalUtilities::authenticationProperty(){
  static std::string s_authenticationProperty = "Authentication";
  return s_authenticationProperty;
}

const std::string& pool::ObjectRelationalUtilities::dictionaryProperty(){
  static std::string s_dictionaryProperty = "Dictionary";
  return s_dictionaryProperty;
}

const std::string& pool::ObjectRelationalUtilities::verbosityProperty(){
  static std::string s_verbosityProperty = "Verbosity";
  return s_verbosityProperty;
}

const std::string& pool::ObjectRelationalUtilities::debugVerbosityPropertyValue(){
  static std::string s_debugVerbosityPropertyValue = "Debug";
  return s_debugVerbosityPropertyValue;
}

const std::string& pool::ObjectRelationalUtilities::infoVerbosityPropertyValue(){
  static std::string s_infoVerbosityPropertyValue = "Info";
  return s_infoVerbosityPropertyValue;
}

 
