#include "XMLDatabaseManagerDriver.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingDescriptor.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingDescriptorElement.h"
#include "ObjectRelationalAccess/ObjectRelationalException.h"
#include "xercesc/util/XMLString.hpp"
#include "xercesc/util/PlatformUtils.hpp"
#include "xercesc/sax2/XMLReaderFactory.hpp"
#include <xercesc/sax2/Attributes.hpp>
#include "xercesc/framework/MemBufInputSource.hpp"
#include <xercesc/util/NameIdPool.hpp>
#include <xercesc/parsers/SAXParser.hpp>
#include <xercesc/validators/DTD/DTDGrammar.hpp>
#include <xercesc/validators/DTD/DTDValidator.hpp>
#include <string>
#include <ostream>

using namespace xercesc;

pool::XMLDatabaseManagerDriver::XMLDatabaseManagerDriver(){}


void pool::XMLDatabaseManagerDriver::init() {
  try {
    XMLPlatformUtils::Initialize();
  } catch (const XMLException& toCatch) {
    char* msg = XMLString::transcode(toCatch.getMessage());
    std::string message(msg);
    XMLString::release(&msg);     
    throw ObjectRelationalException(message,"XMLDatabaseManagerDriver::init");
  }
}

std::string pool::XMLDatabaseManagerDriver::stringValue(const XMLCh* const xmlString){
  if(xmlString){
    char* c_str = XMLString::transcode(xmlString);
    std::string str(c_str);
    XMLString::release(&c_str);
    return str;
  } else {
    return std::string("");
  }     
}

const std::string& pool::XMLDatabaseManagerDriver::databaseTag(){
  static std::string s_databaseTag("PoolDatabase");
  return s_databaseTag;
}

const std::string& pool::XMLDatabaseManagerDriver::containerTag(){
  static std::string s_containerTag("PoolContainer");
  return s_containerTag;
}

const std::string& pool::XMLDatabaseManagerDriver::classTag(){
  static std::string s_classTag("Class");
  return s_classTag;
}

const std::string& pool::XMLDatabaseManagerDriver::nameAttribute(){
  static std::string s_nameAttribute("name");
  return s_nameAttribute;
}

const std::string& pool::XMLDatabaseManagerDriver::mappingVersionAttribute(){
  static std::string s_mappingVersionAttribute("mapping_version");
  return s_mappingVersionAttribute;
}

const std::string& pool::XMLDatabaseManagerDriver::containerTypeAttribute(){
  static std::string s_containerTypeAttribute("tech");
  return s_containerTypeAttribute;
}

const std::string& pool::XMLDatabaseManagerDriver::homogeneousContainerTypeAttributeValue(){
  static std::string s_homogeneousContainerTypeAttributeValue("homogeneous");
  return s_homogeneousContainerTypeAttributeValue;
}

const std::string& pool::XMLDatabaseManagerDriver::policyAttribute(){
  static std::string s_policyAttribute("policy");
  return s_policyAttribute;
}

const std::string& pool::XMLDatabaseManagerDriver::updatePolicyAttributeValue(){
  static std::string s_updatePolicyAttributeValue("update");
  return s_updatePolicyAttributeValue;
}

pool::ObjectRelationalDatabaseDescriptor* pool::XMLDatabaseManagerDriver::read(const std::string& xmlFile){
  this->init();
  
  ObjectRelationalDatabaseDescriptor* dbSetup = new ObjectRelationalDatabaseDescriptor();
  std::auto_ptr<SAX2XMLReader> parser(XMLReaderFactory::createXMLReader());
  std::auto_ptr<DefaultHandler> docHandler(new SAXHandler(dbSetup));
  std::auto_ptr<EntityResolver> entityResolver(new SpecificEntityResolver);
  parser->setContentHandler( docHandler.get() );
  parser->setErrorHandler( docHandler.get() );
  parser->setEntityResolver(entityResolver.get());
  parser->setFeature( XMLUni::fgSAX2CoreValidation, true  );
  parser->setFeature( XMLUni::fgXercesDynamic,      false );
  parser->setFeature( XMLUni::fgSAX2CoreNameSpaces       , true );
  parser->setFeature( XMLUni::fgXercesSchema             , true );
  parser->setFeature( XMLUni::fgXercesSchemaFullChecking , true );
  parser->setFeature( XMLUni::fgSAX2CoreNameSpacePrefixes, true );
  try {
    parser->parse(xmlFile.c_str());
  } catch (const XMLException& toCatch) {
    char* msg = XMLString::transcode(toCatch.getMessage());
    std::string message(msg);
    XMLString::release(&msg);
    delete dbSetup;
    dbSetup = 0;
    throw ObjectRelationalException(message,"XMLDatabaseManagerDriver::read");
  } catch (const SAXParseException& toCatch) {
    char* msg = XMLString::transcode(toCatch.getMessage());
    std::string message(msg);
    XMLString::release(&msg);
    delete dbSetup;
    dbSetup = 0;
    throw ObjectRelationalException(message,"XMLDatabaseManagerDriver::read");
  } catch (const ObjectRelationalException& toCatch) {
    delete dbSetup;
    dbSetup = 0;
    throw ObjectRelationalException(toCatch.what(),"XMLDatabaseManagerDriver::read");
  } catch (...) {
    delete dbSetup;
    dbSetup = 0;
    throw ObjectRelationalException("Unexpected Exception","XMLDatabaseManagerDriver::read");
  }
  return dbSetup;
}

pool::ObjectRelationalDatabaseDescriptor* pool::XMLDatabaseManagerDriver::readFromBuffer(const char* buffer, const char* bufferId){
  this->init();
  
  ObjectRelationalDatabaseDescriptor* dbstruct = new ObjectRelationalDatabaseDescriptor();
  SAX2XMLReader* parser = XMLReaderFactory::createXMLReader();
  DefaultHandler* docHandler = new SAXHandler(dbstruct);
  EntityResolver* entityResolver = new SpecificEntityResolver;
  parser->setContentHandler( docHandler );
  parser->setErrorHandler( docHandler );
  parser->setEntityResolver(entityResolver);
  parser->setFeature( XMLUni::fgSAX2CoreValidation, true  );
  parser->setFeature( XMLUni::fgXercesDynamic,      false );
  parser->setFeature( XMLUni::fgSAX2CoreNameSpaces       , true );
  parser->setFeature( XMLUni::fgXercesSchema             , true );
  parser->setFeature( XMLUni::fgXercesSchemaFullChecking , true );
  parser->setFeature( XMLUni::fgSAX2CoreNameSpacePrefixes, true );

  MemBufInputSource* memBufInputSource = new MemBufInputSource( (const XMLByte*)buffer,
                                                                strlen(buffer),
                                                                bufferId,
                                                                false );
  try {
    parser->parse(*memBufInputSource);
  } catch (const XMLException& toCatch) {
    char* msg = XMLString::transcode(toCatch.getMessage());
    std::string message(msg);
    XMLString::release(&msg);
    delete dbstruct;
    dbstruct = 0;
    throw ObjectRelationalException(message,"XMLDatabaseManagerDriver::readFromBuffer");
  } catch (const SAXParseException& toCatch) {
    char* msg = XMLString::transcode(toCatch.getMessage());
    std::string message(msg);
    XMLString::release(&msg);
    delete dbstruct;
    dbstruct = 0;
    throw ObjectRelationalException(message,"XMLDatabaseManagerDriver::readFromBuffer");
  } catch (const ObjectRelationalException& toCatch) {
    delete dbstruct;
    dbstruct = 0;
    throw ObjectRelationalException(toCatch.what(),"XMLDatabaseManagerDriver::readFromBuffer");
  } catch (...) {
    delete dbstruct;
    dbstruct = 0;
    throw ObjectRelationalException("Unexpected Exception","XMLDatabaseManagerDriver::readFromBuffer");
  }
  delete parser;
  delete docHandler;
  delete entityResolver;
  delete memBufInputSource;
  // Call the termination method
  XMLPlatformUtils::Terminate();
  return dbstruct;
}

bool pool::XMLDatabaseManagerDriver::write(const ObjectRelationalDatabaseDescriptor& databaseStruct, std::ostream& os){
  init();
  static const char*  inMemXML = "\
<?xml version='1.0' encoding='ascii'?>\n\
<!DOCTYPE PoolDatabase SYSTEM \"InMemory\">\n\
<PoolDatabase>\n\
</PoolDatabase >\
";
  static const char*  inMemXMLId = "empty.xml";
  SAXParser::ValSchemes valScheme = SAXParser::Val_Auto;
  DTDValidator* valToUse = new DTDValidator;
  SAXParser* parser = new SAXParser(valToUse);
  parser->setValidationScheme(valScheme);
  parser->setDoNamespaces(true);
  parser->setDoSchema(true);
  parser->setValidationSchemaFullChecking(true);
  std::auto_ptr<EntityResolver> entityResolver(new SpecificEntityResolver);
  parser->setEntityResolver(entityResolver.get());
  MemBufInputSource* memBufIS = 
                     new MemBufInputSource ((const XMLByte*)inMemXML,strlen(inMemXML),inMemXMLId, false);
  int errorCount = 0;
  try {
    parser->parse(*memBufIS);
    errorCount = parser->getErrorCount();
  } catch (const XMLException& e) {
    throw ObjectRelationalException(stringValue(e.getMessage()),"XMLMappingHandler::write");  
  } catch (...){
    throw ObjectRelationalException("Unexpected exception catched ...","XMLMappingHandler::write");  
  }
  DTDGrammar* grammar = (DTDGrammar*) valToUse->getGrammar();
  if(!grammar){
    throw ObjectRelationalException("Grammar initialization failed.","XMLMappingHandler::initGrammar");    
  }

  os << "<?xml version='1.0' encoding=\"UTF-8\"?>" << std::endl;
  os << "<!DOCTYPE PoolDatabase SYSTEM \"InMemory\">" << std::endl;
  os << " <"<<databaseTag()<<" >"<< std::endl;
  const std::vector<ObjectRelationalContainerDescriptor*>& conts = databaseStruct.containers();
  for(std::vector<ObjectRelationalContainerDescriptor*>::const_iterator iC=conts.begin();
      iC!=conts.end();++iC){
    os <<"  <"<<containerTag()<<" name=\""<<(*iC)->name()<<"\"";
    std::string contType = (*iC)->type();
    if(!contType.empty()){
      os<<" tech=\""<<contType<<"\"";
    }
    os<<" >"<<std::endl;
    const std::vector<ObjectRelationalClassDescriptor*>& classes = (*iC)->classStores();
    for(std::vector<ObjectRelationalClassDescriptor*>::const_iterator iClass = classes.begin();
        iClass!=classes.end();++iClass){
      const ObjectRelationalMappingDescriptor* mapping = (*iClass)->mapping();
      if(mapping){
        const ObjectRelationalMappingDescriptorElement* topLevelElement = mapping->topLevelElement();
        if(topLevelElement) {
          for(std::vector<ObjectRelationalMappingDescriptorElement*>::const_iterator iMapElem=topLevelElement->children().begin();
              iMapElem!=topLevelElement->children().end();iMapElem++){
            recursivePrint(*(*iMapElem),*grammar,os,"   ");
          }
        }
      }
    }
    os<<"  </"<<containerTag()<<" >"<<std::endl;
  }
  os<<" </"<<databaseTag()<<" >"<< std::endl;

  delete parser;
  delete memBufIS;
  XMLPlatformUtils::Terminate();

  return true;
}

void pool::XMLDatabaseManagerDriver::recursivePrint(const ObjectRelationalMappingDescriptorElement& mappingElement, DTDGrammar& dtdGrammar,
                                                    std::ostream& os, std::string margin){
  os << margin << "<" << mappingElement.type() << " ";
  // looking up the element in the DTD grammar
  XMLCh* elementName = XMLString::transcode(mappingElement.type().c_str());
  DTDElementDecl* dtdElement = dynamic_cast<DTDElementDecl*>(dtdGrammar.getElemDecl(0,0,elementName,0));
  XMLString::release(&elementName);
  if(dtdElement){
    // retrieve attribute list
    XMLAttDefList& attList = dtdElement->getAttDefList();
    std::map<std::string,std::string>::const_iterator iAttr;
    // print only the attributes defined in the DTD
    while (attList.hasMoreElements()){
      XMLAttDef& curAttDef = attList.nextElement();
      iAttr = mappingElement.attributes().find(stringValue(curAttDef.getFullName()));
      if(iAttr!=mappingElement.attributes().end()){
        std::string attrValue = iAttr->second;
        for(std::string::size_type pos=attrValue.find("<", 0);
            pos!=std::string::npos;
            pos = attrValue.find("<",pos+1)){
          attrValue.replace(pos,1,"&lt;");
        }
        for(std::string::size_type pos=attrValue.find(">", 0);
            pos!=std::string::npos;
            pos = attrValue.find(">",pos+1)){
          attrValue.replace(pos,1,"&gt;");
        }
        os << iAttr->first << "=\"" << attrValue << "\" ";
      }
    }
    int nchilds = mappingElement.children().size();
    if(!nchilds) os << "/";
    os << ">" << std::endl;
    std::vector<ObjectRelationalMappingDescriptorElement*>::const_iterator iMapElem;
    std::string newmargin(margin+"  ");
    for(iMapElem=mappingElement.children().begin();iMapElem!=mappingElement.children().end();iMapElem++){
      recursivePrint(*(*iMapElem),dtdGrammar,os,newmargin);
    }
    if(nchilds){
      os << margin << "</" << mappingElement.type() << " >" << std::endl;
    }
  }
}

pool::XMLDatabaseManagerDriver::SAXHandler::SAXHandler(): m_dbSetup(0),m_containerSetup(0),m_classStoreSetup(0),m_mapping(0),m_currentElement(0){}

pool::XMLDatabaseManagerDriver::SAXHandler::SAXHandler(ObjectRelationalDatabaseDescriptor* dbSetup): m_dbSetup(dbSetup),m_containerSetup(0),m_classStoreSetup(0),m_mapping(0),m_currentElement(0){}

void pool::XMLDatabaseManagerDriver::SAXHandler::startDocument(){}

void pool::XMLDatabaseManagerDriver::SAXHandler::endDocument(){}

void pool::XMLDatabaseManagerDriver::SAXHandler::startElement(const XMLCh* const,
                                                              const XMLCh* const,
                                                              const XMLCh* const qname,
                                                              const Attributes&  attributes  ){
  std::string name(XMLString::transcode(qname));
  if(name!=databaseTag()){
    if (name==containerTag()){
      std::string containerName = 
        stringValue(attributes.getValue(XMLString::transcode(nameAttribute().c_str())));
      std::string containerType =
        stringValue(attributes.getValue(XMLString::transcode(containerTypeAttribute().c_str())));
      std::string policy =
        stringValue(attributes.getValue(XMLString::transcode(policyAttribute().c_str())));
      bool recreate = true;
      if(policy==updatePolicyAttributeValue()){
        recreate = false;
      }
      m_containerSetup = &m_dbSetup->addContainer(containerName,containerType);
      m_containerSetup->setRecreatePolicy(recreate);
    } else {
      if (name==classTag()){
        std::string className =
          stringValue(attributes.getValue(XMLString::transcode(nameAttribute().c_str())));
        std::string mappingVersion =
          stringValue(attributes.getValue(XMLString::transcode(mappingVersionAttribute().c_str())));
        m_containerSetup->addClassStore(className,mappingVersion);
        m_classStoreSetup = m_containerSetup->addClassStore(className,mappingVersion);
        m_mapping = m_classStoreSetup->newMapping();
        m_currentElement = m_mapping->setTopLevelElement(ObjectRelationalMappingDescriptorElement::mappingTag());
        m_currentElement = m_currentElement->addChild(ObjectRelationalMappingDescriptorElement::classTag());
        for( unsigned int i = 0; i < attributes.getLength(); i++ ) {
          std::string  name(XMLString::transcode( attributes.getQName(i) ));
          std::string  value(XMLString::transcode( attributes.getValue(i) ));
          m_currentElement->setAttribute(name,value);
        }
        m_currentElement->setAttribute(ObjectRelationalMappingDescriptorElement::poolContainerNameAttr(),m_containerSetup->name());
      } else {
        if(m_currentElement){
          m_currentElement = m_currentElement->addChild(name);
          for( unsigned int i = 0; i < attributes.getLength(); i++ ) {
            std::string  attrName(XMLString::transcode( attributes.getQName(i) ));
            std::string  attrValue(XMLString::transcode( attributes.getValue(i) ));
            m_currentElement->setAttribute(attrName,attrValue);
          }
        }
      }
    }
  }
}

void pool::XMLDatabaseManagerDriver::SAXHandler::endElement(const XMLCh* const,
                                                            const XMLCh* const,
                                                            const XMLCh* const){
  if(m_currentElement!=0){
    m_currentElement = m_currentElement->parent();
  }
}

void pool::XMLDatabaseManagerDriver::SAXHandler::error(const SAXParseException& exception){
  char* msg = XMLString::transcode(exception.getMessage());
  std::string message(msg);
  XMLString::release(&msg);
  throw ObjectRelationalException(message,"XMLDatabaseManagerDriver::SAXHandler::error");
}
        
void pool::XMLDatabaseManagerDriver::SAXHandler::fatalError(const SAXParseException& exception){
  char* msg = XMLString::transcode(exception.getMessage());
  std::string message(msg);
  XMLString::release(&msg);
  throw ObjectRelationalException(message,"XMLDatabaseManagerDriver::SAXHandler::fatalError");
}

xercesc::InputSource* pool::XMLDatabaseManagerDriver::SpecificEntityResolver::resolveEntity (const XMLCh* const,
					                                                const XMLCh* const){
  static const char* inmemdtdId = "redirectinmem.dtd";

  static std::string sinmemdtd =
    std::string("<!ELEMENT PoolDatabase (PoolContainer)+>\n")+
    std::string("<!ELEMENT PoolContainer (Class)*>\n")+
    std::string("<!ATTLIST PoolContainer name CDATA #REQUIRED>\n")+
    std::string("<!ATTLIST PoolContainer tech (homogeneous|polymorphic) #IMPLIED>\n")+
    std::string("<!ATTLIST PoolContainer policy (recreate|update) #IMPLIED>\n")+
    std::string("<!ELEMENT Class (Primitive|Blob|Container|Array|CArray|InlineCArray|Object|PoolPointer)*>\n")+
    std::string("<!ATTLIST Class name CDATA #REQUIRED>\n")+
    std::string("<!ATTLIST Class table CDATA #IMPLIED>\n")+
    std::string("<!ATTLIST Class id_columns NMTOKENS #IMPLIED>\n")+
    std::string("<!ATTLIST Class mapping_version CDATA #IMPLIED>\n")+
    std::string("<!ATTLIST Class use_mapping CDATA #IMPLIED>\n")+
    std::string("<!ATTLIST Class base_mapping CDATA #IMPLIED>\n")+
    std::string("<!ELEMENT Primitive EMPTY>\n")+
    std::string("<!ATTLIST Primitive name CDATA #REQUIRED>\n")+
    std::string("<!ATTLIST Primitive column CDATA #IMPLIED>\n")+
    std::string("<!ELEMENT Blob EMPTY>\n")+
    std::string("<!ATTLIST Blob name CDATA #REQUIRED>\n")+
    std::string("<!ATTLIST Blob column CDATA #IMPLIED>\n")+
    std::string("<!ELEMENT Container (Primitive|Blob|Container|Array|CArray|InlineCArray|Object|PoolPointer)*>\n")+
    std::string("<!ATTLIST Container name CDATA #REQUIRED>\n")+
    std::string("<!ATTLIST Container table CDATA #IMPLIED>\n")+
    std::string("<!ATTLIST Container position_column CDATA #IMPLIED>\n")+
    std::string("<!ATTLIST Container id_columns NMTOKENS #IMPLIED>\n")+
    std::string("<!ELEMENT Array (Primitive|Blob|Container|Array|CArray|InlineCArray|Object|PoolPointer)?>\n")+
    std::string("<!ATTLIST Array name CDATA #REQUIRED>\n")+
    std::string("<!ATTLIST Array table CDATA #IMPLIED>\n")+
    std::string("<!ATTLIST Array position_column CDATA #IMPLIED>\n")+
    std::string("<!ATTLIST Array id_columns NMTOKENS #IMPLIED>\n")+
    std::string("<!ELEMENT CArray (Primitive|Blob|Container|Array|CArray|InlineCArray|Object|PoolPointer)?>\n")+
    std::string("<!ATTLIST CArray name CDATA #REQUIRED>\n")+
    std::string("<!ATTLIST CArray table CDATA #IMPLIED>\n")+
    std::string("<!ATTLIST CArray position_column CDATA #IMPLIED>\n")+
    std::string("<!ATTLIST CArray id_columns NMTOKENS #IMPLIED>\n")+
    std::string("<!ELEMENT Object (Primitive|Blob|Container|Array|CArray|InlineCArray|Object|PoolPointer)*>\n")+
    std::string("<!ATTLIST Object name CDATA #REQUIRED>\n")+
    std::string("<!ELEMENT InlineCArray (Primitive|Blob|Container|Array|CArray|InlineCArray|Object|PoolPointer)*>\n")+
    std::string("<!ATTLIST InlineCArray name CDATA #REQUIRED>\n")+
    std::string("<!ELEMENT PoolPointer (Primitive|Blob|Container|Array|CArray|InlineCArray|Object|PoolPointer)*>\n")+
    std::string("<!ATTLIST PoolPointer name CDATA #REQUIRED>\n")+
    std::string("<!ATTLIST PoolPointer table CDATA #IMPLIED>\n")+
    std::string("<!ATTLIST PoolPointer id_columns NMTOKENS #IMPLIED>\n");

  const char* inmemdtd = sinmemdtd.c_str();
  
    //  Create MemBufferInputSource from the buffer containing the XML
    //  statements.
    //
    //  NOTE: We are using strlen() here, since we know that the chars in
    //  our hard coded buffer are single byte chars!!! The parameter wants
    //  the number of BYTES, not chars, so when you create a memory buffer
    //  give it the byte size (which just happens to be the same here.)
    //

  MemBufInputSource* memBufIS = 
                     new MemBufInputSource((const XMLByte*)inmemdtd, strlen(inmemdtd), inmemdtdId, false);
  return memBufIS;
}


