#ifndef POOL_PTR_H
#define POOL_PTR_H

#include "POOLCore/Exception.h"
#include <assert.h>

namespace pool    {

   /**
   * @class IPtrLoader Ptr.h POOLCore/Ptr.h
   *  
   * Interface for the lazy loading of the embedded object.
   * Implementations are provided by the storage manager systems.
   */
  class IPtrLoader {
    public:

    // destructor
    virtual ~IPtrLoader(){}

    // triggers the data loading
    virtual void* load() const=0;

    // notify the underlying storage system that the embedded object has been destructed.
    virtual void notify() const=0;

    // invalidates the current loader. Called by the underlying service at his destruction time.
    virtual void invalidate()=0;

    // queries the validity of the current relation with the underlying storage system
    virtual bool isValid() const=0;
    
  };
  
   /**
   * @class Ptr Ptr.h POOLCore/Ptr.h
   *  
   * Templated class for the persistency of associated objects. The embedded object is treated as  
   * a normal C++ pointer in the writing mode, without to require an explicit write request on it.
   * In the reading mode, the embedded object is loaded only at access time (lazy loading).
   * Supported only by the RelationalStorageService backend.
   */
  template <typename T> class Ptr {

    public:
    
    // default constructor
    Ptr();
    
    // from a real pointer
    explicit Ptr(T* anObject);
    
    // copy constructor
    Ptr(const Ptr<T>&); 

    // extended copy constructor
    template <class C> Ptr(const Ptr<C>&);  

    // destructor
    virtual ~Ptr();

    // Assignment operator with real pointer 
    Ptr<T>& operator=(T*);

    // assignment operator
    Ptr<T>& operator=(const Ptr<T>&);

    // extended assignment operator 
    template <class C> Ptr<T>& operator=(const Ptr<C>&);

    // copy operator for up/down casting 
    template <class C> Ptr<T>& cast(const Ptr<C>&);

    // dereference operator
    T* operator->() const;

    // dereference operator
    T& operator*() const;

    // implicit bool conversion
    operator bool () const;

    // return the real pointer
    T* get() const;

    // return the naked pointer, without to trigger the loading.
    void* address() const;

    // 'not' operator for consistency with pointers common use
    bool operator!() const;

    // equality operator between refs
    template <class C>
    bool operator==(const Ptr<C>& aPtr) const {
      return m_ptr == static_cast<C*>(aPtr.address());
    }
    template <class C>
    bool operator!=(const Ptr<C>& aPtr) const {
      return !(this->operator==(aPtr));     
    }
    
    public:

    // clear the embedded pointer and invalidates the loader,if any.
    void reset();

    // returns the current loader
    IPtrLoader* loader() const {
      return m_loader;
    }

    // returns true if the emebedded object data have been loaded.
    bool isLoaded() const {
      return m_isLoaded;
    }
    
    private:
    
    // pointer with throw exception clause
    T* ptr(bool throw_flag) const;

  private:

    // embedded object pointer
    mutable T* m_ptr;

    // data loader, istalled by the storage system
    mutable IPtrLoader* m_loader;

    // object loaded flag
    mutable bool m_isLoaded;
    
};

}

// default ctor
template <class T> pool::Ptr<T>::Ptr() : 
                                 m_ptr(0),m_loader(0),m_isLoaded(false) {}
// ctor from a real pointer
template <class T> pool::Ptr<T>::Ptr(T* anObject) : 
                                 m_ptr(anObject),m_loader(0),m_isLoaded(true) {}

template <class T> pool::Ptr<T>::Ptr(const Ptr<T>& aPtr) :
                                 m_ptr(aPtr.m_ptr),m_loader(aPtr.m_loader),m_isLoaded(false){}

template <class T> 
template <class C> pool::Ptr<T>::Ptr(const Ptr<C>& aPtr) :
                         m_ptr(static_cast<C*>(aPtr.address())),m_loader(aPtr.loader()),m_isLoaded(aPtr.isLoaded()) {
  // compile-time type checking
  C* c = 0; T* t(c); assert(t==0);
}

// destrctor
template <class T> pool::Ptr<T>::~Ptr() {
  if(m_loader) {
    m_loader->notify();
    delete m_loader;
  }
}

// pointer assignment operator
template <class T> pool::Ptr<T>& pool::Ptr<T>::operator=(T* aPtr) {
  reset();
  m_ptr = aPtr;
  m_isLoaded = true;
  return *this;
}

template <class T> pool::Ptr<T>& pool::Ptr<T>::operator=(const Ptr<T>& aPtr){
  reset();
  m_ptr = aPtr.m_ptr;
  m_loader = aPtr.m_loader;
  m_isLoaded = aPtr.m_isLoaded;
  return *this;  
}

template <class T> 
template <class C> pool::Ptr<T>& 
                            pool::Ptr<T>::operator=(const Ptr<C>& aPtr){
  C* c = 0; T* t(c); assert(t==0);
  reset();
  m_ptr = static_cast<C*>(aPtr.address());
  m_loader = aPtr.loader();
  m_isLoaded = aPtr.isLoaded();
  return *this;  
}

template <class T> template <class C> pool::Ptr<T>& 
pool::Ptr<T>::cast(const Ptr<C>& aPtr){
  reset();
  m_ptr = dynamic_cast<T*>(static_cast<C*>(aPtr.address()));
  m_loader = aPtr.loader();
  m_isLoaded = aPtr.isLoaded();
  return *this;  
}

// dereference operator
template <class T> T* pool::Ptr<T>::operator->() const {
  return ptr(true);
}

// dereference operator
template <class T> T& pool::Ptr<T>::operator*() const {
  return *ptr(true);
}

// return the real pointer
template <class T> T* pool::Ptr<T>::get() const  {
  return ptr(false);  
}

// return the naked pointer
template <class T> void* pool::Ptr<T>::address() const  {
  return m_ptr;  
}

// T* conversion operator
template <class T> pool::Ptr<T>::operator bool() const {
  return ptr(false);
}

// 'not' operator for consistency with pointers common use
template <class T> bool pool::Ptr<T>::operator!() const {
  return ptr(false)==0;
}

template <class T> void pool::Ptr<T>::reset(){
  m_ptr = 0;
  if(m_loader) {
    m_loader->notify();
    delete m_loader;
  }
  m_loader = 0;
  m_isLoaded = false;
}

// return the real pointer
template <class T> T* pool::Ptr<T>::ptr(bool throwFlag) const {
  if(!m_ptr){
    if(!m_loader){
      if(throwFlag) throw pool::Exception("Loader is not installed.","Ptr::ptr()","POOLCore");
    }
    if(!m_isLoaded){
      m_ptr = static_cast<T*>(m_loader->load());
      m_isLoaded = true;
    }
  }
  if(!m_ptr){
    if(throwFlag) throw pool::Exception("Underlying pointer is null.","Ptr::ptr()","POOLCore");    
  }
  return m_ptr;
}

#endif  // POOL_PTR_H
