#include "DBLock.h"


#include "RelationalCollectionNames.h"
#include "RelationalCollectionBindVariables.h"

#include "RelationalAccess/TableDescription.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITablePrivilegeManager.h"
#include "RelationalAccess/ITransaction.h"
#include "RelationalAccess/ITableDataEditor.h"

#include "CoralBase/MessageStream.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/TimeStamp.h"


#ifdef WIN32
#   include <windows.h>
#   define sleep(x) Sleep((x)*1000)   /* Windows version */
#else
#   include <unistd.h>
#endif


using namespace pool;
using namespace std;
using namespace pool::RelationalCollection;


 
DBLock::DBLock( coral::ISessionProxy& session, std::string collName )
      : m_session( session ),
        m_name( collName ),
        m_locked( false )
{
}
 
 
DBLock::~DBLock()
{
   if( m_locked )
      unlock();
}


void
DBLock::lock( bool XTrans )
{
   using namespace pool::RelationalCollection;
   coral::MessageStream log( "pool::RelationalCollection::DBLock" );
    
   coral::ISchema& nominalSchema = m_session.nominalSchema();
   if( !nominalSchema.existsTable( RelationalCollectionNames::nameOfCollectionLockTable() ) ) {
      // create the table where the locking is done
      coral::TableDescription description( RelationalCollectionNames::nameOfCollectionLockTable() );
      description.setName( RelationalCollectionNames::nameOfCollectionLockTable() );
      description.insertColumn( 
         RelationalCollectionNames::entryIDVarInCollectionLockTable(),
         coral::AttributeSpecification::typeNameForType<std::string>(), 50, false );
      description.insertColumn( 
         RelationalCollectionNames::collectionNameVarInCollectionLockTable(),
         coral::AttributeSpecification::typeNameForType<std::string>(), 500, false );
      description.insertColumn( 
         RelationalCollectionNames::clientInfoVarInCollectionLockTable(),
         coral::AttributeSpecification::typeNameForType<std::string>(), 500, false );
      description.insertColumn( 
         RelationalCollectionNames::lockTypeVarInCollectionLockTable(),
         coral::AttributeSpecification::typeNameForType<std::string>(), 20, false );
      description.insertColumn( 
         RelationalCollectionNames::timestampVarInCollectionLockTable(),
         coral::AttributeSpecification::typeNameForType<coral::TimeStamp>(), 2, false );
 
      description.setPrimaryKey( RelationalCollectionNames::entryIDVarInCollectionLockTable() );
      log << coral::Debug << "  Creating lock TABLE:" << RelationalCollectionNames::nameOfCollectionLockTable() <<  coral::MessageStream::endmsg;
      nominalSchema.createTable( description ).privilegeManager().grantToPublic( coral::ITablePrivilegeManager::Select );
   }
 
   do { 
      coral::ITable& lockTable = m_session.nominalSchema().tableHandle(
         RelationalCollectionNames::nameOfCollectionLockTable() );
      coral::IQuery* query = lockTable.newQuery();
      query->addToOutputList( RelationalCollectionNames::entryIDVarInCollectionLockTable() );
      query->addToOutputList( RelationalCollectionNames::collectionNameVarInCollectionLockTable() );
      query->addToOutputList( RelationalCollectionNames::lockTypeVarInCollectionLockTable() );
      query->addToOutputList( RelationalCollectionNames::timestampVarInCollectionLockTable() );
 
      coral::AttributeList whereDataForLock;
      whereDataForLock.extend<std::string>(
         RelationalCollectionNames::entryIDVarInCollectionLockTable() );
      whereDataForLock.begin()->data<std::string>() = "LOCK";
      string       whereClauseForLock = RelationalCollectionNames::entryIDVarInCollectionLockTable() + " = :" + RelationalCollectionNames::entryIDVarInCollectionLockTable();
      query->setCondition( whereClauseForLock, whereDataForLock );
      query->limitReturnedRows( 1, 0 );
      query->setForUpdate();
      log << coral::Debug << " Checking the LOCK table:" << RelationalCollectionNames::nameOfCollectionLockTable() << coral::MessageStream::endmsg;
      coral::ICursor& cursor = query->execute();
 
      if( !cursor.next() )    {
         log << coral::Debug << "  Creating lock ROW in " << RelationalCollectionNames::nameOfCollectionLockTable() <<  coral::MessageStream::endmsg;
         // no entry - new table or row removed by hand.  Add the row
         coral::AttributeList rowBuffer;
         rowBuffer.extend< std::string >( RelationalCollectionNames::entryIDVarInCollectionLockTable() );
         rowBuffer.extend< std::string >( RelationalCollectionNames::collectionNameVarInCollectionLockTable() );
         rowBuffer.extend< std::string >( RelationalCollectionNames::lockTypeVarInCollectionLockTable() );
         rowBuffer.extend< coral::TimeStamp >( RelationalCollectionNames::timestampVarInCollectionLockTable() );
 
         rowBuffer[0].data< std::string >() = "LOCK";
         rowBuffer[1].data< std::string >() = m_name;
         rowBuffer[2].data< std::string >() = XTrans? "PERM" : "TEMP";
         rowBuffer[3].data< coral::TimeStamp >() = coral::TimeStamp::now();
         lockTable.dataEditor().insertRow( rowBuffer );
 
      } else {
         string lockType =  cursor.currentRow()[RelationalCollectionNames::lockTypeVarInCollectionLockTable()].data< std::string >();
         log << coral::Debug << "  Current lock type is " << lockType << coral::MessageStream::endmsg;
         if( lockType == "PERM" ) {
            delete query; query = 0;
            // release locks and wait some
            m_session.transaction().commit();
            bool forUpdate(false);
            m_session.transaction().start( forUpdate );
     
            log << coral::Warning << "  Permanent collection database lock detected.  Sleep 5s" << coral::MessageStream::endmsg;
            sleep(5);
            continue;
             
         } else {
            log << coral::Debug << "  Setting database lock to " << (  XTrans? "'PERM'" : "'TEMP'" ) << coral::MessageStream::endmsg;
            // update the row in the LOCK table to set databse lock and store lock type
            string    setClause =
               RelationalCollectionNames::collectionNameVarInCollectionLockTable()
               + " = '" + m_name + "'"
               + ", " +  RelationalCollectionNames::timestampVarInCollectionLockTable()
               + " = :newtime "
               + ", " +  RelationalCollectionNames::lockTypeVarInCollectionLockTable()
               + " = " + (  XTrans? "'PERM'" : "'TEMP'" );
            whereDataForLock.extend< coral::TimeStamp >( "newtime" );
            whereDataForLock[1].data< coral::TimeStamp >() = coral::TimeStamp::now();
            lockTable.dataEditor().updateRows( setClause, whereClauseForLock, whereDataForLock );
         }
      }
      delete query;
      m_locked = true;
      return;
   }while( true );
}
 
 
void
DBLock::unlock()
{
   using namespace pool::RelationalCollection;
 
   coral::ISchema& nominalSchema = m_session.nominalSchema();
   coral::ITable& lockTable = nominalSchema.tableHandle(
      RelationalCollectionNames::nameOfCollectionLockTable() );
 
   string    setClause =
      RelationalCollectionNames::lockTypeVarInCollectionLockTable() + " = 'NONE'";
 
   coral::AttributeList whereDataForLock;
   whereDataForLock.extend<std::string>(
      RelationalCollectionNames::entryIDVarInCollectionLockTable() );
   whereDataForLock.begin()->data<std::string>() = "LOCK";
   string       whereClauseForLock = RelationalCollectionNames::entryIDVarInCollectionLockTable() + " = :" + RelationalCollectionNames::entryIDVarInCollectionLockTable();
 
   lockTable.dataEditor().updateRows( setClause, whereClauseForLock, whereDataForLock ); 
}
 
