#include "MappingManager.h"
#include "MappingInformation.h"
#include "RelationalStorageServiceException.h"
#include "RelationalStorageServiceNames.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingPersistency.h"
#include "ObjectRelationalAccess/ObjectRelationalMapping.h"
#include "ObjectRelationalAccess/ObjectRelationalDefaultMappingBuilder.h"
#include "ObjectRelationalAccess/ObjectRelationalTableNameRationalizer.h"
#include "ObjectRelationalAccess/ObjectRelationalTableVariableNameRationalizer.h"
#include "ObjectRelationalAccess/ObjectRelationalNames.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "CoralBase/MessageStream.h"
#include "RVersion.h"

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
using namespace ROOT;
#endif

pool::RelationalStorageService::MappingManager::MappingManager( coral::ISchema& schema ):
  m_schema( schema ),
  m_persistency( new pool::ObjectRelationalMappingPersistency( schema ) ),
  m_classMappingsForHomogeneousContainers(),
  m_classMappingsForPolymorphicContainers()
{}


pool::RelationalStorageService::MappingManager::~MappingManager()
{
  this->clear();
  delete m_persistency;
}


void
pool::RelationalStorageService::MappingManager::clear()
{
  for ( std::map< const TypeH, std::map< std::string, pool::RelationalStorageService::MappingInformation* > >::iterator iMappingsForClass = m_classMappingsForHomogeneousContainers.begin();
        iMappingsForClass != m_classMappingsForHomogeneousContainers.end(); ++iMappingsForClass ) {
    for ( std::map< std::string, pool::RelationalStorageService::MappingInformation* >::iterator iMappingForVersion = iMappingsForClass->second.begin();
          iMappingForVersion != iMappingsForClass->second.end(); ++iMappingForVersion ) {
      delete iMappingForVersion->second;
    }
  }
  m_classMappingsForHomogeneousContainers.clear();

  for ( std::map< const TypeH, std::map< std::string, pool::RelationalStorageService::MappingInformation* > >::iterator iMappingsForClass = m_classMappingsForPolymorphicContainers.begin();
        iMappingsForClass != m_classMappingsForPolymorphicContainers.end(); ++iMappingsForClass ) {
    for ( std::map< std::string, pool::RelationalStorageService::MappingInformation* >::iterator iMappingForVersion = iMappingsForClass->second.begin();
          iMappingForVersion != iMappingsForClass->second.end(); ++iMappingForVersion ) {
      delete iMappingForVersion->second;
    }
  }
  m_classMappingsForPolymorphicContainers.clear();
}




bool
pool::RelationalStorageService::MappingManager::isTableEmpty( const std::string& tableName )
{
  coral::ITable& table = m_schema.tableHandle( tableName );
  std::auto_ptr< coral::IQuery > query( table.newQuery() );
  query->limitReturnedRows( 1, 0 );
  coral::ICursor& cursor = query->execute();
  if ( cursor.next() ) return false;
  else return true;
}



const pool::RelationalStorageService::MappingInformation&
pool::RelationalStorageService::MappingManager::mappingForPolymorphicContainer( const std::string& containerName, const TypeH& dictionary )
{
  std::map< std::string, pool::RelationalStorageService::MappingInformation* >& mappingsForClass = m_classMappingsForPolymorphicContainers[ dictionary ];
  /// Needed to check if the mapping is in use already
  std::map< std::string, pool::RelationalStorageService::MappingInformation* >& mappingsForHomogeneousContainers = m_classMappingsForHomogeneousContainers[ dictionary ];
  
  if ( ! mappingsForClass.empty() ) {
    return *( mappingsForClass.rbegin()->second ); // Get the most recent
  }

  // The class name
  std::string className = dictionary.Name(Reflex::SCOPED);

  Reflex::PropertyList classProps = dictionary.Properties();
  std::string classVersion = pool::ObjectRelationalAccess::ObjectRelationalNames::defaultClassVersion(className);
  if(classProps.HasProperty(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionPropertyNameInDictionary())){
    classVersion = classProps.PropertyAsString(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionPropertyNameInDictionary());
  }

  m_persistency->initialize();
  std::string mappingVersion = m_persistency->selectMappingVersion( classVersion, containerName );
  if(!mappingVersion.empty()){
    const pool::ObjectRelationalMapping& mappingStructure = m_persistency->retrieveMapping( mappingVersion );
    pool::ObjectRelationalMapping::const_iterator iMappingElement = mappingStructure.find( className );
    if ( iMappingElement != mappingStructure.end() ) {
      // Make sure that this mapping is not used by a polymorphic or another homogeneous container (it should not at this level!).
      if ( mappingsForHomogeneousContainers.find( mappingVersion ) == mappingsForHomogeneousContainers.end() &&
           mappingsForClass.find( mappingVersion ) == mappingsForClass.end() ) {
        // OK, the mapping is valid
        return *( mappingsForClass.insert(
                  std::make_pair( mappingVersion,
                                  new pool::RelationalStorageService::MappingInformation( iMappingElement->second,
                                                                                          mappingVersion ) ) ).first->second );
      } else {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Mapping version \"" + mappingVersion +
                                                                                 "\" for class \"" + className +
                                                                                 "\" is already used by an other container.",
                                                                                 "MappingManager::mappingForPolymorphicContainer" );
      }
    } else {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "No entry for class \"" + className +
                                                                               "\" in mapping version \"" + mappingVersion + "\"",
                                                                               "MappingManager::mappingForPolymorphicContainer" );
    }
  }
  

  std::map<std::string,std::string> existingMappingVersions = m_persistency->retrieveMappingVersions();
  // Create a new mapping using the default rules
  std::string newMappingVersion = "";
  for ( int iteration = 0;; ++iteration ) {
    newMappingVersion = pool::RelationalStorageService::RelationalStorageServiceNames::mappingVersionForClass( className,
                                                                                                               iteration );
    bool found = false;
    for ( std::map<std::string,std::string>::reverse_iterator iVersion = existingMappingVersions.rbegin();
          iVersion != existingMappingVersions.rend(); ++iVersion ) {
      if ( iVersion->first == newMappingVersion ) {
        found = true;
        break;
      }
    }
    if ( ! found ) break;
  }

  pool::ObjectRelationalMapping newMapping( newMappingVersion );
  pool::ObjectRelationalDefaultMappingBuilder builder;
  if ( ! builder.processClass( newMapping, dictionary ) ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not generate a mapping for class \"" + className + "\"",
                                                                             "MappingManager::mappingForPolymorphicContainer" );
  }

  // Rationalize the table and variable names.
  pool::ObjectRelationalTableNameRationalizer rationalizer( m_schema );
  rationalizer.rationalizeMappingElement( newMapping.begin()->second );
  pool::ObjectRelationalTableVariableNameRationalizer vrationalizer;
  vrationalizer.rationalizeMappingElement( newMapping.begin()->second );
  
  // Store and materialize the new mapping.
  if ( ! ( m_persistency->storeMapping( newMapping, classVersion, containerName ) && m_persistency->materializeMapping( newMappingVersion, false ) ) ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not store or materialize mapping for class \"" + className + "\"",
                                                                             "MappingManager::mappingForPolymorphicContainer" );
  }

  return *( mappingsForClass.insert( std::make_pair( newMappingVersion,
                                                     new pool::RelationalStorageService::MappingInformation( m_persistency->retrieveMapping( newMappingVersion ).begin()->second,
                                                                                                             newMappingVersion ) ) ).first->second );
}

const pool::RelationalStorageService::MappingInformation&
pool::RelationalStorageService::MappingManager::mappingForNewHomogeneousContainer( const std::string& containerName, const TypeH& dictionary )
{
  std::map< std::string, pool::RelationalStorageService::MappingInformation* >& mappingsForClass = m_classMappingsForHomogeneousContainers[ dictionary ];
  /// Needed to check if the mapping is in use already
  std::map< std::string, pool::RelationalStorageService::MappingInformation* >& mappingsForPolymorphicContainers = m_classMappingsForPolymorphicContainers[ dictionary ];

  // The class name
  std::string className = dictionary.Name(Reflex::SCOPED);
  
  Reflex::PropertyList classProps = dictionary.Properties();
  std::string classVersion = pool::ObjectRelationalAccess::ObjectRelationalNames::defaultClassVersion(className);
  if(classProps.HasProperty(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionPropertyNameInDictionary())){
    classVersion = classProps.PropertyAsString(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionPropertyNameInDictionary());
  }

  m_persistency->initialize();
  std::string mappingVersion = m_persistency->selectMappingVersion( classVersion, containerName );
  if(!mappingVersion.empty()){
    const pool::ObjectRelationalMapping& mappingStructure = m_persistency->retrieveMapping( mappingVersion );
    pool::ObjectRelationalMapping::const_iterator iMappingElement = mappingStructure.find( className );
    if ( iMappingElement != mappingStructure.end() ) {
      
      // Make sure that this mapping is not used by a polymorphic or another homogeneous container.
      if ( mappingsForPolymorphicContainers.find( mappingVersion ) == mappingsForPolymorphicContainers.end() &&
         mappingsForClass.find( mappingVersion ) == mappingsForClass.end() ) {   
        // OK, this is a valid mapping.
        return *( mappingsForClass.insert(
                  std::make_pair( mappingVersion,
                                  new pool::RelationalStorageService::MappingInformation( iMappingElement->second,
                                                                                          mappingVersion ) ) ).first->second );
      } else {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Mapping version \"" + mappingVersion +
                                                                                 "\" for class \"" + className +
                                                                                 "\" is already used by an other container.",
                                                                                 "MappingManager::mappingForNewHomogeneousContainer" );
      }
    } else {
      throw pool::RelationalStorageService::RelationalStorageServiceException( "No entry for class \"" + className +
                                                                               "\" in mapping version \"" + mappingVersion + "\"",
                                                                               "MappingManager::mappingForNewHomogeneousContainer" );
    }
  }
  
  
  std::map<std::string,std::string> existingMappingVersions = m_persistency->retrieveMappingVersions();
  // Create a new mapping using the default rules
  std::string newMappingVersion = "";
  for ( int iteration = 0;; ++iteration ) {
    newMappingVersion = pool::RelationalStorageService::RelationalStorageServiceNames::mappingVersionForClass( className,
                                                                                                               iteration );
    bool found = false;
    for ( std::map<std::string,std::string>::reverse_iterator iVersion = existingMappingVersions.rbegin();
          iVersion != existingMappingVersions.rend(); ++iVersion ) {
      if ( iVersion->first == newMappingVersion ) {
        found = true;
        break;
      }
    }
    if ( ! found ) break;
  }

  pool::ObjectRelationalMapping newMapping( newMappingVersion );
  pool::ObjectRelationalDefaultMappingBuilder builder;
  if ( ! builder.processClass( newMapping, dictionary ) ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not generate a mapping for class \"" + className + "\"",
                                                                             "MappingManager::mappingForNewHomogeneousContainer" );
  }

  // Rationalize the table and variable names.
  pool::ObjectRelationalTableNameRationalizer rationalizer( m_schema );
  rationalizer.rationalizeMappingElement( newMapping.begin()->second );
  pool::ObjectRelationalTableVariableNameRationalizer vrationalizer;
  vrationalizer.rationalizeMappingElement( newMapping.begin()->second );
  
  // Store and materialize the new mapping.
  if ( ! ( m_persistency->storeMapping( newMapping, classVersion, containerName ) && m_persistency->materializeMapping( newMappingVersion, false ) ) ) {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not store or materialize mapping for class \"" + className + "\"",
                                                                             "MappingManager::mappingForNewHomogeneousContainer" );
  }

  return *( mappingsForClass.insert( std::make_pair( newMappingVersion,
                                                     new pool::RelationalStorageService::MappingInformation( m_persistency->retrieveMapping( newMappingVersion ).begin()->second,
                                                                                                             newMappingVersion ) ) ).first->second );
}



const pool::RelationalStorageService::MappingInformation&
pool::RelationalStorageService::MappingManager::existingMappingForHomogeneousContainer( const std::string& containerName,
                                                                                        const std::string& className,
                                                                                        const std::string& nominalVersion )
{
  const TypeH dictionary = TypeH::ByName( className );

  std::string classVersion = pool::ObjectRelationalAccess::ObjectRelationalNames::defaultClassVersion(className);
  Reflex::PropertyList classProps = dictionary.Properties();
  if(classProps.HasProperty(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionPropertyNameInDictionary())){
    classVersion = classProps.PropertyAsString(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionPropertyNameInDictionary());
  }

  std::string mappingVersion = m_persistency->selectMappingVersion( classVersion, containerName );
  if(mappingVersion.empty()){
    mappingVersion = nominalVersion;
    // if nominalVersion is not considered (as it should be when schema is allowed to evolve), the exception has to be thrown.  
    //throw pool::RelationalStorageService::RelationalStorageServiceException( "No mapping version available for container \"" + containerName +
    //                                                                         "\" and class version \""+classVersion+"\".",
    //                                                                           "MappingManager::existingMappingForHomogeneousContainer" );
  }
  
  std::map< std::string, pool::RelationalStorageService::MappingInformation* >& mappingsForClass = m_classMappingsForHomogeneousContainers[ dictionary ];

  const pool::ObjectRelationalMapping& mappingStructure = m_persistency->retrieveMapping( mappingVersion );
  pool::ObjectRelationalMapping::const_iterator iMappingElement = mappingStructure.find( className );
  if ( iMappingElement != mappingStructure.end() ) {
    return *( mappingsForClass.insert( std::make_pair( mappingVersion,
                                                       new pool::RelationalStorageService::MappingInformation( iMappingElement->second,
                                                                                                               mappingVersion ) ) ).first->second );
  }
  else {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "No entry for class \"" + className +
                                                                             "\" in mapping version \"" + mappingVersion + "\"",
                                                                             "MappingManager::existingMappingForHomogeneousContainer" );
  }
}



const pool::RelationalStorageService::MappingInformation&
pool::RelationalStorageService::MappingManager::existingMappingForPolymorphicContainer( const std::string& containerName,
                                                                                        const TypeH& dictionary,
                                                                                        const std::string& )
{
  std::string className = dictionary.Name(Reflex::SCOPED);
  Reflex::PropertyList classProps = dictionary.Properties();
  std::string classVersion = pool::ObjectRelationalAccess::ObjectRelationalNames::defaultClassVersion(className);
  if(classProps.HasProperty(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionPropertyNameInDictionary())){
    classVersion = classProps.PropertyAsString(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionPropertyNameInDictionary());
  }
  
  std::string mappingVersion = m_persistency->selectMappingVersion( classVersion, containerName );
  if(mappingVersion.empty()){
    throw pool::RelationalStorageService::RelationalStorageServiceException( "No mapping version available for class \""+ dictionary.Name()+
                                                                             "\" in container \"" + containerName + "\"" ,
                                                                               "MappingManager::existingMappingForPolymorphicContainer" );
  }

  std::map< std::string, pool::RelationalStorageService::MappingInformation* >& mappingsForClass = m_classMappingsForPolymorphicContainers[ dictionary ];

  const pool::ObjectRelationalMapping& mappingStructure = m_persistency->retrieveMapping( mappingVersion );
  pool::ObjectRelationalMapping::const_iterator iMappingElement = mappingStructure.find( className );
  if ( iMappingElement != mappingStructure.end() ) {
    return *( mappingsForClass.insert(
                std::make_pair( mappingVersion,
                                new pool::RelationalStorageService::MappingInformation( iMappingElement->second,
                                                                                        mappingVersion ) ) ).first->second );
  }
  else {
    throw pool::RelationalStorageService::RelationalStorageServiceException( "No entry for class \"" + className +
                                                                             "\" in mapping version \"" + mappingVersion + "\"",
                                                                             "MappingManager::existingMappingForPolymorphicContainer" );
  }
}

std::string
pool::RelationalStorageService::MappingManager::versionOfClass( const TypeH& dictionary ){
  std::string className = dictionary.Name(Reflex::SCOPED);
  Reflex::PropertyList classProps = dictionary.Properties();
  std::string classVersion = pool::ObjectRelationalAccess::ObjectRelationalNames::defaultClassVersion(className);
  if(classProps.HasProperty(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionPropertyNameInDictionary())){
    classVersion = classProps.PropertyAsString(pool::ObjectRelationalAccess::ObjectRelationalNames::classVersionPropertyNameInDictionary());
  }
  return classVersion;
}

