#include "ObjectWriter.h"
#include "ObjectRelationalDataBuffer.h"
#include "IContainerHandler.h"
#include "IContainerIteratorHandler.h"
#include "RelationalStorageServiceException.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/IBulkOperation.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/AttributeSpecification.h"
#include "CoralBase/Blob.h"
#include "StorageSvc/DataCallBack.h"
#include "StorageSvc/DbDataHandlerGuard.h"
#include "POOLCore/Reference.h"
#include "POOLCore/Token.h"
#include "Reflex/Member.h"

#include "RVersion.h"
#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
using namespace ROOT;
#endif

pool::RelationalStorageService::ObjectWriter::ObjectWriter( coral::ISchema& schema,
                                                            const TypeH& dictionary,
                                                            const pool::ObjectRelationalMappingElement& mapping,
                                                            pool::DbDatabase* databaseHandle ):
  DbIOHandler(databaseHandle),
  m_dataBuffer( 0 ),
  m_mapping( mapping ),
  m_refDict( TypeH::ByTypeInfo( typeid(pool::Reference) ) )
{
  m_dataBuffer = new pool::RelationalStorageService::ObjectRelationalDataBuffer(m_context);
  m_dataBuffer->build( schema, mapping, dictionary );

}


pool::RelationalStorageService::ObjectWriter::~ObjectWriter()
{
  delete m_dataBuffer;
}

bool
pool::RelationalStorageService::ObjectWriter::write( int oid,
                                                     const void* data,
                                                     pool::DataCallBack& callback )
{
  return write(oid,*m_dataBuffer,data,callback);
}

bool
pool::RelationalStorageService::ObjectWriter::write( int oid,
                                                     ObjectRelationalDataBuffer& dataBuffer,
                                                     const void* data,
                                                     pool::DataCallBack& callback )
{
  coral::AttributeList& buffer = dataBuffer.data();

  // Write first the row of the top level table
  coral::AttributeList::iterator iAttribute = buffer.begin();

  // Skip the identity columns
  for ( size_t iColumn = 0; iColumn < m_mapping.columnNames().size(); ++iColumn ) ++iAttribute;

  // Copy the data
  std::vector<int>::const_iterator iOffset = dataBuffer.offsets().begin();
  while ( iAttribute != buffer.end() ) {
    int offset = *iOffset;
    if ( offset < 0 ) {
      pool::Token::OID_t oid;
      const void* refAddress = static_cast< const char* >( data ) - ( 1 + offset );
      // Call here the ref handler to get the oid...
      pool::DbDataHandlerGuard handlerGuard( &callback, m_databaseContext );
      callback.notify( pool::DataCallBack::PUT, pool::DataCallBack::REFERENCE, m_refDict, const_cast< void* >( refAddress ), &oid );
      iAttribute->data<int>() =  oid.first;
      ++iAttribute;
      ++iOffset;
      iAttribute->data<int>() =  oid.second;
    }
    else {
      if( iAttribute->specification().type() == typeid(coral::Blob) )
      {
        iAttribute->bind<coral::Blob>( dataBuffer.writeBlob( offset, static_cast< const char* >( data ) + offset ) );
      } else 
      {
        iAttribute->setValueFromAddress( static_cast< const char* >( data ) + offset );
      }
    }
    ++iAttribute;
    ++iOffset;
  }

  // Set the oid. For the moment set only the first identity column. In principle there should be a plugin taking
  // care of the OID -> identity values mapping.
  buffer.begin()->data< int >() =  oid;

  // Write the entry into the table
  dataBuffer.table().dataEditor().insertRow( buffer );

  std::vector<int> identity( 1, oid );

  // Loop over the arrays now...
  for ( pool::RelationalStorageService::ObjectRelationalDataBuffer::iterator iBuffer = dataBuffer.begin();
        iBuffer != dataBuffer.end(); ++iBuffer ) {
    if ( ! this->writeArray( identity, **iBuffer,
                             static_cast< const char* >( data ) + (*iBuffer)->offsetToParent(),
                             callback ) ) return false;
  }
  // Loop over the pool pointers now...
  for ( std::vector<ObjectRelationalDataBuffer*>::const_iterator iBuffer = dataBuffer.poolPointerBuffers().begin();
        iBuffer != dataBuffer.poolPointerBuffers().end(); ++iBuffer ) {
    if ( ! this->writePoolPointer( oid, *(*iBuffer),
                                   static_cast< const char* >( data ) + (*iBuffer)->offsetToParent(),
                                   callback ) ) return false;
  }
  return true;
}

bool
pool::RelationalStorageService::ObjectWriter::writeArray( const std::vector<int>& identity,
                                                          pool::RelationalStorageService::ObjectRelationalDataBuffer& buffer,
                                                          const void* address,
                                                          pool::DataCallBack& callback )
{
  const TypeH& dictionary = buffer.dictionary();

  std::vector< int > id( identity.size() + 1, 0 );
  for ( size_t i = 0; i < identity.size(); ++i ) id[i+1] = identity[i];
  
  // Retrieve the container type
  TypeH keyType;
  if ( buffer.keyItems() > 0 ) keyType = dictionary.TemplateArgumentAt(0);

  // Use the iterator to loop over the elements of the container.
  pool::RelationalStorageService::IContainerHandler& containerHandler = buffer.containerHandler();

  size_t containerSize = containerHandler.size( address );
  if ( containerSize == 0 ) return true;

  pool::RelationalStorageService::IContainerIteratorHandler* iteratorHandler = containerHandler.iterate( address );
  const TypeH& iteratorDereferenceReturnType = iteratorHandler->returnType();

  const std::vector<int>& offsets = buffer.offsets();
  coral::AttributeList& data = buffer.data();
  int keyItems = buffer.keyItems();
  bool useBulkInsert = ( buffer.size() == 0 ) ? true : false;
  coral::IBulkOperation* bulkInserter = 0;
  if ( useBulkInsert ) {
    bulkInserter = buffer.table().dataEditor().bulkInsert(data,containerSize);
    if ( ! bulkInserter ) return false;
  }

  for ( size_t iObject = 0; iObject < containerSize; ++iObject ) {
    id[0] = iObject;
    void* objectReference = iteratorHandler->object();
    void* componentData = objectReference;
    void* keyData = 0;

    // Treat the array element.
    coral::AttributeList::iterator iAttribute = data.begin();

    // Skip temporarily the id columns
    for ( std::vector<int>::const_iterator iId = id.begin(); iId != id.end(); ++iId ) ++iAttribute;

    // Copy the data
    std::vector<int>::const_iterator iOffset = offsets.begin();
    if ( keyType ) { // treat the key object first
      Reflex::Member firstField = iteratorDereferenceReturnType.MemberByName( "first" );
      if ( ! firstField ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not retrieve the \"first\" field for the class \"" +
                                                                                 iteratorDereferenceReturnType.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                                 "ObjectWriter::writeArray" );
      }
      keyData = static_cast< char* >( objectReference ) + firstField.Offset();
      for ( int i = 0; i < keyItems; ++i ) {
        int offset = *iOffset;
        if ( offset < 0 ) {
          pool::Token::OID_t oid;
          const void* refAddress = static_cast< const char* >( keyData ) - ( 1 + offset );
          // Call here the ref handler to get the oid...
          callback.notify( pool::DataCallBack::PUT, pool::DataCallBack::REFERENCE, m_refDict, const_cast< void* >( refAddress ), &oid );
          iAttribute->data<int>() =  oid.first;
          ++iAttribute;
          ++iOffset;
          iAttribute->data<int>() = oid.second;
        }
        else {
          if( iAttribute->specification().type() == typeid(coral::Blob) ){
            iAttribute->bind<coral::Blob>( buffer.writeBlob( *iOffset, static_cast< const char* >( keyData ) + *iOffset ) );
          } else {
            iAttribute->setValueFromAddress( static_cast< const char* >( keyData ) + *iOffset );
          }
        }
        ++iAttribute;
        ++iOffset;
      }
      Reflex::Member secondField = iteratorDereferenceReturnType.MemberByName( "second" );
      if ( ! secondField ) {
        throw pool::RelationalStorageService::RelationalStorageServiceException( "Could not retrieve the \"second\" field for the class \"" +
                                                                                 iteratorDereferenceReturnType.Name(Reflex::SCOPED|Reflex::FINAL) + "\"",
                                                                                 "ObjectWriter::writeArray" );
      }
      componentData = static_cast< char* >( objectReference ) + secondField.Offset();
    }
    while ( iAttribute != data.end() ) {
      int offset = *iOffset;
      if ( offset < 0 ) {
        pool::Token::OID_t oid;
        const void* refAddress = static_cast< const char* >( componentData ) - ( 1 + offset );
        // Call here the ref handler to get the oid...
        callback.notify( pool::DataCallBack::PUT, pool::DataCallBack::REFERENCE, m_refDict, const_cast< void* >( refAddress ), &oid );
        iAttribute->data<int>() = oid.first;
        ++iAttribute;
        ++iOffset;
        iAttribute->data<int>() = oid.second;
      }
      else {
        if( iAttribute->specification().type() == typeid(coral::Blob) ){
          iAttribute->bind<coral::Blob>( buffer.writeBlob( *iOffset, static_cast< const char* >( componentData ) + *iOffset ) );
        } else {
          iAttribute->setValueFromAddress( static_cast< const char* >( componentData ) + *iOffset );
        }
      }
      ++iAttribute;
      ++iOffset;
    }

    // Fill in the identity
    iAttribute = data.begin();
    for ( std::vector<int>::const_iterator iId = id.begin(); iId != id.end(); ++iId ) {
      iAttribute->data< int >() = *iId;
      ++iAttribute;
    }

    // Write the entry into the table
    if ( useBulkInsert ) {
      bulkInserter->processNextIteration();
    }
    else {
      buffer.table().dataEditor().insertRow( data );
    }

    // Loop over the arrays now...
    for ( pool::RelationalStorageService::ObjectRelationalDataBuffer::iterator iBuffer = buffer.begin();
          iBuffer != buffer.end(); ++iBuffer ) {
      const void* theAddress = componentData;
      if ( (*iBuffer)->isKeyData() ) {
        theAddress = keyData;
      }
      if ( ! this->writeArray( id, **iBuffer,
                               static_cast< const char* >( theAddress ) + (*iBuffer)->offsetToParent(),
                               callback ) ) return false;
    }

    // Increment the iterator
    iteratorHandler->increment();
  }

  if ( useBulkInsert ) {
    bulkInserter->flush();
  }

  if(bulkInserter) delete bulkInserter;

  delete iteratorHandler;

  return true;
}

bool
pool::RelationalStorageService::ObjectWriter::writePoolPointer( int oid,
                                                                ObjectRelationalDataBuffer& dataBuffer,
                                                                const void* data,
                                                                pool::DataCallBack& callback )
{
  ROOT::Reflex::Object ptrObj(dataBuffer.dictionary(),const_cast<void*>(data));
  bool isNull = *static_cast<bool*>(ptrObj.Invoke("operator!").Address());
  if(!isNull){
    void* ptrData = ptrObj.Invoke("operator*").Address();
    return write(oid,dataBuffer,ptrData,callback);  
  }
  return true;
}
