#ifndef POOLBASECONTAINERMANAGER_H
#define POOLBASECONTAINERMANAGER_H

#include <string>
#include <vector>

#include "RVersion.h"

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
namespace ROOT {
  namespace Reflex {
    class Type;
  }
}
#else
namespace Reflex {
  class Type;
}
#endif

#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
 typedef ROOT::Reflex::Type TypeH;
#else
 typedef Reflex::Type TypeH;
#endif
 
namespace coral {

  class ISchema;
  class ITable;
  class AttributeListSpecification;
  class AttributeList;
}

namespace pool {
  
  class DataCallBack;
  class DbDatabase;

  namespace RelationalStorageService {

    class PoolBaseContainerManager {
    public:
      /// Constructor
      PoolBaseContainerManager( int containerId,
                                const std::string& containerType,
                                coral::ISchema& schema );
      /// Destructor
      virtual ~PoolBaseContainerManager();

      /// Deletes all objects from the container and deletes the container
      virtual bool dropContainer() = 0;

      /// Returns the container id
      int id() const;

      /// Returns the type of the container
      std::string containerType() const;

      virtual std::string mappingVersion() const = 0;

      virtual std::string className() const = 0;

      /// Returns the size of the container (number of objects present in the container)
      unsigned long size();

      /// Returns the number of objects ever written in the container
      unsigned long numberOfObjectsWritten();

      /// Writes an object
      virtual bool writeObject( const void* data,
                                int oid,
                                pool::DataCallBack& callback,
                                const std::string& type_name,
                                const TypeH& dictionary,
                                pool::DbDatabase* databaseHandle ) = 0;

      /// Updates an existing object
      virtual bool updateObject( const void* data,
                                 int oid,
                                 pool::DataCallBack& callback,
                                 const std::string& type_name,
                                 const TypeH& dictionary,
                                 pool::DbDatabase* databaseHandle ) = 0;

      /// Reads an object
      virtual bool readObject( void** data,
                               int oid,
                               pool::DataCallBack& callback,
                               const std::string& type_name,
                               const TypeH& dictionary,
                               pool::DbDatabase* databaseHandle ) = 0;

      /// Reads an object
      virtual bool readAnyObject( void** data,
                                  int oid,
                                  pool::DataCallBack& callback,
                                  const std::string& type_name,
                                  const TypeH& dictionary,
                                  pool::DbDatabase* databaseHandle ) = 0;

      /// Deletes an object
      virtual bool deleteObject( int oid ) = 0;

      /// Force the update of the container contents
      virtual bool updateContent(const std::vector<std::pair<std::string,std::string> >& classInfo, 
                                                                                    bool recreate, bool dry) = 0;

      /// Checks the existence of an OID
      virtual bool existsOid( int oid,
                              int& oidOut,
                              TypeH& dictionary ) = 0;

      /// Ends a transaction
      void endTransaction();

    protected:
      /// Retrieves the class name of the container
      std::string getClassName();

      /// Increments the last OID
      void incrementNumberOfObjectsWritten();

      /// Increments the number of deleted objects
      void incrementNumberOfObjectsDeleted();

      /// Updates the last OID
      bool updateNumberOfObjects(unsigned long writtenObjects,
                                 unsigned long deletedObjects);

      /// Removes the container entry from the headers table
      bool removeContainerEntry();

      /// Returns the schema in use
      coral::ISchema& schema();

    private:
      /// Retrieves the last OID with setting for update optionally
      unsigned long retrieveNumberOfObjectsWritten( bool setForUpdate );

    private:
      /// The container type
      std::string m_containerType;
      
      /// The schema in use
      coral::ISchema& m_schema;

      /// The attribute list for binding the container id variable
      coral::AttributeList* m_dataForId;

      /// The where clause which locates the container record in the container header table
      std::string m_whereClauseForId;

      /// The set clause to increment the last OID
      std::string m_setClauseForIncrementingOid;

      /// Flag indicating whether the container row has been locked.
      bool m_lockedContainerRow;

      /// Flag to indicate whether the number of objects written has been read from the database
      bool m_readNumberOfObjectsWritten;

      /// Cached number of objects written
      unsigned long m_numberOfObjectsWritten;

      /// Flag to indicate whether the container size has been read from the database
      bool m_readContainerSize;

      /// Cached number of objects written
      unsigned long m_containerSize;

      /// Flag to indicate whether the number of objects deleted has been read from the database
      bool m_readNumberOfObjectsDeleted;

      /// The set clause to increment the number of deleted objects
      std::string m_setClauseForIncrementingDeletedObjects;

    };
  }
}


inline coral::ISchema&
pool::RelationalStorageService::PoolBaseContainerManager::schema()
{
  return m_schema;
}

#endif
