#include "PoolHomogeneousContainerManager.h"
#include "ObjectWriter.h"
#include "ObjectReader.h"
#include "MappingInformation.h"
#include "ExistingDataScanner.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/MessageStream.h"
#include "RelationalStorageServiceNames.h"

pool::RelationalStorageService::PoolHomogeneousContainerManager::PoolHomogeneousContainerManager( int containerId,
                                                                                                  coral::ISchema& schema,
                                                                                                  const pool::RelationalStorageService::MappingInformation& mapping ):
  pool::RelationalStorageService::PoolBaseContainerManager( containerId, pool::RelationalStorageService::RelationalStorageServiceNames::homogeneousContainerTypeValueInContainerHeaderTable(), schema ),
  m_mapping( mapping.element ),
  m_mappingVersion( mapping.version ),
  m_dictionary( Reflex::Type::ByName( m_mapping.variableType() ) ),
  m_writer( 0 ),
  m_reader( 0 ),
  m_oidMin( 0 ),
  m_oidMax( 0 ),
  m_validOids()
{}



pool::RelationalStorageService::PoolHomogeneousContainerManager::~PoolHomogeneousContainerManager()
{
  m_validOids.clear();
  if ( m_writer ) delete m_writer;
  if ( m_reader ) delete m_reader;
}

std::string pool::RelationalStorageService::PoolHomogeneousContainerManager::mappingVersion() const {
  return m_mappingVersion;
}

std::string pool::RelationalStorageService::PoolHomogeneousContainerManager::className() const {
  return m_mapping.variableType();
}


bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::dropContainer()
{
  if ( this->schema().existsTable( m_mapping.tableName() ) ) {
    int oid = 0;
    int oidOut = 0;
    Reflex::Type dict;
    while ( this->existsOid( oid, oidOut, dict ) ) {
      this->deleteAnObject( m_mapping, "", oidOut );
      m_validOids.erase( oidOut );
      oid = oidOut;
    }
  }
  return this->removeContainerEntry();
}


bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::writeObject( const void* data,
                                                                              int oid,
                                                                              pool::DataCallBack& callback,
                                                                              const std::string& type_name,
                                                                              const Reflex::Type& dictionary,
                                                                              pool::DbDatabase* databaseHandle )
{
  // Check on the type name
  if ( type_name != m_mapping.variableType() ) return false;

  this->incrementNumberOfObjectsWritten();

  if ( ! m_writer ) {
    m_writer = new pool::RelationalStorageService::ObjectWriter( this->schema(),
                                                                 dictionary,
                                                                 m_mapping,
                                                                 databaseHandle );
  }

  return m_writer->write( oid, data, callback );
}


bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::updateObject( const void* data,
                                                                               int oid,
                                                                               pool::DataCallBack& callback,
                                                                               const std::string& type_name,
                                                                               const Reflex::Type& dictionary,
                                                                               pool::DbDatabase* databaseHandle )
{
  // Check on the type name
  if ( type_name != m_mapping.variableType() ) return false;
  if ( ! m_writer ) {
    m_writer = new pool::RelationalStorageService::ObjectWriter( this->schema(),
                                                                 dictionary,
                                                                 m_mapping,
                                                                 databaseHandle );
  }

  return ( ( this->deleteAnObject( m_mapping, "", oid ) ) && ( m_writer->write( oid, data, callback ) ) );
}


bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::readObject( void** data,
                                                                             int oid,
                                                                             pool::DataCallBack& callback,
                                                                             const std::string& type_name,
                                                                             const Reflex::Type& dictionary,
                                                                             pool::DbDatabase* databaseHandle )
{
  // Check on the type name
  if ( type_name != m_mapping.variableType() ) return false;
  void* objectData = *data;
  if ( ! objectData ) return false;
  if ( ! m_reader ) {
    m_reader = new pool::RelationalStorageService::ObjectReader( this->schema(),
                                                                 dictionary,
                                                                 m_mapping,
                                                                 databaseHandle );
  }
  return m_reader->readObject( objectData, oid, false, callback );
}


bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::readAnyObject( void** data,
                                                                                int oid,
                                                                                pool::DataCallBack& callback,
                                                                                const std::string& type_name,
                                                                                const Reflex::Type& dictionary,
                                                                                pool::DbDatabase* databaseHandle )
{
  // Check on the type name
  if ( type_name != m_mapping.variableType() ) return false;
  void* objectData = *data;
  if ( ! objectData ) return false;
  if ( ! m_reader ) {
    m_reader = new pool::RelationalStorageService::ObjectReader( this->schema(),
                                                                 dictionary,
                                                                 m_mapping,
                                                                 databaseHandle );
  }

  return m_reader->readObject( objectData, oid, true, callback );
}


bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::deleteObject( int oid )
{
  // Increment the number of objects deleted
  this->incrementNumberOfObjectsDeleted();
  if ( this->deleteAnObject( m_mapping, "", oid ) ) {
    m_validOids.erase( oid );
    return true;
  }
  else return false;
}


bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::deleteAnObject( const pool::ObjectRelationalMappingElement& mapping,
                                                                                 const std::string& parentTableName,
                                                                                 int oid )
{
  for ( pool::ObjectRelationalMappingElement::const_iterator iMapping = mapping.begin();
        iMapping != mapping.end(); ++iMapping ) {
    pool::ObjectRelationalMappingElement::ElementType elementType = iMapping->second.elementType();
    if ( elementType == pool::ObjectRelationalMappingElement::Object ||
         elementType == pool::ObjectRelationalMappingElement::PoolReference ||
         elementType == pool::ObjectRelationalMappingElement::Array ) {
      this->deleteAnObject( iMapping->second, mapping.tableName(), oid );
    }
  }

  if ( mapping.tableName() == parentTableName ) return true;
  size_t columnId = mapping.columnNames().size() - 1;
  if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) columnId -= 2;
  std::string oidColumn = mapping.columnNames()[ columnId ];
  coral::AttributeList condition;
  condition.extend< int >( oidColumn );
  condition.begin()->data< int >() = oid;
  this->schema().tableHandle( mapping.tableName() ).dataEditor().deleteRows( oidColumn + " = :" + oidColumn,
                                                                             condition );
  return true;
}

bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::updateContent(const std::vector<std::pair<std::string,std::string> >&, bool, bool dry)
{
  ExistingDataScanner scanner(this->schema());
  scanner.prepareScan(m_mapping);
  scanner.scanAll();
  coral::MessageStream log( pool::RelationalStorageService::serviceName() );
  log << coral::Info << "       Header: objects stored="<<size()<<
                            " ( written="<<numberOfObjectsWritten()<<
                            " deleted="<<numberOfObjectsWritten()-size()<<
                            " )" << coral::MessageStream::endmsg; 
  log << coral::Info << "       Scan on the table "<< m_mapping.tableName() <<
                            ": object stored="<< scanner.writtenObjects()-scanner.deletedObjects() <<
                            " ( written="<<scanner.writtenObjects()<<
                            " deleted="<<scanner.deletedObjects()<<
                            " )" << coral::MessageStream::endmsg;
  std::string status("UPTODATE");
  if(scanner.writtenObjects()!=numberOfObjectsWritten() ||
     scanner.deletedObjects()!=(numberOfObjectsWritten()-size())) status = "OUTOFDATE";
  log << coral::Info << "       Status=[" <<status<<"]"<<coral::MessageStream::endmsg; 
  bool ret = true;
  if(!dry) {
    log << coral::Info << "       Updating container. " << coral::MessageStream::endmsg;
    ret = this->updateNumberOfObjects(scanner.writtenObjects(),scanner.deletedObjects());
  }
  return ret;
}

bool
pool::RelationalStorageService::PoolHomogeneousContainerManager::existsOid( int oid,
                                                                            int& oidOut,
                                                                            Reflex::Type& dictionary )
{
  if ( ! ( oid >= m_oidMin && oid < m_oidMax ) ) {
    m_validOids.clear();

    int intervalSize = 10000; // Need to better define this !!!
    m_oidMin = oid;

    coral::IQuery* query = this->schema().tableHandle( m_mapping.tableName() ).newQuery();
    std::string idColumn = *( m_mapping.columnNames().begin() );
    query->addToOutputList( idColumn );
    query->defineOutputType( idColumn, coral::AttributeSpecification::typeNameForType< int >() );
    query->addToOrderList( idColumn );

    coral::AttributeList oidData;
    oidData.extend< int >( idColumn );
    coral::AttributeList::iterator iData = oidData.begin();
    iData->data< int >() = m_oidMin;
    std::string whereClause = idColumn + ">= :" + idColumn;
    query->setCondition( whereClause, oidData );

    query->setRowCacheSize( intervalSize );
    query->limitReturnedRows( intervalSize, 0 );

    int value = 0;
    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      value = cursor.currentRow().begin()->data< int >();
      m_validOids.insert( value );
      if ( (int)m_validOids.size() > intervalSize ) {
        break;
      }
    }
    delete query;
    m_oidMax = value + 1;
  }

  std::set< int >::const_iterator iResult = m_validOids.lower_bound( oid );
  if ( iResult == m_validOids.end() ) return false;
  oidOut = *iResult;
  
  dictionary = m_dictionary;
  return true;
}
