#include "PoolPolymorphicContainerManager.h"
#include "MappingManager.h"
#include "MappingSequence.h"
#include "MappingInformation.h"
#include "ObjectWriter.h"
#include "ObjectReader.h"
#include "ExistingDataScanner.h"
#include "RelationalStorageServiceNames.h"
#include "ObjectRelationalAccess/ObjectRelationalMappingElement.h"
#include "RelationalAccess/ISchema.h"
#include "RelationalAccess/ITable.h"
#include "RelationalAccess/ITableDescription.h"
#include "RelationalAccess/ITableDataEditor.h"
#include "RelationalAccess/IBulkOperation.h"
#include "RelationalAccess/IQuery.h"
#include "RelationalAccess/ICursor.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/MessageStream.h"
#include "RelationalStorageServiceNames.h"

pool::RelationalStorageService::PoolPolymorphicContainerManager::PoolPolymorphicContainerManager( int containerId,
                                                                                                  const std::string& containerName,
                                                                                                  coral::ISchema& schema,
                                                                                                  const std::string& tableName,
                                                                                                  pool::RelationalStorageService::MappingManager& mappingManager,
                                                                                                  pool::RelationalStorageService::MappingSequence& mappingSequences ):
  pool::RelationalStorageService::PoolBaseContainerManager( containerId, pool::RelationalStorageService::RelationalStorageServiceNames::polymorphicContainerTypeValueInContainerHeaderTable(), schema ),
  m_containerName(containerName),
  m_schema( schema ),
  m_containerTableName( tableName ),
  m_mappingManager( mappingManager ),
  m_mappingSequences( mappingSequences ),
  m_containerRow( 0 ),
  m_writers(),
  m_readers(),
  m_cachedEntries(),
  m_oidMin( 0 ),
  m_oidMax( 0 )
{
  m_containerRow = new coral::AttributeList;
  m_containerRow->extend<int>( pool::RelationalStorageService::RelationalStorageServiceNames::oidVariableNameInPolymorphicContainerTable() );
  m_containerRow->extend<std::string>( pool::RelationalStorageService::RelationalStorageServiceNames::classNameVariableNameInPolymorphicContainerTable() );
  m_containerRow->extend<std::string>( pool::RelationalStorageService::RelationalStorageServiceNames::mappingVersionVariableNameInPolymorphicContainerTable() );
  m_containerRow->extend<int>( pool::RelationalStorageService::RelationalStorageServiceNames::targetOIDVariableNameInPolymorphicContainerTable() );
}



pool::RelationalStorageService::PoolPolymorphicContainerManager::~PoolPolymorphicContainerManager()
{
  for ( std::map< TypeH, pool::RelationalStorageService::ObjectReader* >::iterator iReader = m_readers.begin();
        iReader != m_readers.end(); ++iReader ) delete iReader->second;
  m_readers.clear();

  for ( std::map< TypeH, pool::RelationalStorageService::ObjectWriter* >::iterator iWriter = m_writers.begin();
        iWriter != m_writers.end(); ++iWriter ) delete iWriter->second;
  m_writers.clear();

  delete m_containerRow;
}

std::string
pool::RelationalStorageService::PoolPolymorphicContainerManager::mappingVersion() const{
  return std::string("");
}

std::string
pool::RelationalStorageService::PoolPolymorphicContainerManager::className() const{
  return std::string("");
}

bool
pool::RelationalStorageService::PoolPolymorphicContainerManager::dropContainer()
{
  // Remove the objects first
  int oid = 0;
  int oidOut = 0;
  TypeH dict;
  while ( this->existsOid( oid, oidOut, dict ) ) {
    const pool::RelationalStorageService::PoolPolymorphicContainerManager::Entry& containerEntry = m_cachedEntries.find( oidOut )->second;
    if ( ! dict ) return false;
    std::string nominalMappingVersion = containerEntry.mappingVersion;
    const pool::RelationalStorageService::MappingInformation& mappingInfo =
      m_mappingManager.existingMappingForPolymorphicContainer( m_containerName,dict,nominalMappingVersion );
    this->deleteAnObject( mappingInfo.element, "", containerEntry.targetOid );
    m_cachedEntries.erase( oidOut );
    oid = oidOut;
  }

  return this->removeContainerEntry();
}



bool
pool::RelationalStorageService::PoolPolymorphicContainerManager::writeObject( const void* data,
                                                                              int oid,
                                                                              pool::DataCallBack& callback,
                                                                              const std::string& /*type_name*/,
                                                                              const TypeH& dictionary,
                                                                              pool::DbDatabase* databaseHandle )
{
  // Increment the last oid
  this->incrementNumberOfObjectsWritten();

  // Get a mapping for the class.
  const pool::RelationalStorageService::MappingInformation& mappingInfo = m_mappingManager.mappingForPolymorphicContainer( m_containerName,dictionary );

  // Retrieve a new valid target oid.
  int targetOid = m_mappingSequences.incrementOid( mappingInfo.element.tableName() );

  // Write the entry in the container table.
  coral::AttributeList::iterator iAttribute = m_containerRow->begin();
  iAttribute->data<int>( ) = oid;
  ++iAttribute;
  iAttribute->data<std::string>() = mappingInfo.element.variableType();
  ++iAttribute;
  iAttribute->data<std::string>() = mappingInfo.version;
  ++iAttribute;
  iAttribute->data<int>() = targetOid;

  m_schema.tableHandle( m_containerTableName ).dataEditor().insertRow( *m_containerRow );

  // Write the object.
  pool::RelationalStorageService::ObjectWriter* writer = 0;
  std::map< TypeH, pool::RelationalStorageService::ObjectWriter* >::iterator iWriter = m_writers.find( dictionary );
  if ( iWriter == m_writers.end() ) {
    writer = new pool::RelationalStorageService::ObjectWriter( this->schema(),
                                                               dictionary,
                                                               mappingInfo.element,
                                                               databaseHandle );
    m_writers.insert( std::make_pair( dictionary, writer ) );
  }
  else {
    writer = iWriter->second;
  }

  return writer->write( targetOid, data, callback );
}



bool
pool::RelationalStorageService::PoolPolymorphicContainerManager::updateObject( const void* data,
                                                                               int oid,
                                                                               pool::DataCallBack& callback,
                                                                               const std::string& /*type_name*/,
                                                                               const TypeH& dictionary,
                                                                               pool::DbDatabase* databaseHandle )
{
  int oidOut = 0;
  TypeH dict;
  if (! ( this->existsOid( oid, oidOut, dict ) && oid == oidOut && dict == dictionary ) ) return false;
  const pool::RelationalStorageService::PoolPolymorphicContainerManager::Entry& containerEntry = m_cachedEntries.find( oid )->second;

  // Get a mapping for the class.
  std::string nominalMappingVersion = containerEntry.mappingVersion;
  const pool::RelationalStorageService::MappingInformation& mappingInfo = m_mappingManager.existingMappingForPolymorphicContainer( m_containerName,dict,nominalMappingVersion );
  // Retrieve the writer
  pool::RelationalStorageService::ObjectWriter* writer = 0;
  std::map< TypeH, pool::RelationalStorageService::ObjectWriter* >::iterator iWriter = m_writers.find( dictionary );
  if ( iWriter == m_writers.end() ) {
    writer = new pool::RelationalStorageService::ObjectWriter( this->schema(),
                                                               dictionary,
                                                               mappingInfo.element,
                                                               databaseHandle );
    m_writers.insert( std::make_pair( dictionary, writer ) );
  }
  else {
    writer = iWriter->second;
  }

  return ( ( this->deleteAnObject( mappingInfo.element, "", containerEntry.targetOid ) ) && ( writer->write( containerEntry.targetOid, data, callback ) ) );
}



bool
pool::RelationalStorageService::PoolPolymorphicContainerManager::readObject( void** data,
                                                                             int oid,
                                                                             pool::DataCallBack& callback,
                                                                             const std::string& /*type_name*/,
                                                                             const TypeH& dictionary,
                                                                             pool::DbDatabase* databaseHandle )
{
  void* objectData = *data;
  if ( ! objectData ) return false;

  int oidOut = 0;
  TypeH dict;

  if (! ( this->existsOid( oid, oidOut, dict ) && oid == oidOut && dict == dictionary ) ) return false;
  const pool::RelationalStorageService::PoolPolymorphicContainerManager::Entry& containerEntry = m_cachedEntries.find( oid )->second;

  // Get a mapping for the class.
  std::string nominalMappingVersion = containerEntry.mappingVersion;
  const pool::RelationalStorageService::MappingInformation& mappingInfo = 
       m_mappingManager.existingMappingForPolymorphicContainer( m_containerName,dictionary,nominalMappingVersion );
  // Read the object.
  pool::RelationalStorageService::ObjectReader* reader = 0;
  std::map< TypeH, pool::RelationalStorageService::ObjectReader* >::iterator iReader = m_readers.find( dictionary );
  if ( iReader == m_readers.end() ) {
    reader = new pool::RelationalStorageService::ObjectReader( this->schema(),
                                                               dictionary,
                                                               mappingInfo.element,
                                                               databaseHandle );
    m_readers.insert( std::make_pair( dictionary, reader ) );
  }
  else {
    reader = iReader->second;
  }

  return reader->readObject( objectData, containerEntry.targetOid, false, callback );
}


bool
pool::RelationalStorageService::PoolPolymorphicContainerManager::readAnyObject( void** data,
                                                                                int oid,
                                                                                pool::DataCallBack& callback,
                                                                                const std::string& /*type_name*/,
                                                                                const TypeH& dictionary,
                                                                                pool::DbDatabase* databaseHandle )
{
  void* objectData = *data;
  if ( ! objectData ){
    return false;
  }
  int oidOut = 0;
  bool ok = false;
  std::string mappingVersion = "";
  TypeH dict;
  while ( this->existsOid( oid, oidOut, dict ) ) {
    oid = oidOut;
    const pool::RelationalStorageService::PoolPolymorphicContainerManager::Entry& containerEntry = m_cachedEntries.find( oid )->second;
    // Check whether the type is the correct one
    if ( dict == dictionary ) {
      oidOut = containerEntry.targetOid;
      mappingVersion = containerEntry.mappingVersion;
      ok = true;
      break;
    }
    ++oid;
  }
  if ( ! ok ) return false;

  // Get a mapping for the class.
  const pool::RelationalStorageService::MappingInformation& mappingInfo = m_mappingManager.existingMappingForPolymorphicContainer( m_containerName,dictionary,mappingVersion );

  // Read the object.
  pool::RelationalStorageService::ObjectReader* reader = 0;
  std::map< TypeH, pool::RelationalStorageService::ObjectReader* >::iterator iReader = m_readers.find( dictionary );
  if ( iReader == m_readers.end() ) {
    reader = new pool::RelationalStorageService::ObjectReader( this->schema(),
                                                               dictionary,
                                                               mappingInfo.element,
                                                               databaseHandle );
    m_readers.insert( std::make_pair( dictionary, reader ) );
  }
  else {
    reader = iReader->second;
  }

  return reader->readObject( objectData, oidOut, false, callback );
}


bool
pool::RelationalStorageService::PoolPolymorphicContainerManager::deleteObject( int oid )
{
  int oidOut = 0;
  TypeH dict;
  if (! ( this->existsOid( oid, oidOut, dict ) && oid == oidOut ) ) return false;
  const pool::RelationalStorageService::PoolPolymorphicContainerManager::Entry& containerEntry = m_cachedEntries.find( oid )->second;

  // Get a mapping for the class.
  std::string nominalMappingVersion = containerEntry.mappingVersion;
  const pool::RelationalStorageService::MappingInformation& mappingInfo = m_mappingManager.existingMappingForPolymorphicContainer( m_containerName,dict,nominalMappingVersion );
  // Increment the number of objects deleted
  this->incrementNumberOfObjectsDeleted();

  // Delete the row from the container table
  std::string whereClause = pool::RelationalStorageService::RelationalStorageServiceNames::oidVariableNameInPolymorphicContainerTable() + " = :" +
    pool::RelationalStorageService::RelationalStorageServiceNames::oidVariableNameInPolymorphicContainerTable();
  coral::AttributeList whereData;
  whereData.extend<int>( pool::RelationalStorageService::RelationalStorageServiceNames::oidVariableNameInPolymorphicContainerTable() );
  whereData.begin()->data<int>() = oid;
  m_schema.tableHandle( m_containerTableName ).dataEditor().deleteRows( whereClause, whereData );

  // Delete the object data
  if ( this->deleteAnObject( mappingInfo.element, "", containerEntry.targetOid ) ) {
    m_cachedEntries.erase( oid );
    return true;
  }
  else return false;
}


bool
pool::RelationalStorageService::PoolPolymorphicContainerManager::deleteAnObject( const pool::ObjectRelationalMappingElement& mapping,
                                                                                 const std::string& parentTableName,
                                                                                 int oid )
{
  for ( pool::ObjectRelationalMappingElement::const_iterator iMapping = mapping.begin();
        iMapping != mapping.end(); ++iMapping ) {
    pool::ObjectRelationalMappingElement::ElementType elementType = iMapping->second.elementType();
    if ( elementType == pool::ObjectRelationalMappingElement::Object ||
         elementType == pool::ObjectRelationalMappingElement::PoolReference ||
         elementType == pool::ObjectRelationalMappingElement::Array ) {
      this->deleteAnObject( iMapping->second, mapping.tableName(), oid );
    }
  }

  if ( mapping.tableName() == parentTableName ) return true;
  size_t columnId = mapping.columnNames().size() - 1;
  if ( mapping.elementType() == pool::ObjectRelationalMappingElement::PoolReference ) columnId -= 2;
  std::string oidColumn = mapping.columnNames()[ columnId ];
  coral::AttributeList condition;
  condition.extend< int >( oidColumn );
  condition.begin()->data< int >() = oid;
  this->schema().tableHandle( mapping.tableName() ).dataEditor().deleteRows( oidColumn + " = :" + oidColumn,
                                                                             condition );
  return true;
}

bool 
pool::RelationalStorageService::PoolPolymorphicContainerManager::updateEntries( const ExistingDataScanner* dataScanner, int& newOid ){
  // Write the entry chunk in the container table.
  coral::IBulkOperation* rowInserter = m_schema.tableHandle( m_containerTableName ).dataEditor().bulkInsert(*m_containerRow, 10000);
  for(std::vector<ExistingDataScanner::Entry>::const_iterator iEntry = dataScanner->entries().begin();iEntry!=dataScanner->entries().end();iEntry++) {

    coral::AttributeList::iterator iAttribute = m_containerRow->begin();
    iAttribute->data<int>() = newOid++;
    ++iAttribute;
    iAttribute->data<std::string>() = iEntry->className;
    ++iAttribute;
    iAttribute->data<std::string>() = iEntry->mappingVersion;
    ++iAttribute;
    iAttribute->data<int>() = iEntry->targetOid;
    rowInserter->processNextIteration();
  }
  rowInserter->flush();
  delete rowInserter;
  return true;
}

bool 
pool::RelationalStorageService::PoolPolymorphicContainerManager::updateContent(const std::vector<std::pair<std::string,std::string> >& classInfo, bool recreate, bool dry)
{

  coral::MessageStream log( pool::RelationalStorageService::serviceName() );
  bool upToDate = true;
  int newOid = 0;
  if(recreate && !dry){
    // remove all the current rows
    m_schema.tableHandle( m_containerTableName ).dataEditor().deleteRows("",coral::AttributeList());
  } else {
    newOid = numberOfObjectsWritten();
  }

  ExistingDataScanner* scanner = new ExistingDataScanner(this->schema());

  // loop over the classInfo pairs
  for(std::vector<std::pair<std::string,std::string> >::const_iterator iClInfo=classInfo.begin();iClInfo!=classInfo.end();iClInfo++){
    TypeH dictionary = TypeH::ByName(iClInfo->first);
    const pool::RelationalStorageService::MappingInformation& mappingInfo = m_mappingManager.mappingForPolymorphicContainer( m_containerName,dictionary );
    const ObjectRelationalMappingElement& mappingElement = mappingInfo.element;
    if(recreate){
      scanner->prepareScan(mappingElement,mappingInfo.version);
    } else {
      scanner->prepareScanForNewRows(m_schema.tableHandle( m_containerTableName ).description().name(),mappingElement,mappingInfo.version);
    }
    bool cont = true;
    while(cont){
      if(scanner->scan()){
        if(scanner->isChunkFull()){
          // in this case the chunk is ready to be used for the container table update
          if(!dry){
            if(!updateEntries(scanner,newOid)){
              delete scanner;
              return false;
            }
          }
          // empty the current chunk in the scanner
          scanner->resetChunk();
        }
      } else {
        cont = false;
      }
    }
  
    if(scanner->chunkSize()!=0 && !dry){
      // full chunk or not, what is left in the scanner has to be added in the container table. 
      if(!updateEntries(scanner,newOid)){
        delete scanner;
        return false;
      }    
      scanner->resetChunk();
    }

    unsigned long newObjects = scanner->storedObjects(mappingElement.tableName());
    log << coral::Info << "       Scan on the table "<< mappingElement.tableName() <<
                          ": new objects stored="<< newObjects <<
                          " )" << coral::MessageStream::endmsg;
    if(newObjects) upToDate = false;
        
  }

  unsigned long sz = 0;
  if(!recreate) sz = size();
  if(dry) {
    newOid += scanner->storedObjects();
  }
  unsigned long currentWritten = newOid; 
  unsigned long currentDeleted = newOid-(sz+scanner->storedObjects());

  log << coral::Info << "       Header: objects stored="<<size()<<
                            " ( written="<<numberOfObjectsWritten()<<
                            " deleted="<<numberOfObjectsWritten()-size()<<
                            " )" << coral::MessageStream::endmsg; 
  log << coral::Info << "       Scan on the tables "<< 
                            ": object stored="<< currentWritten-currentDeleted <<
                            " ( written="<<currentWritten<<
                            " deleted="<<currentDeleted<<
                            " )" << coral::MessageStream::endmsg;

  std::string status("UPTODATE");
  if(!upToDate) status = "OUTOFDATE";
  log << coral::Info << "       Status=[" <<status<<"]"<<coral::MessageStream::endmsg; 
  if(!dry){
    log << coral::Info << "       Updating container. " << coral::MessageStream::endmsg;
    if(!this->updateNumberOfObjects(currentWritten,currentDeleted)) {
      delete scanner;
      return false;
    }      
    // update the mapping sequences
    for(std::map<std::string,std::pair<unsigned long,int> >::const_iterator iTable=scanner->tableMap().begin();iTable!=scanner->tableMap().end();iTable++){
      m_mappingSequences.updateOid(iTable->first,iTable->second.second);
    }
  }


  delete scanner;
  return true;
}

bool
pool::RelationalStorageService::PoolPolymorphicContainerManager::existsOid( int oid,
                                                                            int& oidOut,
                                                                            TypeH& dictionary )
{
  if ( ! ( oid >= m_oidMin && oid < m_oidMax ) ) {
    m_cachedEntries.clear();

    unsigned int intervalSize = 10000; // Need to better define this !!!
    m_oidMin = oid;

    coral::IQuery* query = m_schema.tableHandle( m_containerTableName ).newQuery();
    query->addToOutputList( pool::RelationalStorageService::RelationalStorageServiceNames::oidVariableNameInPolymorphicContainerTable() );
    query->defineOutputType( pool::RelationalStorageService::RelationalStorageServiceNames::oidVariableNameInPolymorphicContainerTable(), coral::AttributeSpecification::typeNameForType<int>() );
    
    query->addToOutputList( pool::RelationalStorageService::RelationalStorageServiceNames::classNameVariableNameInPolymorphicContainerTable() );
    query->addToOutputList( pool::RelationalStorageService::RelationalStorageServiceNames::mappingVersionVariableNameInPolymorphicContainerTable() );
    query->addToOutputList( pool::RelationalStorageService::RelationalStorageServiceNames::targetOIDVariableNameInPolymorphicContainerTable() );
    query->defineOutputType( pool::RelationalStorageService::RelationalStorageServiceNames::targetOIDVariableNameInPolymorphicContainerTable(), coral::AttributeSpecification::typeNameForType<int>() );

    query->addToOrderList( pool::RelationalStorageService::RelationalStorageServiceNames::oidVariableNameInPolymorphicContainerTable() );

    coral::AttributeList oidData;
    oidData.extend<int>( pool::RelationalStorageService::RelationalStorageServiceNames::oidVariableNameInPolymorphicContainerTable() );
    oidData.begin()->data< int >() = oid;
    std::string whereClause = pool::RelationalStorageService::RelationalStorageServiceNames::oidVariableNameInPolymorphicContainerTable() + " >= :" +
      pool::RelationalStorageService::RelationalStorageServiceNames::oidVariableNameInPolymorphicContainerTable();
    query->setCondition( whereClause, oidData );
    query->setRowCacheSize( intervalSize );
    query->limitReturnedRows( intervalSize, 0 );

    coral::ICursor& cursor = query->execute();
    while ( cursor.next() ) {
      coral::AttributeList::const_iterator iAttribute = cursor.currentRow().begin();
      m_oidMax = iAttribute->data< int >();
      pool::RelationalStorageService::PoolPolymorphicContainerManager::Entry entry;
      ++iAttribute;
      std::string className = iAttribute->data< std::string >();
      entry.dictionary = TypeH::ByName( className );
      ++iAttribute;
      entry.mappingVersion  = iAttribute->data< std::string >();
      ++iAttribute;
      entry.targetOid = iAttribute->data< int >();
      m_cachedEntries.insert( std::make_pair( m_oidMax,entry ) );
      if ( m_cachedEntries.size() > intervalSize ) {
        break;
      }
    }
    ++m_oidMax;
    delete query;
  }


  std::map< int, pool::RelationalStorageService::PoolPolymorphicContainerManager::Entry >::const_iterator iResult = m_cachedEntries.lower_bound( oid );
  if ( iResult == m_cachedEntries.end() ) return false;
  oidOut = iResult->first;
  dictionary = iResult->second.dictionary;
  return true;
}

