#ifndef RELATIONALSTORAGESERVICE_RELATIONALPOOLDATABASE_H
#define RELATIONALSTORAGESERVICE_RELATIONALPOOLDATABASE_H

#include "StorageSvc/DbDatabaseImp.h"

namespace coral {
  class ISessionProxy;
}

namespace pool {
  
  class DbTransaction;
  class Context;

  namespace RelationalStorageService {

    class PoolDatabaseManager;

    /** @class RelationalPoolDatabase RelationalPoolDatabase.h
     *
     * Extension of the default implementation DbDatabaseImp for the object storage
     * using the POOL Relational backend
     */

    class RelationalPoolDatabase : public pool::DbDatabaseImp
    {
    public:
      /// Standard Constuctor
      /** @param idb      [IN]  Pointer to implementation object
       *
       */
      RelationalPoolDatabase( pool::IOODatabase* idb,
                              pool::Context* context );
      
      /// Standard destructor
      ~RelationalPoolDatabase();

      // Methods specific to this class

      /// Returns the global application context
      pool::Context* context();
      
      /// Returns the database manager
      PoolDatabaseManager& databaseManager();

      /// Static call, for the existence of a database.
      static bool exists( const std::string& name,
                          pool::Context* context );

      /// Returns the session object
      coral::ISessionProxy& session();

      // Methods from the base class
      
      /// Callback after a successfull opening of a pool database
      pool::DbStatus onOpen( const pool::DbDatabase& dbH,
                             pool::DbAccessMode mode );

      /// Open Database object
      /** @param domH    [IN]  Handle to valid domain object
       *                       (validity ensured by upper levels).
       * @param name     [IN]  Name of the database to be opened.
       * @param mode     [IN]  Desired session access mode.
       *
       * @return DbStatus code indicating success or failure.  
       */
      pool::DbStatus open( const pool::DbDomain& domH, 
                           const std::string& name, 
                           pool::DbAccessMode mode );

      /// Re-open database with changing access permissions
      pool::DbStatus reopen( pool::DbAccessMode mode );

      /// Close database access
      /** @param mode     [IN]  Desired session access mode.
       *
       * @return DbStatus code indicating success or failure.  
       */
      pool::DbStatus close( pool::DbAccessMode mode );

      /// Start/Commit/Rollback Database Transaction
      /** @param refTransaction [IN/OUT]  Transaction context.
       *
       * @return DbStatus code indicating success or failure.  
       */
      pool::DbStatus transAct( pool::DbTransaction& refTransaction);

      /// Remove from the database the homogeneous containers (if any) which are using the specific mapping version,
      /// and from the polimorphic containers the entries which are using the specific mapping version
      /** @param mappingVersion [IN] The mapping version to remove
       *
       */

      static void dropContainer( pool::Context* context, 
                                 const std::string& databaseName,
                                 const std::string& containerName,
                                 bool removeDataTables );
      

    private:
      /// The format version number
      std::string m_formatVersion;

      /// The application context
      pool::Context*  m_context;

      /// Flag to set the context ownership
      bool m_ownedContext;

      /// The session object in use
      coral::ISessionProxy* m_session;

      /// Flag indicating whether the database is open in read-only mode
      bool m_readOnlyMode;

      /// The database manager
      PoolDatabaseManager* m_databaseManager;

      /// The PFN name
      std::string m_pfnName;
    };

  }

}

#endif
