//====================================================================
//
//	Package    : RootStorageSvc (The POOL project)
//
//	Author     : M.Frank
//====================================================================
#ifndef POOL_CUSTOMOBJSTREAMER_H
#define POOL_CUSTOMOBJSTREAMER_H

// Framework includes
#include "RootStorageSvc/ICINTStreamer.h"

/* 
 * POOL namespace
 */
namespace pool {

  /** @class CustomObjStreamer CustomObjStreamer.h RootStorageSvc/CustomObjStreamer.h
    * 
    * Example of a custom ROOT streamer to handle a specific I/O request
    * for a given object, which is described by the LCG dictionary.
    * ROOT does not provide any object information in this case; the
    * void pointer in the read/write methods have to handled by the 
    * client.
    * 
    * You are fully responsible to handle schema evolution. ROOT cannot
    * help, if custom streamers are applied. Also the version number 
    * is NOT written by default.
    *
    * @author  M.Frank
    * @version 1.0
    */
  class CustomObjStreamer : virtual public ICINTStreamer  {
  protected:
    /// Reference to reflection class information
    TypeH m_class;

  public:
    /// Standard copnstructor
    /** @param cl   [IN]   Pointer to reflection class.
      * @return  Reference to streamer object.
      */
    CustomObjStreamer(const TypeH& cl) : m_class(cl) {      }

    /// Standard destructor
    virtual ~CustomObjStreamer()       {                                   }

    /// Catalog label
    /** Note: here a naming convention is applied:
      * The catalogLabel **MUST** be CINTStreamer<object_type>,
      * Where object_type is the full qualified name of the object
      * class including namespaces as it occurs in the LCG reflection
      * package.
      */
    static const char* catalogLabel()  { return "CINTStreamer<CustomObj>"; }

    /// Release call: destroy all internally allocated resources.
    virtual void release()             { delete this;                      }

    /// Callback for reading the object
    /** @param obj  [IN]   Pointer to user object.
      * @param cl   [IN]   Pointer to ROOT TClass structure
      * @param b    [IN]   Reference to ROOT Buffer
      * @param arg  [IN]   Currently unused ROOT argument.
      */
    virtual void readData(void*    /* obj  */, 
                          TClass*  /* cl   */,
                          TBuffer& /* b    */, 
                          int      /* arg  */)  {
    }
    /// Callback for writing the object
    /** @param obj  [IN]   Pointer to user object.
      * @param cl   [IN]   Pointer to ROOT TClass structure
      * @param b    [IN]   Reference to ROOT Buffer
      * @param arg  [IN]   Currently unused ROOT argument.
      */
    virtual void writeData(void*    /* obj  */, 
                           TClass*  /* cl   */,
                           TBuffer& /* b    */, 
                           int      /* arg  */) {
    }
  };
}
#endif // POOL_CUSTOMOBJSTREAMER_H
