//====================================================================
//	Root Database file implementation
//--------------------------------------------------------------------
//
//	Package    : RootStorageSvc (The POOL project)
//
//====================================================================
#include "RootDBase.h"
#include "POOLCore/strcasecmp.h"
#include "StorageSvc/DbInstanceCount.h"
#include "StorageSvc/IClassLoader.h"
#include "StorageSvc/IOODatabase.h"
#include "StorageSvc/DbDatabase.h"
#include "StorageSvc/DbOption.h"
#include "StorageSvc/DbDomain.h"
#include "StorageSvc/DbOptionCallback.h"
#include "StorageSvc/DbPrint.h"

#include <iostream>
#include <cerrno>

// Root include files
#include "TFile.h"
#include "TTree.h"
#include "TSystem.h"

using namespace pool;
namespace pool {
  void patchStreamers(RootDBase* db, TFile* file);
}

/// Standard Constuctor
RootDBase::RootDBase(IOODatabase* idb)
: DbDatabaseImp(idb), m_file(0), 
  m_defCompression(1), 
  m_defSplitLevel(99),
  m_defAutoSave(16*1024*1024),
  m_defBufferSize(16*1024),
  m_defWritePolicy(TObject::kOverwrite)   // On write create new versions
{
  m_version = "1.1";
  DbInstanceCount::increment(this);
  m_counters[READ_COUNTER] = m_counters[WRITE_COUNTER] = m_counters[OTHER_COUNTER] = 0;
}

RootDBase::~RootDBase()  {
  DbInstanceCount::decrement(this);
  deletePtr(m_file);
}

bool RootDBase::exists(const std::string& nam)   {
  Bool_t result = gSystem->AccessPathName(nam.c_str(), kFileExists);
  return (result == kFALSE) ? true : false;
}

/// Access the size of the database: May be undefined for some technologies
long long int RootDBase::size()  const   {
  if ( m_file )  {
    return m_file->GetSize();
  }
  return -1;
}

/// Callback after successful open of a database object
DbStatus RootDBase::onOpen(const DbDatabase& dbH, DbAccessMode mode)  {
  std::string par_val;
  DbPrint log(dbH.name());
  if ( !dbH.param("FORMAT_VSN", par_val).isSuccess() )  {
    if ( mode&pool::CREATE || mode&pool::UPDATE ) {
      return dbH.addParam("FORMAT_VSN", m_version);
    }
    log << DbPrintLvl::Warning << "No ROOT data format parameter present "
        << "and file not opened for update." << DbPrint::endmsg;
  }
  else {
    m_version = par_val;
  }
  if ( m_file )  {
    log << DbPrintLvl::Debug << "Root file version:" << int(m_file->GetVersion())
        << DbPrint::endmsg;
  }
  else  {
    log << DbPrintLvl::Error << "Unknown Root file ..."
        << DbPrint::endmsg;
  }
  DbOptionCallback::getOptions(this, DbOptionCallback::_DATABASE_ONOPEN, dbH.name());
  patchStreamers(this, m_file);
  return Success;
}

// Open a new root Database: Access the TFile
DbStatus RootDBase::open(const DbDomain& domH,const std::string& nam,DbAccessMode mode)   {
  DbPrint log(nam);
  const char* fname = nam.c_str();
  Bool_t result = ( mode == pool::READ ) ? kFALSE : gSystem->AccessPathName(fname, kFileExists);
  DbOption opt1("DEFAULT_COMPRESSION","");
  DbOption opt2("DEFAULT_SPLITLEVEL","");
  DbOption opt3("DEFAULT_AUTOSAVE","");
  DbOption opt4("DEFAULT_BUFFERSIZE","");
  domH.getOption(opt1);
  domH.getOption(opt2);
  domH.getOption(opt3);
  domH.getOption(opt4);
  opt1._getValue(m_defCompression);
  opt2._getValue(m_defSplitLevel);
  opt3._getValue(m_defAutoSave);
  opt4._getValue(m_defBufferSize);
  //gDebug = 2;
  DbOptionCallback::getOptions(this, DbOptionCallback::_DATABASE_OPT, nam);
  TDirectory::TContext dirCtxt(0);
  if ( mode == pool::READ )   {
    m_file = TFile::Open(fname);
  }
  else if ( mode&pool::UPDATE && result == kFALSE )    {
    m_file = TFile::Open(fname, "UPDATE", fname, m_defCompression);
  }
  else if ( pool::RECREATE == (mode&pool::RECREATE) )   {
    m_file = TFile::Open(fname, "RECREATE", fname, m_defCompression);
  }
  else if ( mode&pool::CREATE && result == kTRUE )   {
    m_file = TFile::Open(fname, "RECREATE", fname, m_defCompression);
  }
  else if ( mode&pool::CREATE && result == kFALSE )   {
    log << DbPrintLvl::Error << "You cannot open a ROOT file in mode CREATE"
        << " if it already exists. " << DbPrint::endmsg
        << "Use the RECREATE flag if you wish to overwrite."
        << DbPrint::endmsg;
  }
  else if ( mode&pool::UPDATE && result == kTRUE )   {
    log << DbPrintLvl::Error << "You cannot open a ROOT file in mode UPDATE"
        << " if it does not exists. " << DbPrint::endmsg
        << "Use the CREATE or RECREATE flag."
        << DbPrint::endmsg;
  }
  if ( m_file )   {
    log << DbPrintLvl::Always << "Root file version:" << m_file->GetVersion() << DbPrint::endmsg;
    if ( !m_file->IsOpen() )   {
      deletePtr(m_file);
      log << DbPrintLvl::Error << "Failed to open file:" << nam << DbPrint::endmsg;
    }
  }
  else if ( mode == pool::READ )   {
    log << DbPrintLvl::Error << "You cannot open a ROOT file in mode READ"
        << " if it does not exists. " << DbPrint::endmsg;
  }
  return (0 == m_file) ? Error : Success;
}

/// Re-open database with changing access permissions
DbStatus RootDBase::reopen(DbAccessMode mode)   {
  int result = -1;
  if ( m_file )   {
    TDirectory::TContext dirCtxt(0);
    if ( mode == pool::READ ) {
      result = m_file->ReOpen("READ");
    }
    else if ( mode == pool::UPDATE )  {
      result = m_file->ReOpen("UPDATE");
    }
    else  {
      const char* nam = (m_file) ? m_file->GetName() : "UNKNOWN";
      DbPrint log(  nam);
      log << DbPrintLvl::Error << "Failed to reopen file: " << nam;
      log << " in mode " << accessMode(mode) << DbPrint::endmsg;
    }
  }
  return (0 == result) ? Success : Error;
}

static void printErrno(DbPrint& log, const char* nam, int err) {
  switch( err )  {
    case 0:      /// No error - all fine.
      break;
    case EXDEV:  /// Cross-device link. Attempt to move a file to a different device 
    case ENOENT: /// No such file or directory.
    case EINVAL: /// Invalid argument. An invalid value was given
    case EACCES: /// Permission denied. The permission setting does not allow 
    case EBADF:  /// file descriptor is invalid or file not open for writing
    case ENOSPC: /// not enough space left on the device for the operation.
    default:
      {
        log << DbPrintLvl::Error << "Failed to modify file: " << nam
          << " Errno=" << err << " " << strerror(err) << DbPrint::endmsg;
        break;
      }
  }
}

/// Close the root Database: in CREATE/Update mode write the file header...
DbStatus RootDBase::close(DbAccessMode /* mode */ )  {
  if ( m_file )   {
    if ( m_file->IsOpen() )     {
      DbPrint log(  m_file->GetName() );
      const char* nam = (m_file) ? m_file->GetName() : "UNKNOWN";
      DbOptionCallback::getOptions(this, DbOptionCallback::_DATABASE_ONCLOSE, nam);
      if ( byteCount(WRITE_COUNTER) > 0 )   {
        TDirectory::TContext dirCtxt(0);
        m_file->ResetErrno();
        m_file->Write("0", m_defWritePolicy);
        printErrno(log, nam, m_file->GetErrno());
        //m_file->ls();
        m_file->ResetErrno();
        m_file->Close();
        printErrno(log, nam, m_file->GetErrno());
        if (m_file->GetSeekKeys() == 0)   {
          log << DbPrintLvl::Error << "Failed to close file: " << nam
            << " something wrong happened in Close." << DbPrint::endmsg;
        }
      }
      log << DbPrintLvl::Debug
          << "I/O READ  Bytes: " << byteCount(READ_COUNTER)  << DbPrint::endmsg
          << "I/O WRITE Bytes: " << byteCount(WRITE_COUNTER) << DbPrint::endmsg
          << "I/O OTHER Bytes: " << byteCount(OTHER_COUNTER) << DbPrint::endmsg;
    }
  }
  return deletePtr(m_file);
}

/// Do some statistics: add number of bytes read/written/other
void RootDBase::addByteCount(int which, long long int num_bytes)  {
  switch(which)   {
  case OTHER_COUNTER:
  case WRITE_COUNTER:
  case READ_COUNTER:
    m_counters[which] += num_bytes;
    break;
  default:
    break;
  }
}

/// Do some statistics: retrieve number of bytes read/written/other
long long int RootDBase::byteCount(int which) const   {
  switch(which)   {
  case OTHER_COUNTER:
  case WRITE_COUNTER:
  case READ_COUNTER:
    return m_counters[which];
  default:
    return -1;
  }
}

/// Access options
DbStatus RootDBase::getOption(DbOption& opt)  const   { 
  const char* n = opt.name().c_str();
  switch( ::toupper(n[0]) )  {
    case 'C':
      if ( !m_file )
        return Error;
      else if ( !strcasecmp(n, "COMPRESSION_LEVEL") )     // int
        return opt._setValue(int(m_file->GetCompressionLevel()));
      else if ( !strcasecmp(n, "COMPRESSION_FACTOR") )    // float
        return opt._setValue(double(m_file->GetCompressionFactor()));
      break;
    case 'B':
      if ( !m_file )
        return Error;
      else if ( !strcasecmp(n,"BEST_BUFFER") )            // int
        return opt._setValue(int(m_file->GetBestBuffer()));
      else if ( !strcasecmp(n,"BYTES_WRITTEN") )          // double
        return opt._setValue(double(m_file->GetBytesWritten()));
      else if ( !strcasecmp(n,"BYTES_READ") )             // double
        return opt._setValue(double(m_file->GetBytesRead()));
      else if ( !strcasecmp(n,"BYTES_FREE") )             // int
        return opt._setValue(int(m_file->GetNbytesFree()));
      else if ( !strcasecmp(n,"BYTES_INFO") )             // int
        return opt._setValue(int(m_file->GetNbytesInfo()));
      else if ( !strcasecmp(n,"BYTES_KEYS") )             // int
        return opt._setValue(int(m_file->GetNbytesKeys()));
      break;
    case 'D':
      if ( !strcasecmp(n,"DEFAULT_COMPRESSION") )         // int
        return opt._setValue(int(m_defCompression));
      else if ( !strcasecmp(n, "DEFAULT_SPLITLEVEL") )    // int
        return opt._setValue(int(m_defSplitLevel));
      else if ( !strcasecmp(n, "DEFAULT_AUTOSAVE") )      // int
        return opt._setValue(int(m_defAutoSave));   
      else if ( !strcasecmp(n, "DEFAULT_BUFFERSIZE") )    // int
        return opt._setValue(int(m_defBufferSize));
      else if ( !strcasecmp(n, "DEFAULT_WRITEPOLICY") )   // int
        return opt._setValue(int(m_defWritePolicy));
      break;
    case 'N':
      //if ( !strcasecmp(n,"NFREE") )                     // int
      //  return opt._setValue(int(m_file->GetNfree()));
      if ( !m_file )
        return Error;
      else if ( !strcasecmp(n,"NKEYS") )                  // int
        return opt._setValue(int(m_file->GetNkeys()));
      break;
    case 'F':
      if ( !m_file )
        return Error;
      else if ( !strcasecmp(n,"FILEBYTES_WRITTEN") )      // double
        return opt._setValue(double(m_file->GetFileBytesWritten()));
      else if ( !strcasecmp(n,"FILEBYTES_READ") )         // double
        return opt._setValue(double(m_file->GetFileBytesRead()));
      else if ( !strcasecmp(n,"FILE_DESCRIPTOR") )        // int
        return opt._setValue(int(m_file->GetFd()));
      else if ( !strcasecmp(n,"FILE_VERSION") )           // int
        return opt._setValue(int(m_file->GetVersion()));
      else if ( !strcasecmp(n,"FILE_SIZE") )              // int
        return opt._setValue((long long int)m_file->GetSize());
      else if ( !strcasecmp(n,"FILE_ERROR") )             // int
        return opt._setValue(int(m_file->GetErrno()));
      else if ( !strcasecmp(n,"FILE_KEYS") )              // void*
        return opt._setValue((void*)m_file->GetListOfKeys());
      else if ( !strcasecmp(n,"FILE_END") )               // int
        return opt._setValue((long long int)m_file->GetEND());
      break;
    case 'G':
      if ( !m_file )
        return Error;
      else if ( !strcasecmp(n,"GET_OBJECT") )  {         // void*
        const char* key = "";
        opt._getValue(key);
        return opt._setValue((void*)m_file->Get(key));
      }
      break;
    case 'T':
      if ( !strcasecmp(n,"TFILE") )  {                  // void*
        return opt._setValue((void*)m_file);
      }
      else if ( !strcasecmp(n+5,"MAX_SIZE") )  {
        return opt._setValue((long long int)TTree::GetMaxTreeSize());
      }
      break;
    default:
      break;
  }
  return Error;  
}

/// Set options
DbStatus RootDBase::setOption(const DbOption& opt)  { 
  const char* n = opt.name().c_str();
  switch( ::toupper(n[0]) )  {
    case 'C':
      if ( !m_file )
        return Error;
      else if ( !strcasecmp(n, "CD") )  {
        const char* key = "";
        opt._getValue(key);
        m_file->cd(key);
        return Success;
      }
      else if ( !strcasecmp(n, "COMPRESSION_LEVEL") )  {
        int val=1;
        opt._getValue(val);
        m_file->SetCompressionLevel(val);
        return Success;
      }
      break;
    case 'D':
      if (      !strcasecmp(n, "DEFAULT_COMPRESSION") )
        return opt._getValue(m_defCompression);
      else if ( !strcasecmp(n, "DEFAULT_SPLITLEVEL") )
        return opt._getValue(m_defSplitLevel);
      else if ( !strcasecmp(n, "DEFAULT_AUTOSAVE") )      // int
        return opt._getValue(m_defAutoSave);   
      else if ( !strcasecmp(n, "DEFAULT_BUFFERSIZE") )    // int
        return opt._getValue(m_defBufferSize);
      else if ( !strcasecmp(n, "DEFAULT_WRITEPOLICY") )     // int
        return opt._getValue(m_defWritePolicy);
      break;
    case 'F':
      if ( !m_file )
        return Error;
      else if ( !strcasecmp(n,"FILEBYTES_WRITTEN") )   {
        double val = 0;
        opt._getValue(val);
        Long64_t v = (Long64_t)val;
        m_file->SetFileBytesWritten(v);
        return Success;
      }
      else if ( !strcasecmp(n,"FILEBYTES_READ") )  {
        double val = 0;
        opt._getValue(val);
        Long64_t v = (Long64_t)val;
        m_file->SetFileBytesRead(v);
        return Success;
      }
      else if ( !strcasecmp(n,"FILE_FLUSH") )  {
        m_file->Flush();
        return Success;
      }
      break;
    case 'P':
      if ( !m_file )
        return Error;
      else if ( !strcasecmp(n,"PRINT") )  {
        m_file->Print();
        return Success;
      }
      break;
    case 'S':
      if ( !m_file )
        return Error;
      else if ( !strcasecmp(n,"SETMODIFIED") )  {
        m_file->SetModified();
        return Success;
      }
      break;
    case 'T':
      if ( !strcasecmp(n+5,"MAX_SIZE") )  {
        long long int max_size = TTree::GetMaxTreeSize();
        DbStatus sc = opt._getValue(max_size);
        if ( sc.isSuccess() )  {
          TTree::SetMaxTreeSize(max_size);
        }
        return sc;
      }
      break;
    default:
      break;
  }
  return Error;  
}
