// $Id: DbDatabase.h,v 1.26 2008/05/13 14:39:31 witoldp Exp $
//====================================================================
//  DbDatabase and related class definitions
//--------------------------------------------------------------------
//
//  Package    : StorageSvc  (The POOL project)
//  @author      M.Frank
//====================================================================
#ifndef POOL_DBDATABASE_H
#define POOL_DBDATABASE_H

// Framework include files
#include "StorageSvc/pool.h"
#include "StorageSvc/DbHandleBase.h"

// STL include files
#include <vector>

#include "RVersion.h"
// Forward declarations
#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
namespace ROOT { namespace Reflex {  class Type;   }}
#else
namespace Reflex {  class Type;   }
#endif

/*
 *  POOL namespace declaration
 */
namespace pool  {

  // Forward declarations
  class IOODatabase;
  class IDbDatabase;
  class IDbContainer;
  class DbDomain;
  class DbDatabase;
  class DbDatabaseObj;
  class DbContainer;
  class DbContainerObj;
  class DbTransaction;
  class DbOption;
  class DbTypeInfo;
  class Token;
  class Guid;

  /** @class DbDatabase DbDatabase.h StorageSvc/DbDatabase.h
    *
    * Description:
    * Handle managing a DbDatabaseObj, a generic Database object.
    *
    * @author  M.Frank
    * @version 1.0
    */
  class DbDatabase : public DbHandleBase<DbDatabaseObj> {
    /// Reflection class identifier
#if ROOT_VERSION_CODE < ROOT_VERSION(5,19,0)
   typedef const ROOT::Reflex::Type&                 TypeH;
#else
   typedef const Reflex::Type&                 TypeH;
#endif

    /// Parameter definition
    typedef std::pair<std::string, std::string> Parameter;
    /// Parameter container definition
    typedef std::vector< Parameter >            Parameters;

    /// Assign transient object properly (including reference counting)
    void switchPtr(const DbDatabaseObj* obj) const;
  public:
    /// Constructor with initializing arguments
    DbDatabase(const DbType& typ=POOL_StorageType) { m_type = typ;        }
    /// Copy constructor
    DbDatabase(const DbDatabase& cp) : Base()   { switchPtr(cp.m_ptr);    }
    /// Constructor
    DbDatabase(const DbDatabaseObj* obj)        { switchPtr(obj);         }
    /// Standard destructor
    virtual ~DbDatabase()                       { switchPtr(0);           }
    /// Assignment (copy) operator
    DbDatabase& operator=(const DbDatabase& copy) {
      if ( &copy != this )  {
        m_type = copy.m_type;
        switchPtr( copy.m_ptr );
      }
      return *this;
    }
    /// Assignment operator to reset handle using 0
    DbDatabase& operator=(const int /* nuller */ )    {
      switchPtr(0);
      return *this;
    }
    /// Access reference counter
    int refCount()  const;
    /// Allow query if Transaction is active
    bool transactionActive() const;
    /// Access to access mode member
    DbAccessMode openMode() const;
    /// Access to db name
    const std::string& name() const;
    /// Access to db logon string
    const std::string& logon() const;
    /// Access the size of the database: May be undefined for some technologies
    long long int size()  const;
    /// Allow access to the Database implementation
    IOODatabase* db()   const;
    /// Add domain to session
    DbStatus add(const std::string& name, DbContainerObj* cnt) const;
    /// Find domain in session
    DbStatus remove(const DbContainerObj* cnt);
    /// Select container object in Database
    const DbContainerObj* find(const std::string& nam) const;
    /// Access the token of the database object
    const Token* token() const;
    /// Access to domain object
    const DbDomain& containedIn()    const;
    /// Check for existence of Database within domain
    bool exist(const DbDomain& domH, const std::string& nam) const;
    /// Open Database using given domain
    /** @param   domH      [IN]  Valid handle to domain object.
      * @param   pfn       [IN]  Physical file name of the database.
      * @param   lfn       [IN]  Logical file name of the database.
      * @param   mode      [IN]  Open mode (Default=READ).
      * @return Status code indicating success or failure.
      */
    DbStatus open(const DbDomain& domH,
                  const std::string& pfn,
                  const std::string& fid,
                  DbAccessMode mode = pool::READ) const;
    /// Re-open database with changing access permissions
    /** @param   mode      [IN]  Open mode (Valid modes are READ, UPDATE).
      * @return Status code indicating success or failure.
      */
    DbStatus reopen(DbAccessMode mode = pool::READ) const;
    /// Close Database
    DbStatus close() const;
    /// End database access, but still leave database accessible
    DbStatus retire() const;
    /// Check if the database was opened
    bool isOpen() const;
    /// Start/Commit/Rollback Database Transaction
    /** @param   refTr     [IN]  Valid transaction handle.
      * @return Status code indicating success or failure.
      */
    DbStatus transAct(DbTransaction& refTr);
    /// Set options
    /** @param   refOpt    [IN]  Reference to option object.
      * @return Status code indicating success or failure.
      */
    DbStatus setOption(const DbOption& refOpt);
    /// Access options
    /** @param   refOpt    [IN]  Reference to option object.
      * @return Status code indicating success or failure.
      */
    DbStatus getOption(DbOption& refOpt) const;
    /// Add a persistent parameter to the file
    DbStatus addParam(const std::string& nam, const std::string& val) const;
    /// Retrieve the number of user parameters
    int nParam() const;
    /// Retrieve existing parameter by name
    DbStatus param(const std::string& nam, std::string& val) const;
    /// Retrieve all parameters
    DbStatus params(Parameters& vals) const;
    /// Retrieve association link from link container
    DbStatus getLink(const DbLink& lnkH, Token* pToken) const;
    /// Add association link to link container
    DbStatus makeLink(const Token* pToken, DbLink& linkH) const;
    /// Add persistent shape to the Database
    DbStatus addShape (const DbTypeInfo* pShape);
    /// Retrieve persistent type information by class handle
    const DbTypeInfo* objectShape(TypeH typeH)  const;
    /// Retrieve persistent type information by name
    const DbTypeInfo* objectShape(const Guid& nam) const;
    /// Retrieve persistent type information by container
    const DbTypeInfo* contShape(const std::string& nam) const;
    /// Access local container token (if container exists)
    const Token* cntToken(const std::string& cntName) const;
    /// Allow access to all known containers
    DbStatus containers(std::vector<const Token*>& conts, 
                        bool intern=false) const;
    /// Allow access to all known associations between containers
    DbStatus associations(std::vector<const Token*>& assocs) const;
    /// Allow access to all known shapes used by the database
    DbStatus shapes(std::vector<const DbTypeInfo*>& shaps) const;
    /// Let the implementation access the internals
    IDbDatabase* info()  const;
    /// Update database age
    void setAge(int value) const;
    /// Access age value
    int age()  const;
  };
}       // End namespace pool
#endif  // POOL_DBDatabase_H
