// $Id: DbContainer.cpp,v 1.29 2008/10/21 08:39:59 frankb Exp $
//====================================================================
//  DbContainerObj handle implementation
//--------------------------------------------------------------------
//
//  Package    : StorageSvc (The POOL project)
//
//  Description: Generic data persistency
//
//  @author      M.Frank
//====================================================================

// Framework include files
#include "StorageSvc/DbHeap.h"
#include "StorageSvc/DbObject.h"
#include "StorageSvc/DbCallBack.h"
#include "StorageSvc/DbTypeInfo.h"
#include "StorageSvc/DbContainer.h"
#include "StorageSvc/DbTransaction.h"
#include "StorageSvc/DbToken.h"
#include "DbContainerObj.h"

#include <memory>

using namespace std;
using namespace pool;
static string s_empty = "";

/// Add reference count to object if present
int DbContainer::refCount() const  {
  return isValid() ? ptr()->refCount() : int(INVALID);
}

// Open container from handle
DbStatus DbContainer::open( const DbDatabase&  dbH, 
			    const string& nam, 
			    const DbTypeInfo*  typ, 
			    const DbType&      dbtyp,
			    DbAccessMode mod) const
{
  if ( dbH.isValid() )  {
    if ( !(dbH.openMode() == pool::READ && mod != pool::READ) ) {
      const DbContainerObj* q = dbH.find(nam);
      switchPtr(q ? q : new DbContainerObj(dbH, nam, dbtyp, mod));
      if ( mod == pool::READ || mod == pool::UPDATE )  {
        if ( 0 == typ )  {
          typ = dbH.contShape(nam);
        }
      }
      if ( ptr()->open(typ).isSuccess() )  {
        return Success;
      }
      close();
    }
  }
  return Error;
}

void DbContainer::switchPtr(const DbContainerObj* obj) const  {
  if (   obj      ) obj->addRef();
  if ( isValid()  ) m_ptr->release();
  m_ptr = const_cast<DbContainerObj*>(obj);
  if ( m_ptr )  {
    m_type = obj->type();
  }
}

const DbDatabase& DbContainer::containedIn() const {
  static DbDatabase null_dbH(type());
  return isValid() ? m_ptr->database() : null_dbH;
}

long long int DbContainer::size() const {
  return isValid() ? m_ptr->size() : -1;
}

/// Access to access mode member
DbAccessMode DbContainer::openMode() const {
  return isValid() ? m_ptr->mode() : DbAccessMode(pool::NOT_OPEN);
}

/// Access to db name
const string& DbContainer::name() const {
  return isValid() ? m_ptr->name() : s_empty;
}

/// Close container object if handle is valid
DbStatus DbContainer::close() const {
  if ( isValid() )  {
    DbStatus res = m_ptr->close();
    switchPtr(0);
    return res;
  }
  return Error;
}

/// Check if the container was opened
bool DbContainer::isOpen() const  {
  return isValid();
}

/// Allow query if Transaction is active
bool DbContainer::transactionActive() const  {
  return isValid() ? m_ptr->transactionActive() : false;
}

/// Start/Commit/Rollback Database Transaction
DbStatus DbContainer::transAct(DbTransaction& refTr) {
  return (isValid() && refTr.validate(m_ptr).isSuccess()) 
    ? m_ptr->transAct(refTr) : Error;
}

/// Pass options to the implementation
DbStatus DbContainer::setOption(const DbOption& refOpt) {
  return isValid() ? m_ptr->setOption(refOpt) : Error;
}

/// Access options
DbStatus DbContainer::getOption(DbOption& refOpt) const {
  return isValid() ? m_ptr->getOption(refOpt) : Error;
}

/// Create iterator to loop over container elements
IDbCursor* DbContainer::makeCursor(const DbTypeInfo* typ) const {
  return isValid() ? m_ptr->makeCursor(typ) : 0;
}

/// Start/Commit/Rollback Database Transaction
const Token* DbContainer::token() const {
  return isValid() ? m_ptr->token() : 0;
}

/// Access implementation internals
IDbContainer* DbContainer::info() const {
  return isValid() ? m_ptr->info() : 0;
}

/// Allow access to the Database implementation
IOODatabase* DbContainer::db() const {
  return isValid() ? m_ptr->db() : 0;
}

/// Retrieve persistent type information
const DbTypeInfo* 
DbContainer::objectShape(const Guid& guid) const {
  return isValid() ? m_ptr->objectShape(guid) : 0;
}

/// Perform DELETE statement
DbStatus DbContainer::update(DbSelect& sel)  {
  return isValid() ? m_ptr->update(sel) : Error;
}

/// Perform DELETE statement
DbStatus DbContainer::destroy(DbSelect& sel)  {
  return isValid() ? m_ptr->destroy(sel) : Error;
}

/// Perform selection
DbStatus DbContainer::select(DbSelect& sel) const {
  return isValid() ? m_ptr->select(sel) : Error;
}

/// Fetch next object address of the selection to set token
DbStatus DbContainer::fetch(DbSelect& sel) const {
  return isValid() ? m_ptr->fetch(sel) : Error;
}

/// In place allocation of raw memory
void* DbContainer::allocate(unsigned long siz, DataCallBack* call) {
  if ( isValid() && call ) return m_ptr->allocate(siz, *this, call);
  throw bad_alloc();
}

/// In place allocation of object location
DbStatus DbContainer::allocate(DataCallBack* call, DbLink& oid) {
  if ( isValid() && call ) return m_ptr->allocate(*this, call, oid);
  throw bad_alloc();
}

/// In place allocation of raw memory
void* DbContainer::allocate(unsigned long siz, const DbTypeInfo* typ) {
  if ( isValid() )  {
    auto_ptr<DbCallBack> call(new DbCallBack(this, 0, typ));
    void* ptr = m_ptr->allocate(siz, *this, call.get());
    if ( ptr )    {
      call.release();
      return ptr;
    }
  }
  throw bad_alloc();
}

/// In place free of raw memory
DbStatus DbContainer::free(void* ptr, DataCallBack* call) {
  return isValid() && call ? m_ptr->free(ptr,*this,call) : Error;
}

/// In place free of raw memory
DbStatus DbContainer::free(void* ptr, const DbTypeInfo* /* typ*/) {
  return isValid() ? m_ptr->free(ptr, *this, 0) : Error;
}

/// Destroy the persistent representation of the object
DbStatus DbContainer::destroy(const DbLink& linkH)    {
  return isValid() ? m_ptr->destroy(linkH) : Error;
}

/// Load object in the container identified by its handle
DbStatus DbContainer::load(DataCallBack* call, 
			   const DbLink& linkH,
			   DbAccessMode mod)  const {
  if ( isValid() && call )  {
    DbLink oid;
    return m_ptr->load(call, linkH, oid, mod, false);
  }
  return Error;
}

/// Load object in the container identified by its handle
DbStatus DbContainer::loadNext( DataCallBack* call, 
				DbLink& linkH,
				DbAccessMode mod)  const 
{
  if ( isValid() && call )  {
    DbLink oid;
    DbStatus sc = m_ptr->load(call, linkH, oid, mod, true);
    if ( sc.isSuccess() )  {
      linkH = oid;
      return sc;
    }
  }
  return Error;
}

/// Save new object in the container and return its handle
DbStatus DbContainer::save(DataCallBack* call, DbLink& linkH) {
  return isValid() && call ? m_ptr->save(*this,call,linkH) : Error;
}

/// Update object in the container identified by its callback
DbStatus DbContainer::update(DataCallBack* call, const DbLink& linkH) {
  return isValid() && call ? m_ptr->update(*this,call,linkH) : Error;
}

/// Save object in the container identified by its handle
DbStatus DbContainer::_save(const DbObjectHandle<DbObject>& objH,const DbTypeInfo* typ) {
  if ( isValid() && objH.isValid() )    {
    const DbContainer& cnt = objH.containedIn();
    // Object MUST be allocated on this container!
    if ( cnt.info() == m_ptr->info() ) {
      return m_ptr->save( objH, typ );
    }
  }
  return Error;
}

/// Update an object to the container identified by its handle
DbStatus 
DbContainer::_update(const DbObjectHandle<DbObject>& objH,const DbTypeInfo* typ) {
  if ( isValid() && objH.isValid() && typ )    {
    auto_ptr<DbCallBack> call(new DbCallBack(this, &objH, typ));
    DbStatus sc = m_ptr->update(*this, call.get(), objH );
    if ( sc.isSuccess() )  {
      call.release();
    }
    return sc;
  }
  return Error;
}

/// Remove the transient representation of the object from memory
DbStatus DbContainer::_remove(DbObjectHandle<DbObject>& objH)   {
  return isValid() && objH.isValid() ? m_ptr->remove(objH) : Error;
}

/// Destroy the persistent representation of the object
DbStatus DbContainer::_destroy(DbObjectHandle<DbObject>& objH)    {
  if ( isValid() && objH.isValid() ) {
    if ( m_ptr->destroy(objH.oid()).isSuccess() ) {
      return m_ptr->remove(objH);
    }
  }
  return Error;
}

/// Load object in the container identified by its link handle
DbStatus DbContainer::_load(const DbObjectHandle<DbObject>& objH,
			    const DbLink& linkH,
			    const DbTypeInfo* typ,
			    DbAccessMode  mod) const
{
  if ( isValid() && typ )  {
    DbLink oid;
    DbCallBack call(this, &objH, typ);
    if ( m_ptr->load(&call, linkH, oid, mod, false).isSuccess() )  {
      objH.oid() = oid;
      return Success;
    }
  }
  return Error;
}

/// Load object in the container identified by its link handle
DbStatus DbContainer::_loadNext(const DbObjectHandle<DbObject>& objH,
				DbLink& linkH,
				const DbTypeInfo* typ,
				DbAccessMode  mod) const
{
  if ( isValid() && typ )  {
    DbLink oid;
    DbCallBack call(this, &objH, typ);
    if ( m_ptr->load(&call, linkH, oid, mod, true).isSuccess() )  {
      objH.oid() = linkH = oid;
      return Success;
    }
  }
  return Error;
}
