// $Id: DbDatabase.cpp,v 1.22 2008/10/21 08:39:59 frankb Exp $
//====================================================================
//  DbDatabase handle implementation
//--------------------------------------------------------------------
//
//  Package    : StorageSvc (The POOL project)
//
//  Description: Generic data persistency
//
//  @author      M.Frank
//====================================================================

#include "StorageSvc/DbDomain.h"
#include "StorageSvc/DbDatabase.h"
#include "StorageSvc/DbTransaction.h"
#include "DbContainerObj.h"
#include "DbDatabaseObj.h"
using namespace std;
using namespace pool;

static const string& s_empty = "";

/// Add reference count to object if present
int DbDatabase::refCount() const  
{  return isValid() ? m_ptr->refCount() : int(INVALID);                 }

/// Get hosting domain
const DbDomain& DbDatabase::containedIn() const   {
  static DbDomain null_dom(m_type);
  return isValid() ? m_ptr->domain() : null_dom;
}

/// Release handle/switch to new pointer
void DbDatabase::switchPtr(const DbDatabaseObj* obj) const {
  if (   obj ) obj->addRef();
  if ( m_ptr ) m_ptr->release();
  m_ptr = const_cast<DbDatabaseObj*>(obj);
  if ( m_ptr )  {
    setType(obj->type());
  }
}

/// Check if Database exisst within scope of domain
bool DbDatabase::exist(const DbDomain& dH, const string& n) const 
{  return dH.isValid() ? dH.existsDbase(n) : false;                     }

/// Open Database
DbStatus DbDatabase::open(const DbDomain& domH, 
                          const string&   pfn, 
                          const string&   fid, 
                          DbAccessMode    mode)  const  
{
  if ( domH.isValid() )   {
    DbDatabaseObj* db = domH.find(fid);
    if ( !db )   {
      db = new DbDatabaseObj(domH, pfn, fid, mode);
      if ( !db->open().isSuccess() )    {
        switchPtr(0);
        domH.remove(db);
        return Error;
      }
    }
    switchPtr(db);
    m_type = domH.type();
    return Success;
  }
  return Error;
}

/// Close Database
DbStatus DbDatabase::close()  const  {
  if ( isValid() ) m_ptr->close();
  switchPtr(0);  
  return Success;
}

/// Access the size of the database: May be undefined for some technologies
long long int DbDatabase::size()  const   
{  return isValid() ? m_ptr->size() : -1;                               }

/// Re-open database with changing access permissions
DbStatus DbDatabase::reopen(DbAccessMode mod) const 
{  return isValid() ? m_ptr->reopen(mod) : Error;                       }

/// End database access, but still leave database accessible
DbStatus DbDatabase::retire() const  
{  return isValid() ? m_ptr->retire() : Error;                          }

/// Add domain to session
DbStatus DbDatabase::add(const string& nam, DbContainerObj* cnt) const 
{  return (isValid() && cnt) ? m_ptr->add(nam, cnt) : Error;            }

/// Remove domain from session
DbStatus DbDatabase::remove(const DbContainerObj* cnt)   
{  return (isValid() && cnt) ? m_ptr->remove(cnt) : Error;              }

/// Access to access mode member
DbAccessMode DbDatabase::openMode() const 
{  return isValid() ? m_ptr->mode() : DbAccessMode(pool::NOT_OPEN);     }

/// Access to db name
const string& DbDatabase::name() const 
{  return isValid() ? m_ptr->name() : s_empty;                          }

/// Access to db name
const string& DbDatabase::logon() const 
{  return isValid() ? m_ptr->logon() : s_empty;                         }

/// Find domain in session
const DbContainerObj* DbDatabase::find(const string& nam) const 
{  return isValid() ? m_ptr->find(nam) : 0;                             }

/// Check if the database was opened
bool DbDatabase::isOpen() const  
{  return isValid();                                                    }

/// Add association link to link container
DbStatus DbDatabase::makeLink(const Token* pToken, DbLink& refLink) const
{  return isValid() ? m_ptr->makeLink(pToken, refLink) : Error;         }

/// Retrieve association link from link container
DbStatus DbDatabase::getLink(const DbLink& lnkH, Token* pTok) const
{  return isValid() ? m_ptr->getLink(lnkH, pTok) : Error;               }

/// Allow query if Transaction is active
bool DbDatabase::transactionActive() const  
{  return isValid() ? m_ptr->transactionActive() : false;               }

/// Start/Commit/Rollback Database Transaction
DbStatus DbDatabase::transAct(DbTransaction& refTr) {
  return isValid() && refTr.validate(m_ptr).isSuccess() ? m_ptr->transAct(refTr) : Error;
}

/// Pass options to the implementation
DbStatus DbDatabase::setOption(const DbOption& refOpt) 
{  return isValid() ? m_ptr->setOption(refOpt) : Error;                 }

/// Pass options from the implementation to the user
DbStatus DbDatabase::getOption(DbOption& refOpt)  const 
{  return isValid() ? m_ptr->getOption(refOpt) : Error;                 }

/// Add persistent type to the Database
DbStatus DbDatabase::addShape(const DbTypeInfo* refType)   
{  return isValid() ? m_ptr->addShape(refType) : Error;                 }

/// Retrieve persistent type information
const DbTypeInfo* DbDatabase::objectShape(const Guid& guid) const 
{  return isValid() ? m_ptr->objectShape(guid) : 0;                     }

/// Retrieve persistent type information by class handle
const DbTypeInfo* DbDatabase::objectShape(TypeH typeH)  const 
{  return isValid() ? m_ptr->objectShape(typeH) : 0;                    }

/// Retrieve persistent type information
const DbTypeInfo* DbDatabase::contShape(const string& nam) const 
{  return isValid() ? m_ptr->contShape(nam) : 0;                        }

/// Access local container token (if container exists)
const Token* DbDatabase::cntToken(const string& cntName) const 
{  return isValid() ? m_ptr->cntToken(cntName) : 0;                     }

/// Allow access to all known containers
DbStatus DbDatabase::containers(vector<const Token*>& conts,bool with_internal) const 
{  return isValid() ? m_ptr->containers(conts,with_internal) : Error;   }

/// Allow access to all known associations between containers
DbStatus DbDatabase::associations(vector<const Token*>& assocs) const    
{  return isValid() ? m_ptr->associations(assocs) : Error;              }

/// Allow access to all known shapes used by the database
DbStatus DbDatabase::shapes(vector<const DbTypeInfo*>& shaps) const
{  return isValid() ? m_ptr->shapes(shaps) : Error;                     }

/// Retrieve the number of user parameters
int DbDatabase::nParam() const
{  return isValid() ? m_ptr->nParam() : -1;                             }

/// Add a persistent parameter to the file
DbStatus DbDatabase::addParam(const string& nam, const string& val) const
{  return isValid() ? m_ptr->addParam(nam, val) : Error;                }

/// Retrieve existing parameter by name
DbStatus DbDatabase::param(const string& nam, string& val) const
{  return isValid() ? m_ptr->param(nam, val) : Error;                   }

/// Retrieve all parameters
DbStatus DbDatabase::params(Parameters& vals) const
{  return isValid() ? m_ptr->params(vals) : Error;                      }

/// Access the token of the database object
const Token* DbDatabase::token() const
{  return isValid() ? m_ptr->token() : 0;                               }

/// Access implementation internals
IDbDatabase* DbDatabase::info()   const
{  return isValid() ? m_ptr->info() : 0;                                }

/// Allow access to the Database implementation
IOODatabase* DbDatabase::db()    const
{  return isValid() ? m_ptr->db() : 0;                                  }

/// Update database age
void DbDatabase::setAge(int value)  const 
{  if ( isValid() ) m_ptr->setAge(value);                               }

/// Access age value
int DbDatabase::age()  const
{  return isValid() ? m_ptr->age() : 0;                                 }

