#include "Collection/Collection.h"
#include "Collection/CollectionProxy.h"

#include "FileCatalog/IFileCatalog.h"
#include "FileCatalog/URIParser.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/IDataSvc.h"
#include "DataSvc/Ref.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/Placement.h"
#include "StorageSvc/DbType.h"

#include "POOLCore/Exception.h"

#include "TestDictionary/SimpleTestClass.h"
#include "TestDictionary/TestClassPrimitives.h"

#include "StorageSvc/DbStatus.h"
#include "Reflex/SharedLibrary.h"

#include "CoralBase/AttributeListSpecification.h"
#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"

#include "libname.h"

#include <fstream>
#include <stdexcept>
#include <cstdlib>

const std::string TEST_LIBRARY = "test_TestDictionaryDict";
const std::string FILE_CATALOG_TYPE = "xmlcatalog";
const std::string FILE_CATALOG_NAME = "PoolFileCatalog.xml";
const std::string DB_FILE = "TestDbFile.pool";
const pool::DbType DB_TYPE = pool::ROOTTREE_StorageType;  
const std::string CONTAINER_NAME = "TestContainer";
const std::string COLLECTION1_NAME = "TestCollection1";
const std::string COLLECTION2_NAME = "TestCollection2";
const unsigned int NUM_REF_ATTRIBUTES = 3;
const unsigned int NUM_RUNS = 3;
const unsigned int NUM_EVENTS = 5;

static char* setupEnv();



void   printOut( const coral::AttributeListSpecification&  attrListSpec )
{
   std::cout << "[OVAL] ";
   coral::AttributeListSpecification::const_iterator specIter = attrListSpec.begin();
   do{
      std::cout << specIter->name() << ":" << specIter->typeName();
      ++specIter;
      if( specIter != attrListSpec.end() )
         std::cout << ", "; 
   }while( specIter != attrListSpec.end() );
   std::cout << std::endl;
}


void   printOut( const coral::AttributeList&  attrList )
{
   std::cout << "[OVAL] "
             << "attribute list: size = " << attrList.size()
             << " [ ";
   coral::AttributeList::const_iterator iter = attrList.begin();
   do{
      if( iter->specification().typeName() == "bool" ) {
         std::cout << (int) iter->data<bool>();
      } else {
         iter->toOutputStream(std::cout, true);
      }
      std::cout << " as " << iter->specification().name() << ":"
                << iter->specification().typeName();
      ++iter;
      if( iter != attrList.end() )
         std::cout << ", ";
   }while( iter != attrList.end() );
   std::cout << " ]" << std::endl; 
}




int main(int argc, const char *argv[])
{
  try  {
    std::vector<std::string> backends;

    if( argc < 2 )    {
      backends.push_back( "RootCollection" );
      backends.push_back( "RelationalCollectionMySQL" );
      backends.push_back( "RelationalCollectionOracle" );	
    } else {
      backends.push_back( std::string( argv[1] ) );
    }
    char *authPath = setupEnv();

    for ( std::vector<std::string>::const_iterator iBackend = backends.begin();
      iBackend != backends.end(); ++iBackend ) {

        pool::DbStatus sc = pool::DbStatus::Success;
        std::string collectionDb = "";

        std::string collectionType = *iBackend;
        if( collectionType == "RootCollection" )
        {
          collectionDb = "";
        }
        else if( collectionType == "MySQLCollection" || collectionType ==
          "MySQLltCollection" )
        {
          // Retrieve the credentials
#ifdef WIN32
          std::string separator = "\\";
#else
          std::string separator = "/";
#endif
          std::string fileForCredentials = std::string(authPath) + separator + "mysql.pool.pw";
          std::ifstream is( fileForCredentials.c_str() );
          is >> collectionDb;
        }
        else if ( collectionType == "RelationalCollectionMySQL" )
        {
          //collectionDb = "mysql://pcitdb59.cern.ch/ralcolls";
          //collectionDb = "mysql://lxfs6021/CollectionsTest";
          collectionDb = "mysql://itrac424.cern.ch/RAL_WRITER";
          collectionType = "RelationalCollection";
        }
        else if ( collectionType == "RelationalCollectionOracle" )
        {
          collectionDb = "oracle://devdb10/ral_writer";
          collectionType = "RelationalCollection";
        }
        else
        {
          std::cerr << "Collection_Write: Collection type " << collectionType << " is not presently handled by this program." << std::endl;
          return 1;
        }

        std::cout << "[OVAL] Collection_Write: Initialising the SEAL services."
          << std::endl;

        std::cout << "[OVAL] Collection_Write: Loading the shared libraries." 
          << std::endl;

	Reflex::SharedLibrary libloader( libname(TEST_LIBRARY) );
	
	if(!libloader.Load()){
	  std::cout << libloader.Error() << std::endl;
	}

        std::cout << "[OVAL] Collection_Write: Creating a file catalog." 
          << std::endl;
        pool::IFileCatalog* fileCatalog = new pool::IFileCatalog;
        pool::URIParser p( FILE_CATALOG_TYPE + "_file:" + FILE_CATALOG_NAME );
        p.parse();
        fileCatalog->setWriteCatalog( p.contactstring() );

        std::cout << "[OVAL] Collection_Write: Establishing a connection to the "
          << "file catalog." << std::endl;
        fileCatalog->connect();

        std::cout << "[OVAL] Collection_Write: Starting the file catalog "
          << "transaction." << std::endl;
        fileCatalog->start();

        std::cout << "[OVAL] Collection_Write: Creating a data service instance." 
          << std::endl;
        pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( fileCatalog );

        std::cout << "[OVAL] Collection_Write: Setting the database connection "
          << "policy for this session." << std::endl;
        pool::DatabaseConnectionPolicy policy;
        policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
        policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
        dataSvc->session().setDefaultConnectionPolicy( policy );
        std::cout << "[OVAL] Collection_Write: Write mode for a non-existing "
          << "database is set to: " << policy.writeModeForNonExisting() << std::endl;
        std::cout << "[OVAL] Collection_Write: Write mode for an existing "
          << "database is set to: " << policy.writeModeForExisting() << std::endl;

        std::cout << "[OVAL] Collection_Write: Creating a Placement object " 
          << "to store the data object" << std::endl;
        pool::Placement placement;
        placement.setDatabase( DB_FILE, pool::DatabaseSpecification::PFN );
        placement.setContainerName( CONTAINER_NAME );
        placement.setTechnology( DB_TYPE.type() );

        char attrNum[128];

        std::vector<std::string> refAttrName(NUM_REF_ATTRIBUTES);
        for(unsigned int i=0; i<NUM_REF_ATTRIBUTES; i++)
        {
          std::sprintf(attrNum, "%i", i+1);
          refAttrName[i] = "test_ref_";
          refAttrName[i] += attrNum;
        }

        std::cout << "[OVAL] Collection_Write: Creating a Placement object" 
          << " to store the Ref attributes." << std::endl;
        std::vector<pool::Placement> placeRefAttr(NUM_REF_ATTRIBUTES);
        std::string contName;
        for(unsigned int i=0; i<NUM_REF_ATTRIBUTES; i++)
        {
          std::sprintf(attrNum, "%i", i+1);
          placeRefAttr[i].setDatabase( DB_FILE, pool::DatabaseSpecification::PFN );
          contName = CONTAINER_NAME;
          contName += "Ref";
          contName += attrNum;
          placeRefAttr[i].setContainerName( contName );
          placeRefAttr[i].setTechnology( DB_TYPE.type() );
        }

        std::cout << "[OVAL] Collection_Write: Starting a database transaction " 
          << "for UPDATE." << std::endl;
        sc = dataSvc->transaction().start( pool::ITransaction::UPDATE );
        if( !sc.isSuccess() )
        {
          std::cout << "Collection_Write: Error when starting database transaction."
            << std::endl;
        }

        std::cout << "[OVAL] Collection_Write: Creating primitive test attributes" 
          << std::endl;
        TestClassPrimitives tcPrim;
        tcPrim.setNonZero();

        std::cout << std::endl;
        std::cout << "[OVAL] Collection_Write: Creating a collection object:" 
          << std::endl;
        std::cout << "[OVAL] Collection type is " << collectionType << std::endl;
        std::cout << "[OVAL] Collection location is " << collectionDb << std::endl;
        std::cout << "[OVAL] Collection name is " << COLLECTION1_NAME << std::endl;
        pool::Collection<SimpleTestClass> collection(dataSvc, collectionType, 
          collectionDb, COLLECTION1_NAME,
          pool::ICollection::CREATE_AND_OVERWRITE);
        if( !collection.isOpen() ) {
          std::cout << "[OVAL] Collection creation FAILED!";
          std::cout << std::endl;
          exit( 1 );
        }
        std::cout << std::endl;

        std::cout << "[OVAL] Collection_Write: Defining the attribute list "
          << "specification." << std::endl;
        coral::AttributeListSpecification& attribSpec1 = 
           collection.attributeListSpecification();
        attribSpec1.extend( "run", "unsigned int" );
        attribSpec1.extend( "event", "unsigned int" );
        attribSpec1.extend( "test_bool", "bool" );
        attribSpec1.extend( "test_short", "short" );
        attribSpec1.extend( "test_ushort", "unsigned short" );
        attribSpec1.extend( "test_sshort", "short" );
        attribSpec1.extend( "test_shortint", "short" );
        attribSpec1.extend( "test_ushortint", "unsigned short" );
        attribSpec1.extend( "test_sshortint", "short" );
        attribSpec1.extend( "test_int", "int" );
        attribSpec1.extend( "test_uint", "unsigned int" );
        attribSpec1.extend( "test_sint", "int" );
        attribSpec1.extend( "test_s", "int" );
        attribSpec1.extend( "test_u", "unsigned int" );
        attribSpec1.extend( "test_long", "long" );
        attribSpec1.extend( "test_ulong", "unsigned long" );
        attribSpec1.extend( "test_slong", "long" );
        attribSpec1.extend( "test_longint", "long" );
        attribSpec1.extend( "test_ulongint", "unsigned long" );
        attribSpec1.extend( "test_slongint", "long" );
        attribSpec1.extend( "test_float", "float" );
        attribSpec1.extend( "test_double", "double" );
        attribSpec1.extend( "test_string", "string" );
        for (unsigned int i =0; i<NUM_REF_ATTRIBUTES; i++)
        {
          attribSpec1.extend( refAttrName[i], "string" );
        }

        printOut(attribSpec1);

        std::cout << std::endl;
        std::cout << "[OVAL] Collection_Write: Adding objects and attributes "
          << "to the collection:" << std::endl;
        std::cout << std::endl;

        for (unsigned int run = 1; run <= NUM_RUNS; run++)
        {
          for (unsigned int evt = 1; evt <= NUM_EVENTS; evt++)
          {
            std::cout << "[OVAL] Creating a SimpleTestClass object."
              << std::endl;
            pool::Ref<SimpleTestClass> 
              simpleTestClass( dataSvc, new SimpleTestClass() );

            std::cout << "[OVAL] Marking the object for writing."
              << std::endl;
            simpleTestClass.markWrite( placement );

            std::cout << "[OVAL] Setting the object's data." << std::endl;
            simpleTestClass->setNonZero();

            std::cout << "[OVAL] The data is: ";
            simpleTestClass->streamOut(std::cout);
            std::cout << std::endl;

            std::cout << "[OVAL] Setting the object's attributes."
              << std::endl;
            coral::AttributeList attributeList( attribSpec1 );
            attributeList[ "run" ].setValue( run );
            attributeList[ "event" ].setValue( evt );
            attributeList[ "test_bool" ].setValue( tcPrim.m_bool );
            attributeList[ "test_short" ].setValue( tcPrim.m_short );
            attributeList[ "test_ushort" ].setValue( tcPrim.m_ushort );
            attributeList[ "test_sshort" ].setValue( tcPrim.m_sshort );
            attributeList[ "test_shortint" ].setValue( tcPrim.m_shortint );
            attributeList[ "test_ushortint" ].setValue( tcPrim.m_ushortint );
            attributeList[ "test_sshortint" ].setValue( tcPrim.m_sshortint );
            attributeList[ "test_int" ].setValue( tcPrim.m_int );
            attributeList[ "test_uint" ].setValue( tcPrim.m_uint );
            attributeList[ "test_sint" ].setValue( tcPrim.m_sint );
            attributeList[ "test_s" ].setValue( tcPrim.m_s );
            attributeList[ "test_u" ].setValue( tcPrim.m_u );
            attributeList[ "test_long" ].setValue( tcPrim.m_long );
            attributeList[ "test_ulong" ].setValue( tcPrim.m_ulong );
            attributeList[ "test_slong" ].setValue( tcPrim.m_slong );
            attributeList[ "test_longint" ].setValue( tcPrim.m_longint );
            attributeList[ "test_ulongint" ].setValue( tcPrim.m_ulongint );
            attributeList[ "test_slongint" ].setValue( tcPrim.m_slongint );
            attributeList[ "test_float" ].setValue( tcPrim.m_float );
            attributeList[ "test_double" ].setValue( tcPrim.m_double );
            attributeList[ "test_string" ].setValue( std::string("Test String") );
            for (unsigned int i=0; i<NUM_REF_ATTRIBUTES; i++)
            {
              pool::Ref<SimpleTestClass> 
                refAttr( dataSvc, new SimpleTestClass() );
              refAttr.markWrite( placeRefAttr[i] );
              refAttr->setNonZero();
              std::cout << "[OVAL] Ref attribute " << refAttrName[i] 
              << " has its data set to ";
              refAttr->streamOut(std::cout);
              std::cout << std::endl;
              attributeList[ refAttrName[i] ].setValue( refAttr.toString() );
            }

            printOut( attributeList );
            std::cout << std::endl;

            std::cout << "[OVAL] Adding object and attributes to "
              << "collection." << std::endl; 
            collection.add( simpleTestClass, attributeList );

            std::cout << std::endl;
          }
        }

        std::cout << "[OVAL] Collection_Write: Committing all changes made to the "
          "collection." << std::endl;
        collection.commit();

        std::cout << "[OVAL] Collection_Write: Closing the collection." 
          << std::endl;
        collection.close();

        std::cout << std::endl;
        std::cout << "[OVAL] Collection_Write: Creating a collection proxy "
          << "object:" << std::endl;
        std::cout << "[OVAL] Collection type is " << collectionType << std::endl;
        std::cout << "[OVAL] Collection location is " << collectionDb << std::endl;
        std::cout << "[OVAL] Collection name is " << COLLECTION2_NAME
          << std::endl;
        pool::CollectionProxy collectionProxy(dataSvc, collectionType,
          collectionDb, COLLECTION2_NAME, pool::ICollection::CREATE_AND_OVERWRITE);
        std::cout << std::endl;

        std::cout << "[OVAL] Collection_Write: Defining the attribute list "
          << "specification." << std::endl;
        coral::AttributeListSpecification& attribSpec2 =
          collectionProxy.attributeListSpecification();
        attribSpec2.extend( "run", "unsigned int" );
        attribSpec2.extend( "event", "unsigned int" );
        attribSpec2.extend( "test_bool", "bool" );
        attribSpec2.extend( "test_short", "short" );
        attribSpec2.extend( "test_ushort", "unsigned short" );
        attribSpec2.extend( "test_sshort", "short" );
        attribSpec2.extend( "test_shortint", "short" );
        attribSpec2.extend( "test_ushortint", "unsigned short" );
        attribSpec2.extend( "test_sshortint", "short" );
        attribSpec2.extend( "test_int", "int" );
        attribSpec2.extend( "test_uint", "unsigned int" );
        attribSpec2.extend( "test_sint", "int" );
        attribSpec2.extend( "test_s", "int" );
        attribSpec2.extend( "test_u", "unsigned int" );
        attribSpec2.extend( "test_long", "long" );
        attribSpec2.extend( "test_ulong", "unsigned long" );
        attribSpec2.extend( "test_slong", "long" );
        attribSpec2.extend( "test_longint", "long" );
        attribSpec2.extend( "test_ulongint", "unsigned long" );
        attribSpec2.extend( "test_slongint", "long" );
        attribSpec2.extend( "test_float", "float" );
        attribSpec2.extend( "test_double", "double" );
        attribSpec2.extend( "test_string", "string" );
        for (unsigned int i =0; i<NUM_REF_ATTRIBUTES; i++)
        {
          attribSpec2.extend( refAttrName[i], "string" );
        }
        printOut( attribSpec2 );

        std::cout << std::endl;
        std::cout << "[OVAL] Collection_Write: Adding objects and attributes "
          << "to the collection:" << std::endl;
        std::cout << std::endl;
        for (unsigned int run = 1; run <= NUM_RUNS; run++)
        {
          for (unsigned int evt = 1; evt <= NUM_EVENTS; evt++)
          {
            std::cout << "[OVAL] Creating a SimpleTestClass object."
              << std::endl;
            pool::Ref<SimpleTestClass> 
              simpleTestClass( dataSvc, new SimpleTestClass() );

            std::cout << "[OVAL] Marking the object for writing."
              << std::endl;
            simpleTestClass.markWrite( placement );
            if( evt == 1 )
            {
              fileCatalog->commit();
              fileCatalog->start();
            }

            std::cout << "[OVAL] Setting the object's data." << std::endl;
            simpleTestClass->setNonZero();

            std::cout << "[OVAL] The data is: ";
            simpleTestClass->streamOut(std::cout);
            std::cout << std::endl;

            std::cout << "[OVAL] Setting the object's attributes."
              << std::endl;
            coral::AttributeList attributeList( attribSpec2 );
            attributeList[ "run" ].setValue( run );
            attributeList[ "event" ].setValue( evt );
            attributeList[ "test_bool" ].setValue( tcPrim.m_bool );
            attributeList[ "test_short" ].setValue( tcPrim.m_short );
            attributeList[ "test_ushort" ].setValue( tcPrim.m_ushort );
            attributeList[ "test_sshort" ].setValue( tcPrim.m_sshort );
            attributeList[ "test_shortint" ].setValue( tcPrim.m_shortint );
            attributeList[ "test_ushortint" ].setValue( tcPrim.m_ushortint );
            attributeList[ "test_sshortint" ].setValue( tcPrim.m_sshortint );
            attributeList[ "test_int" ].setValue( tcPrim.m_int );
            attributeList[ "test_uint" ].setValue( tcPrim.m_uint );
            attributeList[ "test_sint" ].setValue( tcPrim.m_sint );
            attributeList[ "test_s" ].setValue( tcPrim.m_s );
            attributeList[ "test_u" ].setValue( tcPrim.m_u );
            attributeList[ "test_long" ].setValue( tcPrim.m_long );
            attributeList[ "test_ulong" ].setValue( tcPrim.m_ulong );
            attributeList[ "test_slong" ].setValue( tcPrim.m_slong );
            attributeList[ "test_longint" ].setValue( tcPrim.m_longint );
            attributeList[ "test_ulongint" ].setValue( tcPrim.m_ulongint );
            attributeList[ "test_slongint" ].setValue( tcPrim.m_slongint );
            attributeList[ "test_float" ].setValue( tcPrim.m_float );
            attributeList[ "test_double" ].setValue( tcPrim.m_double );
            attributeList[ "test_string" ].setValue( std::string("Test String") );
            for (unsigned int i=0; i<NUM_REF_ATTRIBUTES; i++)
            {
              pool::Ref<SimpleTestClass> 
                refAttr( dataSvc, new SimpleTestClass() );
              refAttr.markWrite( placeRefAttr[i] );
              refAttr->setNonZero();
              std::cout << "[OVAL] Ref attribute " << refAttrName[i] 
              << " has its data set to ";
              refAttr->streamOut(std::cout);
              std::cout << std::endl;
              attributeList[ refAttrName[i] ].setValue( refAttr.toString() );
            }

            printOut( attributeList );

            std::cout << "[OVAL] Adding object and attributes to "
              << "collection." << std::endl; 
            collectionProxy.add(simpleTestClass, attributeList);

            std::cout << std::endl;
          }
        }

        std::cout << "[OVAL] Collection_Write: Committing all changes made to the "
          "collection." << std::endl;
        collectionProxy.commit();

        std::cout << "[OVAL] Collection_Write: Closing the collection proxy." 
          << std::endl;
        collectionProxy.close();

        std::cout << "[OVAL] Collection_Write: Ending the database transaction."
          << std::endl;
        sc = dataSvc->transaction().commit();
        if( !sc.isSuccess() )
        {
          std::cout << "Collection_Write: Error when ending database transaction."
            << std::endl;
        }

        std::cout << "[OVAL] Collection_Write: Committing all file catalog "
          << "changes." << std::endl;
        fileCatalog->commit();

        std::cout << "[OVAL] Collection_Write: Disconnecting from the database." 
          << std::endl;
        delete dataSvc;

        std::cout << "[OVAL] Collection_Write: Disconnecting from the file "
          << "catalog." << std::endl;
        fileCatalog->disconnect();
        delete fileCatalog;

        std::cout << "[OVAL] Collection_Write: Returning from main()" << std::endl;
      }
      return 0;
  }
  catch( pool::Exception& poolException )
  {
    std::cerr << "pool::Exception: " << poolException.what() << std::endl;
    return 1;
  }
  catch( std::exception& exception )
  {
    std::cerr << "std::Exception: " << exception.what() << std::endl;
    return 1;
  }
  catch( ... )
  {
    std::cerr << "Unknown exception. " << std::endl;
    return 1;
  }
}
#ifdef _WIN32
#include <windows.h>
#endif

static char* setupEnv()  {
  const char* CoralEVar =  "CORAL_AUTH_PATH";
#ifdef _WIN32
  ::SetEnvironmentVariable(CoralEVar,"/afs/cern.ch/sw/lcg/app/pool/db");
#else
  ::setenv(CoralEVar, "/afs/cern.ch/sw/lcg/app/pool/db", 0 );
#endif

  char *authPath = ::getenv( CoralEVar );
  std::cout << "Using Coral Auth path: " << authPath << std::endl;
  return authPath;
}
