#include <stdexcept>

#include "EventService.h"
#include "RefHandler.h"
#include "OriginObject.h"
#include "TargetObject.h"

#include "POOLCore/Token.h"

#include "FileCatalog/URIParser.h"
#include "FileCatalog/IFileCatalog.h"

#include "StorageSvc/DbType.h"

#include "PersistencySvc/IPersistencySvc.h"
#include "PersistencySvc/IPersistencySvcFactory.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/IDatabase.h"
#include "PersistencySvc/IContainer.h"
#include "PersistencySvc/ITokenIterator.h"

EventService::EventService():
  m_originClassType( Reflex::Type::ByName( "OriginObject" ) ),
  m_targetClassType( Reflex::Type::ByName( "TargetObject" ) ),
  m_fileCatalog( 0 ),
  m_persistencySvc( 0 ),
  m_dataHandler(0),
  m_placementHintOrigin( 0 ),
  m_placementHintTarget( 0 ),
  m_originObject( 0 ),
  m_tokenIterator( 0 ),
  m_linksAndTokens(),
  m_tokens()
{
  m_dataHandler = new RefHandler( static_cast< ITokenCache& >( *this ) );
  m_fileCatalog = new pool::IFileCatalog;
  pool::URIParser p;
  p.parse();
  m_fileCatalog->setWriteCatalog( p.contactstring() );
  m_fileCatalog->connect();
  m_fileCatalog->start();

  pool::IPersistencySvcFactory* psfactory = pool::IPersistencySvcFactory::get();
  if ( ! psfactory ) {
    throw std::runtime_error( "Could not retrieve an IPersistencySvc factory" );
  }

  m_persistencySvc = psfactory->create( "PersistencySvc", *m_fileCatalog );
  if ( ! m_persistencySvc ) {
    throw std::runtime_error( "Could not create a PersistencySvc" );
  }

  // Set up the policy.
  pool::DatabaseConnectionPolicy policy;
  policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
  policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
  m_persistencySvc->session().setDefaultConnectionPolicy( policy );

  // Setting up the placement hints
  m_placementHintOrigin = new pool::Placement;
  m_placementHintOrigin->setDatabase( "pool_1.root", pool::DatabaseSpecification::PFN );
  m_placementHintOrigin->setContainerName( "containerOrigin" );
  m_placementHintOrigin->setTechnology( pool::ROOTTREE_StorageType.type() );

  m_placementHintTarget = new pool::Placement;
  m_placementHintTarget->setDatabase( "pool_2.root", pool::DatabaseSpecification::PFN );
  m_placementHintTarget->setContainerName( "containerTarget" );
  m_placementHintTarget->setTechnology( pool::ROOTTREE_StorageType.type() );
}


EventService::~EventService()
{
  m_fileCatalog->commit();
  delete m_fileCatalog;
  delete m_persistencySvc;
  delete m_dataHandler;
  delete m_placementHintOrigin;
  delete m_placementHintTarget;
}


IDataCache*
EventService::cache()
{
  return static_cast< IDataCache* >( this );
}


void
EventService::writeObjectTree( const OriginObject* object )
{
  // Start an update transaction
  if ( ! ( m_persistencySvc->session().transaction().start( pool::ITransaction::UPDATE ) ) ) {
    throw std::runtime_error( "Could not start an update transaction" );
  }

  // Write the top level object
  pool::Token* token = m_persistencySvc->registerForWrite( *m_placementHintOrigin,
							   object,
							   m_originClassType,
							   m_dataHandler );

  // Commit the transaction
  if ( ! m_persistencySvc->session().transaction().commit() ) {
    throw std::runtime_error( "Could not commit the transaction." );
  }

  // Clear the object tree from the memory
  this->clearCache();
  delete object;
  token->release();
}


const OriginObject*
EventService::readNextEvent()
{
  if ( ! m_tokenIterator ) {
    // Initialize the event loop
    if ( ! m_persistencySvc->session().transaction().start( pool::ITransaction::READ ) ) {
      throw std::runtime_error( "Could not start a read transaction.h" );
    }
    pool::IDatabase* db = m_persistencySvc->session().databaseHandle( m_placementHintOrigin->databaseName(),
								      pool::DatabaseSpecification::PFN );
    if ( ! db ) {
      throw std::runtime_error( "Could not retrieve a database handle" );
    }
    db->connectForRead();
    pool::IContainer* container = db->containerHandle( m_placementHintOrigin->containerName() );
    if ( ! container ) {
      throw std::runtime_error( "Could not retrieve the container" );
    }
    m_tokenIterator = container->tokens( "" );
    if ( ! m_tokenIterator ) {
      throw std::runtime_error( "Could not obtain a token iterator" );
    }
  }
  else { // Otherwise clear the previous top level object
    // Clear the object tree from the memory
    this->clearCache();
  }

  pool::Token* token = m_tokenIterator->next();
  if ( token ) {
    std::pair< void*, Reflex::Type > objectAndClass = m_persistencySvc->readObject( *token,
												   m_dataHandler );
    if ( objectAndClass.first == 0 ) {
      throw std::runtime_error( "Could not read the stored data" );
    }
    if ( ! objectAndClass.second ) {
      throw std::runtime_error( "Could not retrieve the stored class" );
    }
    if ( objectAndClass.second != m_originClassType ) {
      throw std::runtime_error( "Did not retrieve the correct class type" );
    }
    m_originObject = reinterpret_cast< OriginObject* >( objectAndClass.first );
    for ( std::vector< DataLink >::const_iterator iLink = m_originObject->data.begin();
	  iLink != m_originObject->data.end(); ++iLink ) {
      this->registerKey( iLink->cacheLocationIndex(), iLink->key() );
    }
    token->release();
  }

  if ( ! m_originObject ) {
    delete m_tokenIterator;
    m_tokenIterator = 0;
    m_persistencySvc->session().transaction().commit();
  }

  return m_originObject;
}


void
EventService::registerObject( TargetObject* object, const std::string& key )
{
  std::map< std::string, std::pair< pool::Token*, TargetObject* > >::iterator iObj = m_linksAndTokens.find( key );
  if ( iObj == m_linksAndTokens.end() ) {
    std::pair< pool::Token*, TargetObject* > p( 0, 0 );
    iObj = m_linksAndTokens.insert( std::make_pair( key, p ) ).first;
  }
  iObj->second.second = object;
}


TargetObject*
EventService::lookup( const std::string& key )
{
  TargetObject* result = 0;
  std::map< std::string, std::pair< pool::Token*, TargetObject* > >::iterator iObj = m_linksAndTokens.find( key );
  if ( iObj != m_linksAndTokens.end() ) {
    result = iObj->second.second;
    if ( result == 0 && iObj->second.first ) {

      std::pair< void*, Reflex::Type >
        objectAndClass = m_persistencySvc->readObject( *(iObj->second.first) );
      if ( objectAndClass.first == 0 ) {
	throw std::runtime_error( "Could not read the stored data" );
      }
      if ( !objectAndClass.second ) {
	throw std::runtime_error( "Could not retrieve the stored class" );
      }
      if ( objectAndClass.second != m_targetClassType ) {
	throw std::runtime_error( "Did not retrieve the correct class type" );
      }
      result = iObj->second.second = reinterpret_cast< TargetObject* >( objectAndClass.first );
    }
  }
  return result;
}


std::string
EventService::key( TargetObject* object ) const
{
  std::string result = "";
  for ( std::map< std::string, std::pair< pool::Token*, TargetObject* > >::const_iterator iObj = m_linksAndTokens.begin();
	iObj != m_linksAndTokens.end(); ++iObj ) {
    if ( iObj->second.second == object ) {
      result = iObj->first;
      break;
    }
  }
  return result;
}


int
EventService::registerToken( pool::Token* token )
{
  int index = 0;
  if ( m_tokens.size() > 0 ) {
    index = m_tokens.rbegin()->first + 1;
  }
  m_tokens.insert( std::make_pair( index, token ) );
  return index;
}


void
EventService::registerKey( int locationIndex, const std::string& key )
{
  std::map< int, pool::Token* >::iterator iEl = m_tokens.find( locationIndex );
  if ( iEl != m_tokens.end() ) {
    pool::Token* token = iEl->second;
    m_tokens.erase( iEl );
    std::map< std::string, std::pair< pool::Token*, TargetObject* > >::iterator iElement = m_linksAndTokens.find( key );
    if ( iElement == m_linksAndTokens.end() ) {
      std::pair< pool::Token*, TargetObject* > p( token, 0 );
      m_linksAndTokens.insert( std::make_pair( key, p ) );
    }
    else {
      iElement->second.first = token;
    }
  }
}


pool::Token*
EventService::saveObject( const TargetObject* object )
{
  pool::Token* token = m_persistencySvc->registerForWrite( *m_placementHintTarget,
							   object,
							   m_targetClassType );
  for ( std::map< std::string, std::pair< pool::Token*, TargetObject* > >::iterator iObj = m_linksAndTokens.begin();
	iObj != m_linksAndTokens.end(); ++iObj ) {
    if ( iObj->second.second == object ) {
      iObj->second.first = token;
      break;
    }
  }
  return token;
}


void
EventService::clearCache()
{
  if ( m_originObject ) {
    delete m_originObject;
    m_originObject = 0;
  }

  for ( std::map< std::string, std::pair< pool::Token*, TargetObject* > >::iterator iElement = m_linksAndTokens.begin();
	iElement != m_linksAndTokens.end(); ++iElement ) {
    pool::Token* token = iElement->second.first;
    if ( token ) token->release();
    TargetObject* object = iElement->second.second;
    if ( object ) delete object;
  }
  m_linksAndTokens.clear();
}
