#include "TestDriver.h"
#include "TestTransactionObserver.h"
#include "libname.h"
#include <stdexcept>
#include <iostream>
#include <sstream>

#include "Reflex/SharedLibrary.h"
#include "Reflex/Type.h"

#include "POOLCore/Token.h"


#include "StorageSvc/DbType.h"

#include "FileCatalog/URIParser.h"
#include "FileCatalog/IFileCatalog.h"

#include "PersistencySvc/Placement.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/ITransaction.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/IDatabase.h"
#include "PersistencySvc/ITechnologySpecificAttributes.h"
#include "PersistencySvc/IContainer.h"
#include "PersistencySvc/ITokenIterator.h"

#include "PersistencySvc/ITokenValidator.h"
#include "PersistencySvc/ITokenValidatorFactory.h"

#include "PersistencySvc/IDataTransform.h"
#include "PersistencySvc/IDataTransformFactory.h"

#include "PersistencySvc/IPersistencySvc.h"
#include "PersistencySvc/IPersistencySvcFactory.h"


static const Reflex::Type emptyType = Reflex::Type();

namespace pool  {
  int numTokenInstances();
}

pool::TestDriver::TestDriver():
  m_fileCatalog( 0 ),
  m_fileName1( "pool1.root" ),
  m_fileName2( "pool2.root" ),
  m_events( 100 ),
  m_eventsToCommitAndHold( 10 )
{
  
  //// pool::POOLContext::setMessageVerbosityLevel( coral::Info );
  
  std::cout << "[OVAL] Creating a file catalog" << std::endl;
  m_fileCatalog = new pool::IFileCatalog;
  if ( ! m_fileCatalog ) {
    throw std::runtime_error( "Could not create a file catalog" );
  }
  pool::URIParser p;
  p.parse();
  m_fileCatalog->setWriteCatalog( p.contactstring() );
  m_fileCatalog->connect();
}

pool::TestDriver::~TestDriver()
{
  if ( m_fileCatalog ) delete m_fileCatalog;
  std::cout << "[OVAL] Number of floating tokens : " << pool::numTokenInstances() << std::endl;
}

void
pool::TestDriver::clearCache()
{
  for ( std::vector< pool::Token* >::iterator iToken = m_tokens.begin();
	iToken != m_tokens.end(); ++iToken ) {
    (*iToken)->release();
  }
  m_tokens.clear();
}

void
pool::TestDriver::loadLibraries( const std::vector<std::string>& libraries )
{
  for ( std::vector<std::string>::const_iterator iLibrary = libraries.begin();
	iLibrary != libraries.end(); ++iLibrary ) {
    const std::string& library = *iLibrary;
    std::cout << "Loading library " << library << std::endl;

    Reflex::SharedLibrary libloader( libname(library) );
    
    if(!libloader.Load()){
      std::cout << libloader.Error() << std::endl;
    }
  }
}

void
pool::TestDriver::write()
{
  pool::IFileCatalog& catalog = *m_fileCatalog;
  catalog.start();

  std::cout << "Creating the persistency service" << std::endl;
  pool::IPersistencySvcFactory* psfactory = pool::IPersistencySvcFactory::get();
  if ( ! psfactory ) {
    throw std::runtime_error( "Could not retrieve an IPersistencySvc factory" );
  }
  std::auto_ptr< pool::IPersistencySvc > persistencySvc( psfactory->create( "PersistencySvc", catalog ) );
  if ( ! persistencySvc.get() ) {
    throw std::runtime_error( "Could not create a PersistencySvc" );
  }

  // Set up the policy.
  pool::DatabaseConnectionPolicy policy;
  policy.setWriteModeForNonExisting( pool::DatabaseConnectionPolicy::CREATE );
  policy.setWriteModeForExisting( pool::DatabaseConnectionPolicy::OVERWRITE );
  persistencySvc->session().setDefaultConnectionPolicy( policy );

  // Start an update transaction
  if ( ! ( persistencySvc->session().transaction().start( pool::ITransaction::UPDATE ) ) ) {
    throw std::runtime_error( "Could not start an update transaction" );
  }

  // Retrieving the class descriptions
  Reflex::Type class_SimpleTestClass = Reflex::Type::ByName( "SimpleTestClass" );
  Reflex::Type class_TestClassPrimitives = Reflex::Type::ByName( "TestClassPrimitives" );
  Reflex::Type class_TestClassSimpleContainers = Reflex::Type::ByName( "TestClassSimpleContainers" );
  Reflex::Type class_TestClassVectors = Reflex::Type::ByName( "TestClassVectors" );

  // Defining the placement objects
  pool::Placement placementHint_SimpleTestClass;
  placementHint_SimpleTestClass.setDatabase( m_fileName1, pool::DatabaseSpecification::PFN );
  placementHint_SimpleTestClass.setContainerName( "SimpleTestClass_Container" );
  placementHint_SimpleTestClass.setTechnology( pool::ROOTKEY_StorageType.type() );

  pool::Placement placementHint_TestClassPrimitives;
  placementHint_TestClassPrimitives.setDatabase( m_fileName1, pool::DatabaseSpecification::PFN );
  placementHint_TestClassPrimitives.setContainerName( "TestClassPrimitives_Container" );
  placementHint_TestClassPrimitives.setTechnology( pool::ROOTTREE_StorageType.type() );

  pool::Placement placementHint_TestClassSimpleContainers;
  placementHint_TestClassSimpleContainers.setDatabase( m_fileName2, pool::DatabaseSpecification::PFN );
  placementHint_TestClassSimpleContainers.setContainerName( "TestClassSimpleContainers_Container" );
  placementHint_TestClassSimpleContainers.setTechnology( pool::ROOTTREE_StorageType.type() );

  pool::Placement placementHint_TestClassVectors;
  placementHint_TestClassVectors.setDatabase( m_fileName2, pool::DatabaseSpecification::PFN );
  placementHint_TestClassVectors.setContainerName( "TestClassVectors_Container" );
  placementHint_TestClassVectors.setTechnology( pool::ROOTTREE_StorageType.type() );

  std::vector< SimpleTestClass* > v_simpleTestClass;
  std::vector< TestClassPrimitives* > v_testClassPrimitives;
  std::vector< TestClassSimpleContainers* > v_testClassSimpleContainers;
  std::vector< TestClassVectors* > v_testClassVectors;
  std::cout << "Writing " << m_events << " objects of each type" << std::endl;
  for ( int i = 0; i < m_events; ++i ) {

    ///////////////////////////////////////////////////////////
    SimpleTestClass* object_SimpleTestClass = new SimpleTestClass();
    v_simpleTestClass.push_back( object_SimpleTestClass );
    object_SimpleTestClass->setNonZero();
    pool::Token* token_SimpleTestClass = persistencySvc->registerForWrite( placementHint_SimpleTestClass,
                                                                           object_SimpleTestClass,
                                                                           class_SimpleTestClass );
    if ( ! token_SimpleTestClass ) {
      throw std::runtime_error( "Could not write an object" );
    }
    m_tokens.push_back( token_SimpleTestClass );
    m_simpleTestClass.push_back( *object_SimpleTestClass );

    ///////////////////////////////////////////////////////////
    TestClassPrimitives* object_TestClassPrimitives = new TestClassPrimitives();
    v_testClassPrimitives.push_back( object_TestClassPrimitives );
    object_TestClassPrimitives->setNonZero();
    pool::Token* token_TestClassPrimitives = persistencySvc->registerForWrite( placementHint_TestClassPrimitives,
                                                                               object_TestClassPrimitives,
                                                                               class_TestClassPrimitives );
    if ( ! token_TestClassPrimitives ) {
      throw std::runtime_error( "Could not write an object" );
    }
    m_tokens.push_back( token_TestClassPrimitives );
    m_testClassPrimitives.push_back( *object_TestClassPrimitives );

    ///////////////////////////////////////////////////////////
    TestClassSimpleContainers* object_TestClassSimpleContainers = new TestClassSimpleContainers();
    v_testClassSimpleContainers.push_back( object_TestClassSimpleContainers );
    object_TestClassSimpleContainers->setNonZero();
    pool::Token* token_TestClassSimpleContainers = persistencySvc->registerForWrite( placementHint_TestClassSimpleContainers,
                                                                                     object_TestClassSimpleContainers,
                                                                                     class_TestClassSimpleContainers );
    if ( ! token_TestClassSimpleContainers ) {
      throw std::runtime_error( "Could not write an object" );
    }
    m_tokens.push_back( token_TestClassSimpleContainers );
    m_testClassSimpleContainers.push_back( *object_TestClassSimpleContainers );

    ///////////////////////////////////////////////////////////
    TestClassVectors* object_TestClassVectors = new TestClassVectors();
    v_testClassVectors.push_back( object_TestClassVectors );
    object_TestClassVectors->setNonZero();
    pool::Token* token_TestClassVectors = persistencySvc->registerForWrite( placementHint_TestClassVectors,
                                                                            object_TestClassVectors,
                                                                            class_TestClassVectors );
    if ( ! token_TestClassVectors ) {
      throw std::runtime_error( "Could not write an object" );
    }
    m_tokens.push_back( token_TestClassVectors );
    m_testClassVectors.push_back( *object_TestClassVectors );


    // Commit and hold the transaction
    if ( ( i + 1 ) % m_eventsToCommitAndHold == 0 ) {
      if ( ! persistencySvc->session().transaction().commitAndHold() ) {
        throw std::runtime_error( "Could not commit and hold the transaction." );
      }
    }
  }


  // Committing the transaction
  std::cout << "Committing the transaction." << std::endl;
  if ( ! persistencySvc->session().transaction().commit() ) {
    throw std::runtime_error( "Could not commit the transaction." );
  }

  // Removing the "cache"
  for ( std::vector< SimpleTestClass* >::iterator iObject = v_simpleTestClass.begin();
        iObject != v_simpleTestClass.end(); ++iObject ) {
    delete *iObject;
  }
  v_simpleTestClass.clear();
  for ( std::vector< TestClassPrimitives* >::iterator iObject = v_testClassPrimitives.begin();
        iObject != v_testClassPrimitives.end(); ++iObject ) {
    delete *iObject;
  }
  v_testClassPrimitives.clear();
  
  for ( std::vector< TestClassSimpleContainers* >::iterator iObject = v_testClassSimpleContainers.begin();
        iObject != v_testClassSimpleContainers.end(); ++iObject ) {
    delete *iObject;
  }
  v_testClassSimpleContainers.clear();

  for ( std::vector< TestClassVectors* >::iterator iObject = v_testClassVectors.begin();
        iObject != v_testClassVectors.end(); ++iObject ) {
    delete *iObject;
  }
  v_testClassVectors.clear();

  // Start an update transaction
  if ( ! ( persistencySvc->session().transaction().start( pool::ITransaction::UPDATE ) ) ) {
    throw std::runtime_error( "Could not start an update transaction" );
  }

  // Committing the transaction
  std::cout << "Committing the transaction." << std::endl;
  if ( ! persistencySvc->session().transaction().commit() ) {
    throw std::runtime_error( "Could not commit the transaction." );
  }

  catalog.commit();
}


void
pool::TestDriver::read()
{
  pool::IFileCatalog& catalog = *m_fileCatalog;
  catalog.start();

  std::cout << "Creating the persistency service" << std::endl;
  pool::IPersistencySvcFactory* psfactory = pool::IPersistencySvcFactory::get();
  if ( ! psfactory ) {
    throw std::runtime_error( "Could not retrieve an IPersistencySvc factory" );
  }
  std::auto_ptr< pool::IPersistencySvc > persistencySvc( psfactory->create( "PersistencySvc", catalog ) );
  if ( ! persistencySvc.get() ) {
    throw std::runtime_error( "Could not create a PersistencySvc" );
  }

  // Starting a read transaction
  if ( ! persistencySvc->session().transaction().start( pool::ITransaction::READ ) ) {
    throw std::runtime_error( "Could not start a read transaction." );
  }

  std::cout << "Reading back the objects." << std::endl;
  const unsigned int numberOfTypes = m_tokens.size() / m_events;
  for ( int i = 0; i < m_events; ++i ) {
    int j = 0;

    ///////////////////////////////////////////////////////////
    std::pair< void*, Reflex::Type > objectAndClass_simpleTestClass = persistencySvc->readObject( *( m_tokens[numberOfTypes*i + j ] ) );
    if ( objectAndClass_simpleTestClass.first == 0 ) {
      throw std::runtime_error( "Could not read the stored data" );
    }
    if ( objectAndClass_simpleTestClass.second == emptyType ) {
      throw std::runtime_error( "Could not retrieve the stored class" );
    }
    if ( objectAndClass_simpleTestClass.second.Name() != "SimpleTestClass" ) {
      throw std::runtime_error( "read wrong class type." );
    }
    SimpleTestClass* object_simpleTestClass = reinterpret_cast< SimpleTestClass* >( objectAndClass_simpleTestClass.first );
    if ( *object_simpleTestClass != m_simpleTestClass[i] ) {
      std::ostringstream error;
      error << "SimpleTestClass object written is different from object read:" << std::endl << "Original : ";
      m_simpleTestClass[i].streamOut( error );
      error << std::endl << "Read from persistency : ";
      object_simpleTestClass->streamOut( error );
      throw std::runtime_error( error.str() );
    }
    delete object_simpleTestClass;
    ++j;

    ///////////////////////////////////////////////////////////
    std::pair< void*, Reflex::Type > objectAndClass_testClassPrimitives = persistencySvc->readObject( *( m_tokens[numberOfTypes*i + j] ) );
    if ( objectAndClass_testClassPrimitives.first == 0 ) {
      throw std::runtime_error( "Could not read the stored data" );
    }
    if ( objectAndClass_testClassPrimitives.second == emptyType ) {
      throw std::runtime_error( "Could not retrieve the stored class" );
    }
    if ( objectAndClass_testClassPrimitives.second.Name() != "TestClassPrimitives" ) {
      throw std::runtime_error( "read wrong class type." );
    }
    TestClassPrimitives* object_testClassPrimitives = reinterpret_cast< TestClassPrimitives* >( objectAndClass_testClassPrimitives.first );
    if ( *object_testClassPrimitives != m_testClassPrimitives[i] ) {
      std::ostringstream error;
      error << "TestClassPrimitives object written is different from object read:" << std::endl << "Original : ";
      m_testClassPrimitives[i].streamOut( error );
      error << std::endl << "Read from persistency : ";
      object_testClassPrimitives->streamOut( error );
      throw std::runtime_error( error.str() );
    }
    delete object_testClassPrimitives;
    ++j;

    ///////////////////////////////////////////////////////////
    std::pair< void*, Reflex::Type > objectAndClass_testClassSimpleContainers = persistencySvc->readObject( *( m_tokens[numberOfTypes*i + j] ) );
    if ( objectAndClass_testClassSimpleContainers.first == 0 ) {
      throw std::runtime_error( "Could not read the stored data" );
    }
    if ( objectAndClass_testClassSimpleContainers.second == emptyType ) {
      throw std::runtime_error( "Could not retrieve the stored class" );
    }
    if ( objectAndClass_testClassSimpleContainers.second.Name() != "TestClassSimpleContainers" ) {
      throw std::runtime_error( "read wrong class type." );
    }
    TestClassSimpleContainers* object_testClassSimpleContainers = reinterpret_cast< TestClassSimpleContainers* >( objectAndClass_testClassSimpleContainers.first );
    if ( *object_testClassSimpleContainers != m_testClassSimpleContainers[i] ) {
      std::ostringstream error;
      error << "TestClassSimpleContainers object written is different from object read:" << std::endl << "Original : ";
      m_testClassSimpleContainers[i].streamOut( error );
      error << std::endl << "Read from persistency : ";
      object_testClassSimpleContainers->streamOut( error );
      throw std::runtime_error( error.str() );
    }
    delete object_testClassSimpleContainers;
    ++j;

    ///////////////////////////////////////////////////////////
    std::pair< void*, Reflex::Type > objectAndClass_testClassVectors = persistencySvc->readObject( *( m_tokens[numberOfTypes*i + j] ) );
    if ( objectAndClass_testClassVectors.first == 0 ) {
      throw std::runtime_error( "Could not read the stored data" );
    }
    if ( objectAndClass_testClassVectors.second == emptyType ) {
      throw std::runtime_error( "Could not retrieve the stored class" );
    }
    if ( objectAndClass_testClassVectors.second.Name() != "TestClassVectors" ) {
      throw std::runtime_error( "read wrong class type." );
    }
    TestClassVectors* object_testClassVectors = reinterpret_cast< TestClassVectors* >( objectAndClass_testClassVectors.first );
    if ( *object_testClassVectors != m_testClassVectors[i] ) {
      std::ostringstream error;
      error << "TestClassVectors object written is different from object read:" << std::endl << "Original : ";
      m_testClassVectors[i].streamOut( error );
      error << std::endl << "Read from persistency : ";
      object_testClassVectors->streamOut( error );
      throw std::runtime_error( error.str() );
    }
    delete object_testClassVectors;
    ++j;

    // Commit and hold the transaction
    if ( ( i + 1 ) % m_eventsToCommitAndHold == 0 ) {
      if ( ! persistencySvc->session().transaction().commitAndHold() ) {
        throw std::runtime_error( "Could not commit and hold the transaction." );
      }
    }

  }

  // Committing the transaction
  std::cout << "Committing the transaction." << std::endl;
  if ( ! persistencySvc->session().transaction().commit() ) {
    throw std::runtime_error( "Could not commit the transaction." );
  }

  catalog.commit();
}


void
pool::TestDriver::testPlugins()
{
  pool::ITokenValidatorFactory* tvfactory = pool::ITokenValidatorFactory::get();
  if ( ! tvfactory ) {
    throw std::runtime_error( "Could not retrieve an ITokenValidator factory" );
  }
  std::cout << "Creating a token validator" << std::endl;
  std::auto_ptr< pool::ITokenValidator > tokenValidator( tvfactory->create( "DummyTokenValidator" ) );
  if ( ! tokenValidator.get() ) {
    throw std::runtime_error( "Could not create an DummyTokenValidator" );
  }

  pool::IDataTransformFactory* dtfactory = pool::IDataTransformFactory::get();
  if ( ! dtfactory ) {
    throw std::runtime_error( "Could not retrieve an IDataTransform factory" );
  }
  std::cout << "Creating a data transform" << std::endl;
  std::auto_ptr< pool::IDataTransform > dataTransform( dtfactory->create( "DataTransform" ) );
  if ( ! dataTransform.get() ) {
    throw std::runtime_error( "Could not create a DataTransform" );
  }

  pool::IPersistencySvcFactory* psfactory = pool::IPersistencySvcFactory::get();
  if ( ! psfactory ) {
    throw std::runtime_error( "Could not retrieve an IPersistencySvc factory" );
  }
}



void
pool::TestDriver::readCollections()
{
  pool::IFileCatalog& catalog = *m_fileCatalog;
  catalog.start();

  std::cout << "Creating the persistency service" << std::endl;
  pool::IPersistencySvcFactory* psfactory = pool::IPersistencySvcFactory::get();
  if ( ! psfactory ) {
    throw std::runtime_error( "Could not retrieve an IPersistencySvc factory" );
  }
  std::auto_ptr< pool::IPersistencySvc > persistencySvc( psfactory->create( "PersistencySvc", catalog ) );
  if ( ! persistencySvc.get() ) {
    throw std::runtime_error( "Could not create a PersistencySvc" );
  }

  // Starting a read transaction
  if ( ! persistencySvc->session().transaction().start( pool::ITransaction::READ ) ) {
    throw std::runtime_error( "Could not start a read transaction." );
  }

  // Opening again a database
  pool::IDatabase* db = persistencySvc->session().databaseHandle( m_fileName1, pool::DatabaseSpecification::PFN );
  if ( ! db ) {
    throw std::runtime_error( "Could not retrieve a database handle" );
  }

  db->connectForRead();

  std::vector< std::string > containers = db->containers();
  for ( std::vector< std::string >::const_iterator iContainer = containers.begin();
	iContainer != containers.end(); ++iContainer ) {
    pool::IContainer* container = db->containerHandle( *iContainer );
    if ( ! container ) {
      throw std::runtime_error( "Could not retrieve the container" );
    }
    std::cout << "Container : " << container->name() << std::endl;
    pool::ITokenIterator* tokenIterator = container->tokens( "" );
    if ( ! tokenIterator ) {
      throw std::runtime_error( "Could not obtain a token iterator" );
    }

    int numberOfObjects = 0;
    pool::Token* token = tokenIterator->next();
    while ( token ) {
      ++numberOfObjects;
      token->release();
      token = tokenIterator->next();
    }

    if ( numberOfObjects != m_events ) {
      throw std::runtime_error( "Unexpected number of objects in that container." );
    }
    else {
      std::cout << numberOfObjects << " objects found in the container." << std::endl;
    }

    delete tokenIterator;
    delete container;
  }

  delete db;

  std::cout << "Committing the transaction." << std::endl;
  if ( ! persistencySvc->session().transaction().commit() ) {
    throw std::runtime_error( "Could not commit the transaction." );
  }

  catalog.commit();
}



void
pool::TestDriver::updateObjects()
{
  pool::IFileCatalog& catalog = *m_fileCatalog;
  catalog.start();

  std::cout << "Creating the persistency service" << std::endl;
  pool::IPersistencySvcFactory* psfactory = pool::IPersistencySvcFactory::get();
  if ( ! psfactory ) {
    throw std::runtime_error( "Could not retrieve an IPersistencySvc factory" );
  }
  std::auto_ptr< pool::IPersistencySvc > persistencySvc( psfactory->create( "PersistencySvc", catalog ) );
  if ( ! persistencySvc.get() ) {
    throw std::runtime_error( "Could not create a PersistencySvc" );
  }

  // Registering transaction observers.
  TestTransactionObserver observer1( "obsevrer 1", std::cout );
  if ( ! persistencySvc->registerObserver( observer1 ) ) {
    throw std::runtime_error( "Could not register observer 1" );
  }
  TestTransactionObserver observer2( "obsevrer 2", std::cout );
  if ( ! persistencySvc->registerObserver( observer2 ) ) {
    throw std::runtime_error( "Could not register observer 2" );
  }

  if ( persistencySvc->registerObserver( observer1 ) ) {
    throw std::runtime_error( "Registered observer 1 twice" );
  }

  // Starting an update transaction
  if ( ! persistencySvc->session().transaction().start( pool::ITransaction::UPDATE ) ) {
    throw std::runtime_error( "Could not start a read transaction." );
  }

  pool::IDatabase* db = persistencySvc->session().databaseHandle( m_fileName1, pool::DatabaseSpecification::PFN );
  if ( ! db ) {
    throw std::runtime_error( "Could not retrieve a database handle" );
  }

  db->connectForRead(); // will open db in read mode.

  pool::IContainer* container = db->containerHandle( "SimpleTestClass_Container" );
  if ( ! container ) {
    throw std::runtime_error( "Could not retrieve the container" );
  }

  pool::ITokenIterator* tokenIterator = container->tokens( "" );
  if ( ! tokenIterator ) {
    throw std::runtime_error( "Could not obtain a token iterator" );
  }

  int i = 0;
  std::vector< SimpleTestClass* > v_simpleTestClass;
  pool::Token* token = tokenIterator->next();
  while ( token ) {
    std::pair< void*, Reflex::Type > objectAndClass_simpleTestClass = persistencySvc->readObject( *token );
    if ( objectAndClass_simpleTestClass.first == 0 ) {
      throw std::runtime_error( "Could not read the stored data" );
    }
    if ( objectAndClass_simpleTestClass.second == emptyType ) {
      throw std::runtime_error( "Could not retrieve the stored class" );
    }
    if ( objectAndClass_simpleTestClass.second.Name() != "SimpleTestClass" ) {
      throw std::runtime_error( "read wrong class type." );
    }
    SimpleTestClass* object_simpleTestClass = reinterpret_cast< SimpleTestClass* >( objectAndClass_simpleTestClass.first );
    object_simpleTestClass->data = i;
    ++i;

    if ( ! persistencySvc->updateObject( object_simpleTestClass,
                                         objectAndClass_simpleTestClass.second,
                                         *token ) ) {
      throw std::runtime_error( "Could not update the obect" );
    }
    v_simpleTestClass.push_back( object_simpleTestClass );
    token->release();
    token = tokenIterator->next();
  }

  delete tokenIterator;
  delete container;
  delete db;

  // Committing
  std::cout << "Committing the transaction." << std::endl;
  if ( ! persistencySvc->session().transaction().commit() ) {
    throw std::runtime_error( "Could not commit the transaction." );
  }

  std::cout << "Clearing the cache" << std::endl;
  for ( std::vector< SimpleTestClass* >::iterator iObject = v_simpleTestClass.begin();
        iObject != v_simpleTestClass.end(); ++iObject ) {
    delete *iObject;
  }

  catalog.commit();

  if ( ! persistencySvc->deregisterObserver( observer1 ) ) {
    throw std::runtime_error( "Could not deregister observer 1" );
  }
  if ( ! persistencySvc->deregisterObserver( observer2 ) ) {
    throw std::runtime_error( "Could not deregister observer 2" );
  }

  if ( persistencySvc->deregisterObserver( observer1 ) ) {
    throw std::runtime_error( "Deregistered observer 1 twice" );
  }
  if ( persistencySvc->deregisterObserver( observer2 ) ) {
    throw std::runtime_error( "Deregistered observer 2 twice" );
  }

}


void
pool::TestDriver::readBackUpdatedObjects()
{
  pool::IFileCatalog& catalog = *m_fileCatalog;
  catalog.start();
  
  std::cout << "Creating the persistency service" << std::endl;
  pool::IPersistencySvcFactory* psfactory = pool::IPersistencySvcFactory::get();
  if ( ! psfactory ) {
    throw std::runtime_error( "Could not retrieve an IPersistencySvc factory" );
  }
  std::auto_ptr< pool::IPersistencySvc > persistencySvc( psfactory->create( "PersistencySvc", catalog ) );
  if ( ! persistencySvc.get() ) {
    throw std::runtime_error( "Could not create a PersistencySvc" );
  }

  // Registering transaction observers.
  TestTransactionObserver observer1( "obsevrer 1", std::cout );
  persistencySvc->registerObserver( observer1 );
  TestTransactionObserver observer2( "obsevrer 2", std::cout );
  persistencySvc->registerObserver( observer2 );

  // Starting a read transaction
  if ( ! persistencySvc->session().transaction().start( pool::ITransaction::READ ) ) {
    throw std::runtime_error( "Could not start a read transaction." );
  }

  pool::IDatabase* db = persistencySvc->session().databaseHandle( m_fileName1, pool::DatabaseSpecification::PFN );
  if ( ! db ) {
    throw std::runtime_error( "Could not retrieve a database handle" );
  }

  db->connectForRead(); // will open db in read mode.

  pool::IContainer* container = db->containerHandle( "SimpleTestClass_Container" );
  if ( ! container ) {
    throw std::runtime_error( "Could not retrieve the container" );
  }

  pool::ITokenIterator* tokenIterator = container->tokens( "" );
  if ( ! tokenIterator ) {
    throw std::runtime_error( "Could not obtain a token iterator" );
  }

  int i = 0;
  std::vector< SimpleTestClass* > v_simpleTestClass;
  pool::Token* token = tokenIterator->next();
  while ( token ) {
    std::pair< void*, Reflex::Type > objectAndClass_simpleTestClass = persistencySvc->readObject( *token );
    if ( objectAndClass_simpleTestClass.first == 0 ) {
      throw std::runtime_error( "Could not read the stored data" );
    }
    if ( objectAndClass_simpleTestClass.second == emptyType ) {
      throw std::runtime_error( "Could not retrieve the stored class" );
    }
    if ( objectAndClass_simpleTestClass.second.Name() != "SimpleTestClass" ) {
      throw std::runtime_error( "read wrong class type." );
    }
    SimpleTestClass* object_simpleTestClass = reinterpret_cast< SimpleTestClass* >( objectAndClass_simpleTestClass.first );
    if ( object_simpleTestClass->data != i ) {
      throw std::runtime_error( "Unexpected data read back" );
    }
    delete object_simpleTestClass;
    token->release();
    token = tokenIterator->next();

    // Commit and hold the transaction
    if ( ( i + 1 ) % m_eventsToCommitAndHold == 0 ) {
      if ( ! persistencySvc->session().transaction().commitAndHold() ) {
        throw std::runtime_error( "Could not commit and hold the transaction." );
      }
    }

    ++i;
  }

  delete tokenIterator;
  delete container;
  delete db;

  // Committing
  std::cout << "Committing the transaction." << std::endl;
  if ( ! persistencySvc->session().transaction().commit() ) {
    throw std::runtime_error( "Could not commit the transaction." );
  }

  catalog.commit();
}


void
pool::TestDriver::readFileSizes()
{
  pool::IFileCatalog& catalog = *m_fileCatalog;
  catalog.start();

  std::cout << "Creating the persistency service" << std::endl;
  pool::IPersistencySvcFactory* psfactory = pool::IPersistencySvcFactory::get();
  if ( ! psfactory ) {
    throw std::runtime_error( "Could not retrieve an IPersistencySvc factory" );
  }
  std::auto_ptr< pool::IPersistencySvc > persistencySvc( psfactory->create( "PersistencySvc", catalog ) );
  if ( ! persistencySvc.get() ) {
    throw std::runtime_error( "Could not create a PersistencySvc" );
  }

  // Starting a read transaction
  if ( ! persistencySvc->session().transaction().start( pool::ITransaction::READ ) ) {
    throw std::runtime_error( "Could not start a read transaction." );
  }

  // Committing
  std::cout << "Committing the transaction." << std::endl;
  if ( ! persistencySvc->session().transaction().commit() ) {
    throw std::runtime_error( "Could not commit the transaction." );
  }

  catalog.commit();
}
