/**
 * @file CollListAttrib.cpp
 * @brief Utility to list the metadata of a POOL collection
 * @author K. Karr <Kristo.Karr@cern.ch>
 * @author Marcin.Nowak@cern.ch
 * $Id: CollListAttrib.cpp,v 1.24 2008/03/27 16:13:14 mnowak Exp $
 */

#include "CollectionBase/ICollection.h"
#include "CollectionBase/ICollectionDescription.h"
#include "CollectionBase/ICollectionColumn.h"
#include "CollectionBase/ICollectionQuery.h"
#include "CollectionBase/ICollectionCursor.h"
#include "CollectionBase/CollectionRowBuffer.h"
#include "CollectionBase/CollectionService.h"

#include "POOLCore/Token.h"
#include "POOLCore/Exception.h"

#include "CoralBase/AttributeList.h"
#include "CoralBase/Attribute.h"
#include "CoralBase/MessageStream.h"

#include "CmdLineArgs.h"

#include "StorageSvc/DbStatus.h"

#include <iostream>
#include <iomanip>

using namespace std;

// fake TimeInfo, SEAL replacement
class TimeInfo
{
public:
  TimeInfo() {}

  void init() {}
  void processTimes( double, double, double ) {}
};



int main(int argc, const char *argv[])
{
   try  { 
      pool::DbStatus sc = pool::DbStatus::Success;
      std::vector<std::string> collNameVec;
      std::vector<std::string> collTypeVec;
      std::vector<std::string> collConnectVec;
      std::vector<std::string> collCatalogReadVec;
      std::vector<double> collOpenUserTimeVec;
      std::vector<double> collOpenSystemTimeVec;
      std::vector<double> collOpenRealTimeVec;
      std::vector<double> queryUserTimeVec;
      std::vector<double> querySystemTimeVec;
      std::vector<double> queryRealTimeVec;
      std::vector<double> eventCountUserTimeVec;
      std::vector<double> eventCountSystemTimeVec;
      std::vector<double> eventCountRealTimeVec;
      std::vector<double> metadataDeliveryUserTimeVec;
      std::vector<double> metadataDeliverySystemTimeVec;
      std::vector<double> metadataDeliveryRealTimeVec;
      std::vector<double> metadataReadUserTimeVec;
      std::vector<double> metadataReadSystemTimeVec;
      std::vector<double> metadataReadRealTimeVec;
      std::vector<double> metadataAnalysisUserTimeVec;
      std::vector<double> metadataAnalysisSystemTimeVec;
      std::vector<double> metadataAnalysisRealTimeVec;
      std::vector<double> totalMetadataAccessUserTimeVec;
      std::vector<double> totalMetadataAccessSystemTimeVec;
      std::vector<double> totalMetadataAccessRealTimeVec;
      std::vector<double> totalEventProcessingUserTimeVec;
      std::vector<double> totalEventProcessingSystemTimeVec;
      std::vector<double> totalEventProcessingRealTimeVec;
      std::vector<double> totalUserTimeVec;
      std::vector<double> totalSystemTimeVec;
      std::vector<double> totalRealTimeVec;

      coral::MessageStream log( "CollListAttrib" );

      CmdLineArgs cmdLineArgs(argc, argv);

      if( cmdLineArgs.hasQual("-help") ||
	  !cmdLineArgs.hasQual("-src") )
      {
	 std::cerr << "Usage: CollListAttrib "
		   << "<-src [+]<<input collection name> <input collection type>> > "
		   << "(List of name-type pairs is space separated; "
		   << "Choices for collection types are RootCollection, "
		   << "MySQLCollection, MySQLltCollection, RelationalCollection, "
		   << "PHYSICAL_NAME, LOGICAL_NAME or GUID) "
		   << "Optional arguments: "
		   << "[-srcconnect [Input database connection string]] "
		   << "(Note: The default is \"\" but this argument MUST be "
		   << "specified for collections of type MySQLCollection, "
		   << "MySQLltCollection and RelationalCollection. Also, for "
		   << "collections of type RelationalCollection the environment " 
		   << "variables POOL_AUTH_USER and POOL_AUTH_PASSWORD "
		   << "must be set) "
		   << "[-query [predicate string (DEFAULT = \"\")]] "
		   << "[-queryopt <comma separated attribute list> "
		   << "(DEFAULT = * ) "
		   << "[-ccread [+][read collection catalog connection string] "
		   << "(DEFAULT = \"\")]] "
		   << "[-nevtscan [Number of events starting from first event "
		   << "for which to scan and perform a simple count. If set to "
		   << "smaller value than set by \"-nevtdmp\", \"-nevtstat\" or "
		   << "\"-nevtread\", scan number will be reset to largest of "
		   << "these three values. \"all\" => scan all events "
		   << "(DEFAULT = \"all\")]] "
		   << "[-nevtread [Number of events starting from first event "
		   << "for which to read in metadata entries specified by "
		   << "\"-queryopt\" (default is all metadata). If set to "
		   << "smaller value than set by \"-nevtdmp\" or \"-nevtstat\', "
		   << "read number will be reset to larger of these two values. "
		   << "\"all\" => read all events (DEFAULT = 0)]] "
		   << "[-nevtstat [Number of events starting from first event "
		   << "for which to calculate max, min and average values. "
		   << "\"all\" => analyze all events (DEFAULT = 0)]] "
		   << "[-nevtdmp [Number of events starting from first event for "
		   << "which to dump the attribute values. \"all\" => dump for "
		   << "all events (DEFAULT = 0)]] "
		   << "[-nevtperprint [Number of events to scan before "
		   << "each print of status to the screen. \"all\" => "
		   << "skip printing for all events (DEFAULT = \"all\")]] "
		   << "[-quiet Qualifier that takes no arguments and specifies "
		   << "that no information except the (optional) event dump and "
		   << "event scan status information is to be printed to screen; "
		   << "Useful for performance testing] "
		   << "[-timing Qualifier that takes no arguments and specifies "
		   << "that time stamp information for key processes is to be "
		   << "printed out at the end of program execution]"
		   << std::endl;
	 return 1;
      }

      std::string connection = cmdLineArgs.getArg( "-srcconnect" );

      std::string collName = "";
      collName = cmdLineArgs.getArg( "-src" );
      if( (collName == "") ||
	  (collName == "-src") ||
	  (collName == "-query") ||
	  (collName == "-queryopt") ||
	  (collName == "-srcconnect") ||
	  (collName == "-ccread" ) ||
	  (collName == "-nevtscan") ||
	  (collName == "-nevtread") ||
	  (collName == "-nevtstat") ||
	  (collName == "-nevtdmp") ||
	  (collName == "-nevtperprint") ||
	  (collName == "-quiet") ||
	  (collName == "-timing") )
      {
	 std::cerr << "CollListAttrib: ERROR No input collection specified on " 
		   << "command line." << std::endl;
	 return 1;
      }

      std::string collType = "";
      std::string collConnect = "";
      while( (collName != "") &&
	     (collName != "-src") &&
	     (collName != "-query") &&
	     (collName != "-queryopt") &&
	     (collName != "-srcconnect") &&
	     (collName != "-ccread") &&
	     (collName != "-nevtscan") &&
	     (collName != "-nevtread") &&
	     (collName != "-nevtstat") &&
	     (collName != "-nevtdmp") &&
	     (collName != "-nevtperprint") &&
	     (collName != "-quiet") &&
	     (collName != "-timing") )
      {
	 collType = cmdLineArgs.getArg( cmdLineArgs.getPos() );
	 if( collType == "MySQLCollection" ||
	     collType == "MySQLltCollection" || 
	     collType == "RelationalCollection" )
	 {
	    if ( connection != "" )
	    {
	       collConnect = connection;
	    }
	    else
	    {
	       std::cerr << "CollListAttrib: ERROR Must specify a database "
			 << "connection string for collections of type " 
			 << "MySQL or Oracle." << std::endl;
	       return 1;
	    }
	 }
	 else if ( collType == "RootCollection" ||
		   collType == "PHYSICAL_NAME" ||
		   collType == "LOGICAL_NAME" ||
		   collType == "GUID" )
	 {
	    collConnect = "";
	 }
	 else
	 {
	    std::cerr << "CollListAttrib: ERROR Collection type " << collType 
		      << " is not presently handled by this program." 
		      << std::endl;
	    return 1;
	 }
	 collNameVec.push_back( collName );
	 collTypeVec.push_back( collType );
	 collConnectVec.push_back( collConnect );
	 collName = cmdLineArgs.getArg( cmdLineArgs.getPos() );
      }

      std::string query = cmdLineArgs.getArg( "-query" );

      std::string queryOptions = "";
      if ( cmdLineArgs.hasQual( "-queryopt" ) )
      {
	 queryOptions = cmdLineArgs.getArg("-queryopt");
	 if ( (queryOptions == "-src") ||
	      (queryOptions == "-query") ||
	      (queryOptions == "-queryopt") ||
	      (queryOptions == "-srcconnect") ||
	      (queryOptions == "-ccread") ||
	      (queryOptions == "-nevtscan") ||
	      (queryOptions == "-nevtread") ||
	      (queryOptions == "-nevtstat") ||
	      (queryOptions == "-nevtdmp") ||
	      (queryOptions == "-nevtperprint") ||
	      (queryOptions == "-quiet") ||
	      (queryOptions == "-timing") )
	 {
	    std::string queryOptions = "";
	 }
      }

      std::string collCatalogRead = cmdLineArgs.getArg( "-ccread" );
      while( collCatalogRead.size() && collCatalogRead[0] != '-' )  {
	 collCatalogReadVec.push_back( collCatalogRead );
	 collCatalogRead = cmdLineArgs.getArg( cmdLineArgs.getPos() );
      }

      int numEventsDump = 0;
      if ( cmdLineArgs.hasQual( "-nevtdmp" ) )
      {
	 std::string numEventsDumpStr = cmdLineArgs.getArg("-nevtdmp");
	 if ( numEventsDumpStr == "all" )
	 {
	    numEventsDump = -1;
	 }
	 else if ( (numEventsDumpStr != "") &&
		   (numEventsDumpStr != "-src") &&
		   (numEventsDumpStr != "-query") &&
		   (numEventsDumpStr != "-queryopt") &&
		   (numEventsDumpStr != "-srcconnect") &&
		   (numEventsDumpStr != "-ccread") &&
		   (numEventsDumpStr != "-nevtscan") &&
		   (numEventsDumpStr != "-nevtread") &&
		   (numEventsDumpStr != "-nevtstat") &&
		   (numEventsDumpStr != "-nevtdmp") &&
		   (numEventsDumpStr != "-nevtperprint") &&
		   (numEventsDumpStr != "-quiet") &&
		   (numEventsDumpStr != "-timing") )
	 {
	    numEventsDump = atoi( numEventsDumpStr.c_str() );
	 }
      }

      int numEventsAnalyze = 0;
      if ( cmdLineArgs.hasQual( "-nevtstat" ) )
      {
	 std::string numEventsAnalyzeStr = cmdLineArgs.getArg("-nevtstat");
	 if ( numEventsAnalyzeStr == "all" )
	 {
	    numEventsAnalyze = -1;
	 }
	 else if ( (numEventsAnalyzeStr != "") &&
		   (numEventsAnalyzeStr != "-src") &&
		   (numEventsAnalyzeStr != "-query") &&
		   (numEventsAnalyzeStr != "-queryopt") &&
		   (numEventsAnalyzeStr != "-srcconnect") &&
		   (numEventsAnalyzeStr != "-ccread") &&
		   (numEventsAnalyzeStr != "-nevtscan") &&
		   (numEventsAnalyzeStr != "-nevtread") &&
		   (numEventsAnalyzeStr != "-nevtstat") &&
		   (numEventsAnalyzeStr != "-nevtdmp") &&
		   (numEventsAnalyzeStr != "-nevtperprint") &&
		   (numEventsAnalyzeStr != "-quiet") &&
		   (numEventsAnalyzeStr != "-timing") )
	 {
	    numEventsAnalyze = atoi( numEventsAnalyzeStr.c_str() );
	 }
      }

      int numEventsRead = 0;
      if ( cmdLineArgs.hasQual( "-nevtread" ) )
      {
	 std::string numEventsReadStr = cmdLineArgs.getArg("-nevtread");
	 if ( numEventsReadStr == "all" )
	 {
	    numEventsRead = -1;
	 }
	 else if ( (numEventsReadStr != "") &&
		   (numEventsReadStr != "-src") &&
		   (numEventsReadStr != "-query") &&
		   (numEventsReadStr != "-queryopt") &&
		   (numEventsReadStr != "-srcconnect") &&
		   (numEventsReadStr != "-ccread") &&
		   (numEventsReadStr != "-nevtscan") &&
		   (numEventsReadStr != "-nevtread") &&
		   (numEventsReadStr != "-nevtstat") &&
		   (numEventsReadStr != "-nevtdmp") &&
		   (numEventsReadStr != "-nevtperprint") &&
		   (numEventsReadStr != "-quiet") &&
		   (numEventsReadStr != "-timing") )
	 {
	    numEventsRead = atoi( numEventsReadStr.c_str() );
	 }
      }
      if ( numEventsRead >= 0 )
      {
	 if ( (numEventsDump < 0) || (numEventsAnalyze < 0) )
	 {
	    numEventsRead = -1;
	 }
	 else 
	 {
	    if ( numEventsRead < numEventsDump ) numEventsRead = numEventsDump;
	    if ( numEventsRead < numEventsAnalyze ) numEventsRead = numEventsAnalyze;
	 }
      }

      int numEventsScan = -1;
      if ( cmdLineArgs.hasQual( "-nevtscan" ) )
      {
	 std::string numEventsScanStr = cmdLineArgs.getArg("-nevtscan");
	 if ( numEventsScanStr == "all" )
	 {
	    numEventsScan = -1;
	 }
	 else if ( (numEventsScanStr != "") &&
		   (numEventsScanStr != "-src") &&
		   (numEventsScanStr != "-query") &&
		   (numEventsScanStr != "-queryopt") &&
		   (numEventsScanStr != "-srcconnect") &&
		   (numEventsScanStr != "-ccread") &&
		   (numEventsScanStr != "-nevtscan") &&
		   (numEventsScanStr != "-nevtread") &&
		   (numEventsScanStr != "-nevtstat") &&
		   (numEventsScanStr != "-nevtdmp") &&
		   (numEventsScanStr != "-nevtperprint") &&
		   (numEventsScanStr != "-quiet") &&
		   (numEventsScanStr != "-timing") )
	 {
	    numEventsScan = atoi( numEventsScanStr.c_str() );
	 }
      }
      if ( numEventsScan >= 0 )
      {
	 if ( (numEventsRead < 0) ||
	      (numEventsDump < 0) || 
	      (numEventsAnalyze < 0) )
	 {
	    numEventsScan = -1;
	 }
	 else 
	 {
	    if ( numEventsScan < numEventsRead ) numEventsScan = numEventsRead;
	    if ( numEventsScan < numEventsDump ) numEventsScan = numEventsDump;
	    if ( numEventsScan < numEventsAnalyze ) numEventsScan = numEventsAnalyze;
	 }
      }

      int numEventsPerPrint = -1;
      if ( cmdLineArgs.hasQual( "-nevtperprint" ) )
      {
	 std::string numEventsPerPrintStr = cmdLineArgs.getArg("-nevtperprint");
	 if ( numEventsPerPrintStr == "all" )
	 {
	    numEventsPerPrint = -1;
	 }
	 else if ( (numEventsPerPrintStr != "") &&
		   (numEventsPerPrintStr != "-src") &&
		   (numEventsPerPrintStr != "-query") &&
		   (numEventsPerPrintStr != "-queryopt") &&
		   (numEventsPerPrintStr != "-srcconnect") &&
		   (numEventsPerPrintStr != "-ccread") &&
		   (numEventsPerPrintStr != "-nevtscan") &&
		   (numEventsPerPrintStr != "-nevtread") &&
		   (numEventsPerPrintStr != "-nevtstat") &&
		   (numEventsPerPrintStr != "-nevtdmp") &&
		   (numEventsPerPrintStr != "-nevtperprint") &&
		   (numEventsPerPrintStr != "-quiet") &&
		   (numEventsPerPrintStr != "-timing") )
	 {
	    numEventsPerPrint = atoi( numEventsPerPrintStr.c_str() );
	 }
      }

      bool noPrint = false;
      if ( cmdLineArgs.hasQual( "-quiet" ) )
      {
	 noPrint = true;
      }

      bool getTimingInfo = false;
      if ( cmdLineArgs.hasQual( "-timing" ) )
      {
         log << coral::Warning << " Timing measurements are not working " << coral::MessageStream::endmsg;
         // getTimingInfo = true;
      }


      if( !::getenv( "POOL_OUTMSG_LEVEL" )) {
         // pool::POOLContext::setMessageVerbosityLevel(coral::Info);
      } 
      else {
         // pool::POOLContext::setMessageVerbosityLevel( pms.threshold() );
      }

      log << coral::Info << "creating CollectionService" << coral::MessageStream::endmsg;
      auto_ptr<pool::CollectionService> collectionService( new pool::CollectionService() );

      // FIX - does the next line connect to the catalog by itself?
      for( unsigned int i=0; i<collCatalogReadVec.size(); i++ )      {
	 collectionService->addReadCollectionCatalog( collCatalogReadVec[i] );
      }
      // collCatalog->connect();

      unsigned int collCounter = 0;
      unsigned int evtCounterTotal = 0;
      std::string attribName = "";
      std::string attribTypeName = "";
      double utime1 = 0;
      double stime1 = 0;
      double rtime1 = 0;
      double utime2 = 0;
      double stime2 = 0;
      double rtime2 = 0;
      double utime3 = 0;
      double stime3 = 0;
      double rtime3 = 0;
      double utime4 = 0;
      double stime4 = 0;
      double rtime4 = 0;
      double utime5 = 0;
      double stime5 = 0;
      double rtime5 = 0;
      double utime6 = 0;
      double stime6 = 0;
      double rtime6 = 0;
      double utime7 = 0;
      double stime7 = 0;
      double rtime7 = 0;
      double utime8 = 0;
      double stime8 = 0;
      double rtime8 = 0;
      double utime9 = 0;
      double stime9 = 0;
      double rtime9 = 0;
      double utime10 = 0;
      double stime10 = 0;
      double rtime10 = 0;
      double utime11 = 0;
      double stime11 = 0;
      double rtime11 = 0;
      double utime12 = 0;
      double stime12 = 0;
      double rtime12 = 0;
      double utime13 = 0;
      double stime13 = 0;
      double rtime13 = 0;
      double utime14 = 0;
      double stime14 = 0;
      double rtime14 = 0;
      double utime15 = 0;
      double stime15 = 0;
      double rtime15 = 0;

      std::map<int, pool::ICollection*> collMap;
      TimeInfo timeInfo;
      for( unsigned int i=0; i<collNameVec.size(); i++ )    {
	 if( getTimingInfo )      {
	    timeInfo.init();
	    timeInfo.processTimes( utime1, stime1, rtime1 );
	 }
	 bool readOnly( true );
	 pool::ICollection* collection = collectionService->handle( collNameVec[i], collTypeVec[i], collConnectVec[i], readOnly );
	 collMap[i] = collection;

	 if( getTimingInfo ) {
	    timeInfo.init();
	    timeInfo.processTimes( utime2, stime2, rtime2 );

	    collOpenUserTimeVec.push_back( (utime2-utime1)*(1e-9) );
	    collOpenSystemTimeVec.push_back( (stime2-stime1)*(1e-9) );
	    collOpenRealTimeVec.push_back( (rtime2-rtime1)*(1e-9) );
	 }

	 collCounter++;

	 const pool::ICollectionDescription &description = collection->description();

	 unsigned int maxNameSize = 0;
	 unsigned int maxTypeNameSize = 0;
	 for( int f=0; f < description.numberOfCollectionFragments(); f++ ) {
	    for( int c=0; c < description.numberOfAttributeColumns( f ); c++ ) {
	       const pool::ICollectionColumn&	column = description.attributeColumn(c, f);
	       if( maxNameSize < column.name().size() ) 
		  maxNameSize = column.name().size();
	       if( maxTypeNameSize < column.type().size() ) 
		  maxTypeNameSize = column.type().size();
	    }
	 }

	 if( !noPrint ) {
	    std::cout << std::endl;
	    std::cout << "Collection name: " << description.name() << std::endl;
	    std::cout << "Collection type: " << description.type() << std::endl;
	    std::cout << "Number of collection fragments: " << description.numberOfCollectionFragments() << std::endl;
	    std::cout << "Number of attributes is: " << description.numberOfAttributeColumns() << std::endl;
	    std::cout << "Attributes are: " << std::endl;
	    for( int f=0; f < description.numberOfCollectionFragments(); f++ ) {
	       for( int c=0; c < description.numberOfAttributeColumns( f ); c++ ) {
		  const pool::ICollectionColumn&	column = description.attributeColumn(c, f);
		  std::cout.setf(std::ios::left);
		  std::cout << "NAME: ";
		  std::cout.width(maxNameSize+5);
		  std::cout << column.name();
		  std::cout << "TYPE: ";
		  std::cout.width(maxTypeNameSize+5);
		  std::cout << column.type();
		  std::cout << "INFO: ";
		  std::cout << column.annotation();
		  std::cout << std::endl;
	       }
	    }

	    if ( getTimingInfo )      {
	       timeInfo.init();
	       timeInfo.processTimes( utime3, stime3, rtime3 );
	    }

	    cout << ">> Creating query for the collection" << endl;
	    pool::ICollectionQuery *collQuery = collection->newQuery();
	    collQuery->setCondition( query );
	    if( queryOptions.size() ) {
	       collQuery->addToOutputList( queryOptions );
	    } else {
	       collQuery->selectAllAttributes();
	    }
	    cout << ">> Executing the query " << endl;
	    pool::ICollectionCursor& cursor = collQuery->execute();
	       
	    if ( getTimingInfo )
	    {
	       timeInfo.init();
	       timeInfo.processTimes( utime4, stime4, rtime4 );

	       queryUserTimeVec.push_back( (utime4-utime3)*(1e-9) );
	       querySystemTimeVec.push_back( (stime4-stime3)*(1e-9) );
	       queryRealTimeVec.push_back( (rtime4-rtime3)*(1e-9) );
	    }

	    std::map<std::string, short> shortAttrSumMap;
	    std::map<std::string, short> shortAttrMaxMap;
	    std::map<std::string, short> shortAttrMinMap;
	    std::map<std::string, unsigned short> uShortAttrSumMap;
	    std::map<std::string, unsigned short> uShortAttrMaxMap;
	    std::map<std::string, unsigned short> uShortAttrMinMap;
	    std::map<std::string, int> intAttrSumMap;
	    std::map<std::string, int> intAttrMaxMap;
	    std::map<std::string, int> intAttrMinMap;
	    std::map<std::string, unsigned int> uIntAttrSumMap;
	    std::map<std::string, unsigned int> uIntAttrMaxMap;
	    std::map<std::string, unsigned> uIntAttrMinMap;
	    std::map<std::string, long> longAttrSumMap;
	    std::map<std::string, long> longAttrMaxMap;
	    std::map<std::string, long> longAttrMinMap;
	    std::map<std::string, unsigned long> uLongAttrSumMap;
	    std::map<std::string, unsigned long> uLongAttrMaxMap;
	    std::map<std::string, unsigned long> uLongAttrMinMap;
	    std::map<std::string, float> floatAttrSumMap;
	    std::map<std::string, float> floatAttrMaxMap;
	    std::map<std::string, float> floatAttrMinMap;
	    std::map<std::string, double> doubleAttrSumMap;
	    std::map<std::string, double> doubleAttrMaxMap;
	    std::map<std::string, double> doubleAttrMinMap;
	    int evtCounter = 0;
	    double eventCountUserTime = 0;
	    double eventCountSystemTime = 0;
	    double eventCountRealTime = 0;
	    double metadataDeliveryUserTime = 0;
	    double metadataDeliverySystemTime = 0;
	    double metadataDeliveryRealTime = 0;
	    double metadataReadUserTime = 0;
	    double metadataReadSystemTime = 0;
	    double metadataReadRealTime = 0;
	    double metadataAnalysisUserTime = 0;
	    double metadataAnalysisSystemTime = 0;
	    double metadataAnalysisRealTime = 0;
	    double totalMetadataAccessUserTime = 0;
	    double totalMetadataAccessSystemTime = 0;
	    double totalMetadataAccessRealTime = 0;

	    if ( getTimingInfo )
	    {
	       timeInfo.init();
	       timeInfo.processTimes( utime5, stime5, rtime5 );
	    }

	    cout << ">> Iterating over query results" << endl;
	    while( cursor.next() ) {
	       if ( getTimingInfo )        {      
		  timeInfo.init();
		  timeInfo.processTimes( utime6, stime6, rtime6 );
	       }

	       if ( (numEventsScan >= 0) && (numEventsScan <= evtCounter) )
	       {
		  break;
	       }

	       evtCounter++; 
	       evtCounterTotal++;

	       if( numEventsPerPrint >= 0 && !(evtCounterTotal % numEventsPerPrint) ) {
		  log << coral::Info << evtCounterTotal << " events scanned" << coral::MessageStream::endmsg;
	       }     

	       if( getTimingInfo )        {
		  timeInfo.init();
		  timeInfo.processTimes( utime7, stime7, rtime7 );
        
		  eventCountUserTime += (utime7-utime6)*(1e-9);
		  eventCountSystemTime += (stime7-stime6)*(1e-9);
		  eventCountRealTime += (rtime7-rtime6)*(1e-9);
	       }

	       if ( (numEventsDump == 0) && 
		    (numEventsAnalyze == 0) &&
		    (numEventsRead == 0) )
	       {
		  continue;
	       }

	       if ( (numEventsDump < 0) || (numEventsDump >= evtCounter) )
	       {
		  std::cout << std::endl;
		  std::cout << "Collection name: " << collNameVec[i] << std::endl;
		  std::cout << "Collection type: " << collTypeVec[i] << std::endl;
		  std::cout << "Number of events counted: " 
			    << evtCounter << std::endl;
		  std::cout << "Attributes are: " << std::endl;
	       }

	       if ( getTimingInfo )
	       {
		  timeInfo.init();
		  timeInfo.processTimes( utime8, stime8, rtime8 );
	       }

	       const coral::AttributeList &attribList = cursor.currentRow().attributeList();

	       if ( getTimingInfo )        {
		  timeInfo.init();
		  timeInfo.processTimes( utime9, stime9, rtime9 );

		  metadataDeliveryUserTime += (utime9-utime8)*(1e-9);
		  metadataDeliverySystemTime += (stime9-stime8)*(1e-9);
		  metadataDeliveryRealTime += (rtime9-rtime8)*(1e-9);
	       }

	       for ( coral::AttributeList::const_iterator iter = attribList.begin();
		     iter != attribList.end(); ++iter ) 
	       {
		  if ( getTimingInfo )
		  {
		     timeInfo.init();
		     timeInfo.processTimes( utime10, stime10, rtime10 );
		  }

		  attribName = iter->specification().name();
		  attribTypeName = iter->specification().typeName();

		  if ( getTimingInfo )
		  {
		     timeInfo.init();
		     timeInfo.processTimes( utime11, stime11, rtime11 );

		     metadataReadUserTime += (utime11-utime10)*(1e-9);
		     metadataReadSystemTime += (stime11-stime10)*(1e-9);
		     metadataReadRealTime += (rtime11-rtime10)*(1e-9);
		  }

		  if ( (numEventsDump < 0) || (numEventsDump >= evtCounter) )
		  {
		     std::cout.setf(std::ios::left);
		     std::cout << "NAME: ";
		     std::cout.width(maxNameSize+5);
		     std::cout << attribName;
		     std::cout << "TYPE: ";
		     std::cout.width(maxTypeNameSize+5);
		     std::cout << attribTypeName;
		     std::cout << "VALUE: ";
		     iter->toOutputStream(cout);
		     cout<< std::endl;
		  }

		  if( getTimingInfo ) {
		     timeInfo.init();
		     timeInfo.processTimes( utime12, stime12, rtime12 );
		  }

		  if( numEventsAnalyze < 0 || numEventsAnalyze >= evtCounter )	  {
		     if( attribTypeName == "short" ) {
			short val = iter->data<short>();
			if ( shortAttrSumMap.count(attribName) == 0 ) {
			   shortAttrMaxMap[attribName] = val;
			   shortAttrMinMap[attribName] = val;
			}
			shortAttrSumMap[attribName] += val;
			if ( val > shortAttrMaxMap[attribName] )
			{
			   shortAttrMaxMap[attribName] = val;
			}
			if ( val < shortAttrMinMap[attribName] )
			{
			   shortAttrMinMap[attribName] = val;
			}
		     }
		     else if ( attribTypeName == "unsigned short" )
		     {
			unsigned short val = iter->data<unsigned short>();
			if ( uShortAttrSumMap.count(attribName) == 0 )
			{
			   uShortAttrMaxMap[attribName] = val;
			   uShortAttrMinMap[attribName] = val;
			}
			uShortAttrSumMap[attribName] += val;
			if ( val > uShortAttrMaxMap[attribName] )
			{
			   uShortAttrMaxMap[attribName] = val;
			}
			if ( val < uShortAttrMinMap[attribName] )
			{
			   uShortAttrMinMap[attribName] = val;
			}
		     }
		     else if ( attribTypeName == "int" )
		     {
			int val = iter->data<int>();
			if ( intAttrSumMap.count(attribName) == 0 )
			{
			   intAttrMaxMap[attribName] = val;
			   intAttrMinMap[attribName] = val;
			}
			intAttrSumMap[attribName] += val;
			if ( val > intAttrMaxMap[attribName] )
			{
			   intAttrMaxMap[attribName] = val;
			}
			if ( val < intAttrMinMap[attribName] )
			{
			   intAttrMinMap[attribName] = val;
			}
		     }
		     else if ( attribTypeName == "unsigned int" )
		     {
			unsigned int val = iter->data<unsigned>();
			if ( uIntAttrSumMap.count(attribName) == 0 ) {
			   uIntAttrMaxMap[attribName] = val;
			   uIntAttrMinMap[attribName] = val;
			}
			uIntAttrSumMap[attribName] += val;
			if ( val > uIntAttrMaxMap[attribName] )
			{
			   uIntAttrMaxMap[attribName] = val;
			}
			if ( val < uIntAttrMinMap[attribName] )
			{
			   uIntAttrMinMap[attribName] = val;
			}
		     }
		     else if ( attribTypeName == "long" )
		     {
			long val = iter->data<long>();
			if ( longAttrSumMap.count(attribName) == 0 )
			{
			   longAttrMaxMap[attribName] = val;
			   longAttrMinMap[attribName] = val;
			}
			longAttrSumMap[attribName] += val;
			if ( val > longAttrMaxMap[attribName] )
			{
			   longAttrMaxMap[attribName] = val;
			}
			if ( val < longAttrMinMap[attribName] )
			{
			   longAttrMinMap[attribName] = val;
			}
		     }
		     else if ( attribTypeName == "unsigned long" )	    {
			unsigned long val = iter->data<unsigned long>();
			if ( uLongAttrSumMap.count(attribName) == 0 )	      {
			   uLongAttrMaxMap[attribName] = val;
			   uLongAttrMinMap[attribName] = val;
			}
			uLongAttrSumMap[attribName] += val;
			if ( val > uLongAttrMaxMap[attribName] )
			{
			   uLongAttrMaxMap[attribName] = val;
			}
			if ( val < uLongAttrMinMap[attribName] )
			{
			   uLongAttrMinMap[attribName] = val;
			}
		     }
		     else if ( attribTypeName == "float" )
		     {
			float val = iter->data<float>();
			if ( floatAttrSumMap.count(attribName) == 0 ) {
			   floatAttrMaxMap[attribName] = val;
			   floatAttrMinMap[attribName] = val;
			}
			floatAttrSumMap[attribName] += val;
			if ( val > floatAttrMaxMap[attribName] )
			{
			   floatAttrMaxMap[attribName] = val;
			}
			if ( val < floatAttrMinMap[attribName] )
			{
			   floatAttrMinMap[attribName] = val;
			}
		     }
		     else if ( attribTypeName == "double" ) {
			double val = iter->data<double>();
			if( doubleAttrSumMap.count(attribName) == 0 ) {
			   doubleAttrMaxMap[attribName] = val;
			   doubleAttrMinMap[attribName] = val;
			}
			doubleAttrSumMap[attribName] += val;
			if ( val > doubleAttrMaxMap[attribName] )
			{
			   doubleAttrMaxMap[attribName] = val;
			}
			if ( val < doubleAttrMinMap[attribName] )
			{
			   doubleAttrMinMap[attribName] = val;
			}
		     }
		  }

		  if ( getTimingInfo )
		  {
		     timeInfo.init();
		     timeInfo.processTimes( utime13, stime13, rtime13 );

		     metadataAnalysisUserTime += (utime13-utime12)*(1e-9);
		     metadataAnalysisSystemTime += (stime13-stime12)*(1e-9);
		     metadataAnalysisRealTime += (rtime13-rtime12)*(1e-9);
		  }
	       }

	       if ( getTimingInfo )
	       {
		  timeInfo.init();
		  timeInfo.processTimes( utime14, stime14, rtime14 );

		  totalMetadataAccessUserTime += (utime14-utime9)*(1e-9);
		  totalMetadataAccessSystemTime += (stime14-stime9)*(1e-9);
		  totalMetadataAccessRealTime += (rtime14-rtime9)*(1e-9);
	       }
	    }

	    if ( getTimingInfo )
	    {
	       timeInfo.init();
	       timeInfo.processTimes( utime15, stime15, rtime15 );

	       eventCountUserTimeVec.push_back( eventCountUserTime );
	       eventCountSystemTimeVec.push_back( eventCountSystemTime );
	       eventCountRealTimeVec.push_back( eventCountRealTime );

	       metadataDeliveryUserTimeVec.push_back( metadataDeliveryUserTime );
	       metadataDeliverySystemTimeVec.push_back( metadataDeliverySystemTime );
	       metadataDeliveryRealTimeVec.push_back( metadataDeliveryRealTime );

	       metadataReadUserTimeVec.push_back( metadataReadUserTime );
	       metadataReadSystemTimeVec.push_back( metadataReadSystemTime );
	       metadataReadRealTimeVec.push_back( metadataReadRealTime );

	       metadataAnalysisUserTimeVec.push_back( metadataAnalysisUserTime );
	       metadataAnalysisSystemTimeVec.push_back( metadataAnalysisSystemTime );
	       metadataAnalysisRealTimeVec.push_back( metadataAnalysisRealTime );

	       totalMetadataAccessUserTimeVec.push_back( totalMetadataAccessUserTime );
	       totalMetadataAccessSystemTimeVec.push_back( totalMetadataAccessUserTime );
	       totalMetadataAccessRealTimeVec.push_back( totalMetadataAccessUserTime );

	       totalEventProcessingUserTimeVec.push_back( (utime15-utime5)*(1e-9) );
	       totalEventProcessingSystemTimeVec.push_back( (stime15-stime5)*(1e-9) );
	       totalEventProcessingRealTimeVec.push_back( (rtime15-rtime5)*(1e-9) );

	       totalUserTimeVec.push_back( (utime15-utime1)*(1e-9) );
	       totalSystemTimeVec.push_back( (stime15-stime1)*(1e-9) );
	       totalRealTimeVec.push_back( (rtime15-rtime1)*(1e-9) );
	    }

	    if ( !noPrint )
	    {
	       std::cout << std::endl;           
	       std::cout << "Total number of events scanned in collection " 
			 << collNameVec[i] 
			 << " (type: " << collTypeVec[i] << ") is: " << evtCounter 
			 << std::endl;
	       std::cout << std::endl;
	    }

	    if( numEventsAnalyze && !noPrint ) {
	       for( int f=0; f < description.numberOfCollectionFragments(); f++ ) {
		  for( int c=0; c < description.numberOfAttributeColumns( f ); c++ ) {
		     const pool::ICollectionColumn&	column = description.attributeColumn(c, f);
		     attribName = column.name();
		     attribTypeName = column.type();

		     std::cout.setf(std::ios::left);
		     std::cout << "NAME: ";
		     std::cout.width(maxNameSize+5);
		     std::cout << attribName;
		     std::cout << "TYPE: ";
		     std::cout.width(maxTypeNameSize+5);
		     std::cout << attribTypeName << std::endl;
		     if ( attribTypeName == "short" )
		     {
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "AVERAGE VALUE: "
				  << shortAttrSumMap[attribName]/evtCounter 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MAX VALUE: " 
				  << shortAttrMaxMap[attribName] 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MIN VALUE: " 
				  << shortAttrMinMap[attribName] 
				  << std::endl;
		     }
		     else if ( attribTypeName == "unsigned short" )
		     {
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "AVERAGE VALUE: "
				  << uShortAttrSumMap[attribName]/evtCounter 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MAX VALUE: "
				  << uShortAttrMaxMap[attribName] 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MIN VALUE: "
				  << uShortAttrMinMap[attribName] 
				  << std::endl;
		     }
		     else if ( attribTypeName == "int" )
		     {
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "AVERAGE VALUE: "
				  << intAttrSumMap[attribName]/evtCounter 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MAX VALUE: "
				  << intAttrMaxMap[attribName] 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MIN VALUE: "
				  << intAttrMinMap[attribName] 
				  << std::endl;
		     }
		     else if ( attribTypeName == "unsigned int" )
		     {
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "AVERAGE VALUE: "
				  << uIntAttrSumMap[attribName]/evtCounter 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MAX VALUE: "
				  << uIntAttrMaxMap[attribName] 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MIN VALUE: "
				  << uIntAttrMinMap[attribName] 
				  << std::endl;
		     }
		     else if ( attribTypeName == "long" )
		     {
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "AVERAGE VALUE: "
				  << longAttrSumMap[attribName]/evtCounter 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MAX VALUE: "
				  << longAttrMaxMap[attribName] 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MIN VALUE: "
				  << longAttrMinMap[attribName] 
				  << std::endl;
		     }
		     else if ( attribTypeName == "unsigned long" )
		     {
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "AVERAGE VALUE: "
				  << uLongAttrSumMap[attribName]/evtCounter 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MAX VALUE: "
				  << uLongAttrMaxMap[attribName] 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MIN VALUE: "
				  << uLongAttrMinMap[attribName] 
				  << std::endl;
		     }
		     else if ( attribTypeName == "float" )
		     {
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "AVERAGE VALUE: "
				  << floatAttrSumMap[attribName]/evtCounter 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MAX VALUE: "
				  << floatAttrMaxMap[attribName] 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MIN VALUE: "
				  << floatAttrMinMap[attribName] 
				  << std::endl;
		     }
		     else if ( attribTypeName == "double" )
		     {
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "AVERAGE VALUE: "
				  << doubleAttrSumMap[attribName]/evtCounter 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MAX VALUE: "
				  << doubleAttrMaxMap[attribName] 
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MIN VALUE: "
				  << doubleAttrMinMap[attribName] 
				  << std::endl;
		     }
		     else
		     {
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "AVERAGE VALUE: N/A"
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MAX VALUE: N/A"
				  << std::endl;
			std::cout.setf(std::ios::left);
			std::cout.width(maxNameSize+11);
			std::cout << " ";
			std::cout << "MIN VALUE: N/A"
				  << std::endl;
		     }
		  }
	       }
	       std::cout << std::endl;
	    }

	    if ( getTimingInfo )
	    {
	       std::cout.setf(std::ios::left);
	       std::cout << "TIMING INFORMATION FOR COLLECTION: "
			 << collNameVec[i]
			 << std::endl;
	       std::cout << std::endl;

	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "ACTION: "
			 << "Collection Access"
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (USER): "
			 << std::setw(10)
			 << collOpenUserTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (SYSTEM): "
			 << std::setw(10)
			 << collOpenSystemTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "REAL TIME: "
			 << std::setw(10)
			 << collOpenRealTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout << std::endl;

	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "ACTION: "
			 << "Server Side Query"
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (USER): "
			 << std::setw(10)
			 << queryUserTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (SYSTEM): "
			 << std::setw(10)
			 << querySystemTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "REAL TIME: "
			 << std::setw(10)
			 << queryRealTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout << std::endl;

	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "ACTION: "
			 << "Event Count"
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (USER): "
			 << std::setw(10)
			 << eventCountUserTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (SYSTEM): "
			 << std::setw(10)
			 << eventCountSystemTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "REAL TIME: "
			 << std::setw(10)
			 << eventCountRealTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout << std::endl;

	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "ACTION: "
			 << "Metadata Delivery"
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (USER): "
			 << std::setw(10)
			 << metadataDeliveryUserTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (SYSTEM): "
			 << std::setw(10)
			 << metadataDeliverySystemTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "REAL TIME: "
			 << std::setw(10)
			 << metadataDeliveryRealTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout << std::endl;

	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "ACTION: "
			 << "Metadata Read"
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (USER): "
			 << std::setw(10)
			 << metadataReadUserTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (SYSTEM): "
			 << std::setw(10)
			 << metadataReadSystemTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "REAL TIME: "
			 << std::setw(10)
			 << metadataReadRealTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout << std::endl;

	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "ACTION: "
			 << "Metadata Analysis"
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (USER): "
			 << std::setw(10)
			 << metadataAnalysisUserTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (SYSTEM): "
			 << std::setw(10)
			 << metadataAnalysisSystemTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "REAL TIME: "
			 << std::setw(10)
			 << metadataAnalysisRealTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout << std::endl;

	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "ACTION: "
			 << "Total Metadata Processing"
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (USER): "
			 << std::setw(10)
			 << totalMetadataAccessUserTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (SYSTEM): "
			 << std::setw(10)
			 << totalMetadataAccessSystemTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "REAL TIME: "
			 << std::setw(10)
			 << totalMetadataAccessRealTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout << std::endl;

	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "ACTION: "
			 << "Total Event Processing"
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (USER): "
			 << std::setw(10)
			 << totalEventProcessingUserTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (SYSTEM): "
			 << std::setw(10)
			 << totalEventProcessingSystemTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "REAL TIME: "
			 << std::setw(10)
			 << totalEventProcessingRealTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout << std::endl;

	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "ACTION: "
			 << "Total Processing"
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (USER): "
			 << std::setw(10)
			 << totalUserTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "CPU TIME (SYSTEM): "
			 << std::setw(10)
			 << totalSystemTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout.setf(std::ios::left);
	       std::cout << std::setw(20)
			 << "REAL TIME: "
			 << std::setw(10)
			 << totalRealTimeVec[i]
			 << " sec."
			 << std::endl;
	       std::cout << std::endl;
	    }
	 }
      }
      if ( !noPrint )
      {
	 std::cout << std::endl;
	 std::cout << "Number of collections scanned is: " 
		   << collCounter << std::endl;
	 std::cout << "Total number of events scanned in all collections is: " 
		   << evtCounterTotal << std::endl;
	 std::cout << std::endl;
      }

      for( unsigned int i=0; i<collNameVec.size(); i++ ) {      
	 collMap[i]->close();
      }

      return 0;
   }
   catch( pool::Exception& poolException )
   {
      std::cerr << "pool::Exception: " << poolException.what() << std::endl;;
      return 1;
   }
   catch( std::exception& exception )
   {
      std::cerr << "std::exception: " << exception.what() << std::endl;;
      return 1;
   }
   catch( ... )
   {
      std::cerr << "Unknown exception caught in main()." << std::endl;;
      return 1;
   }
}
