/**
 * @file CollSplitByGUID.cpp
 * @brief Utility to list the file GUIDs used by a POOL collection and split the
 * collection into sub-collections by GUID
 * @author K. Karr <Kristo.Karr@cern.ch>, C. Nicholson <Caitriana.Nicholson@cern.ch>
 * $Id: CollSplitByGUID.cpp,v 1.4 2008/10/30 16:14:50 mnowak Exp $
 */

#include "CollectionBase/ICollection.h"
#include "CollectionBase/CollectionService.h"
#include "CollectionBase/CollectionDescription.h"
#include "CollectionBase/CollectionRowBuffer.h"
#include "CollectionBase/TokenList.h"
#include "CollectionBase/ICollectionQuery.h"
#include "CollectionBase/ICollectionCursor.h"
#include "CollectionBase/ICollectionColumn.h"
#include "CollectionBase/ICollectionDataEditor.h"

#include "CoralBase/Attribute.h"
#include "CoralBase/MessageStream.h"

#include "FileCatalog/IFileCatalog.h"
#include "DataSvc/DataSvcFactory.h"
#include "DataSvc/IDataSvc.h"
#include "PersistencySvc/ISession.h"
#include "PersistencySvc/DatabaseConnectionPolicy.h"
#include "PersistencySvc/ITransaction.h"

#include "POOLCore/Token.h"
#include "POOLCore/Exception.h"
#include "CmdLineArgs.h"

#include <iostream>
#include <string>

const std::string FILE_CATALOG_CONNECTION = 
  "xmlcatalog_file:PoolFileCatalog.xml";
const std::string COLLECTION_CATALOG_CONNECTION =
  "xmlcatalog_file:CollectionCatalog.xml";

const int DEFAULT_MIN_EVENTS = 0;

int main(int argc, const char *argv[])
{
  try
  {
    std::vector<std::string> collNameVec;
    std::vector<std::string> collTypeVec;
    std::vector<std::string> collConnectVec;
    std::vector<std::string> fileCatalogReadVec;
    std::vector<std::string> collCatalogReadVec;

    CmdLineArgs cmdLineArgs(argc, argv);

    if( cmdLineArgs.hasQual("-help") ||
        !cmdLineArgs.hasQual("-src") )
    {
      std::cerr << "*******************************************************************\n"
                << "* CollSplitByGUID is a tool for querying an event collection, or  *\n"
                << "* collections, and outputting the resultant events as a number of *\n"
                << "* sub-collections, one for each GUID of the files containing the  *\n"
                << "* events. Currently, these sub-collections are output as LOCAL,   *\n"
                << "* ROOT-based collections and will appear in the directory from    *\n"
                << "* which you call CollSplitByGUID.                                 *\n"
                << "*******************************************************************\n"
                << "Usage: CollSplitByGUID  " << std::endl
                << std::endl
                << "Required arguments: " << std::endl
                << "< -src [+]<<input collection name> <input collection type>> > " << std::endl
                << "  (List of name-type pairs is space separated; " << std::endl
                << "  Choices for collection types are RootCollection, " << std::endl
                << "  MySQLCollection, MySQLltCollection, RelationalCollection, " << std::endl
                << "  PHYSICAL_NAME, LOGICAL_NAME or GUID)" << std::endl
                << std::endl
                << "Optional arguments: " << std::endl
                << "[-srcconnect [Input database connection string]] " << std::endl
                << "  (Note: The default is \"\" and this argument MUST be " << std::endl
                << "  specified for collections of type MySQLCollection, " << std::endl
                << "  MySQLltCollection and RelationalCollection. Also, for " << std::endl
                << "  collections of type RelationalCollection the environment " << std::endl
                << "  variables POOL_AUTH_USER and POOL_AUTH_PASSWORD " << std::endl
                << "  must be set) \n"
                << "[-query [predicate string (DEFAULT = \"\")]] \n"
                << "[-queryopt [Choices are space separated list of the " << std::endl
                << "  following: \"FetchOne\" (no caching) or \"FetchAll\" " << std::endl
                << "  (caching) and/or \"SELECT <comma separated attribute list>\" " << std::endl
                << "  (DEFAULT = \"FetchOne SELECT *\")]]" << std::endl
                << "[-minevents [minimum number of events to write to output collection]]\n"
                << "[-fcread [+][read file catalog connection string "
                << "(DEFAULT = \"\")]] \n"
                << "[-ccread [+][read collection catalog connection string "
                << "(DEFAULT = \"\")]] \n"
                << std::endl;
      return 1;
    }

    std::string connection = cmdLineArgs.getArg( "-srcconnect" );

    std::string collName = "";
    collName = cmdLineArgs.getArg( "-src" );
    if( (collName == "") ||
        (collName == "-src") ||
        (collName == "-query") ||
        (collName == "-queryopt") ||
        (collName == "-srcconnect") ||
        (collName == "-minevents") ||
        (collName == "-fcread") ||
        (collName == "-ccread") )
    {
        std::cerr << "CollSplitByGUID: ERROR No input collection specified on " 
                  << "command line." << std::endl;
        return 1;
    }

    std::string collType = "";
    std::string collConnect = "";
    while( (collName != "") &&
           (collName != "-src") &&
           (collName != "-query") &&
           (collName != "-queryopt") &&
           (collName != "-minevents") &&
           (collName != "-srcconnect") &&
           (collName != "-fcread") &&
           (collName != "-ccread") )
    {
      collType = cmdLineArgs.getArg( cmdLineArgs.getPos() );
      if( collType == "MySQLCollection" ||
          collType == "MySQLltCollection" ||
          collType == "RelationalCollection" )
      {
        if ( connection != "" )
        {
          collConnect = connection;
        }
        else
        {
          std::cerr << "CollSplitByGUID: ERROR Must specify a database "
                    << "connection string for collections of type " 
                    << "MySQL or Oracle." << std::endl;
          return 1;
        }
      }
      else if ( collType == "RootCollection" ||
                collType == "PHYSICAL_NAME" ||
                collType == "LOGICAL_NAME" ||
                collType == "GUID" )
      {
        collConnect = "";
      }
      else
      {
        std::cerr << "CollSplitByGUID: ERROR Collection type " << collType 
                  << " is not presently handled by this program." 
                  << std::endl;
        return 1;
      }
      collNameVec.push_back( collName );
      collTypeVec.push_back( collType );
      collConnectVec.push_back( collConnect );
      collName = cmdLineArgs.getArg( cmdLineArgs.getPos() );
    }

    std::string query = cmdLineArgs.getArg( "-query" );

    std::string queryOptions = "";
    if ( cmdLineArgs.hasQual( "-queryopt" ) )
    {
      queryOptions = cmdLineArgs.getArg("-queryopt");
      if( (queryOptions == "-src") ||
          (queryOptions == "-query") ||
          (queryOptions == "-queryopt") ||
          (queryOptions == "-minevents") ||
          (queryOptions == "-srcconnect") ||
          (queryOptions == "-fcread") ||
          (queryOptions == "-ccread") )
      {
        std::string queryOptions = "";
      }
    }

    std::string fileCatalogRead = cmdLineArgs.getArg( "-fcread" );
    if( (fileCatalogRead == "") ||
        (fileCatalogRead == "-src") ||
        (fileCatalogRead == "-query") ||
        (fileCatalogRead == "-queryopt") ||
        (fileCatalogRead == "-minevents") ||
        (fileCatalogRead == "-srcconnect") ||
        (fileCatalogRead == "-fcread") ||
        (fileCatalogRead == "-ccread") )
    {
      fileCatalogRead = FILE_CATALOG_CONNECTION;
    }
    while( (fileCatalogRead != "") &&
           (fileCatalogRead != "-src") &&
           (fileCatalogRead != "-query") &&
           (fileCatalogRead != "-queryopt") &&
           (fileCatalogRead != "-minevents") &&
           (fileCatalogRead != "-srcconnect") &&
           (fileCatalogRead != "-fcread") &&
           (fileCatalogRead != "-ccread") )
    {
      fileCatalogReadVec.push_back( fileCatalogRead );
      fileCatalogRead = cmdLineArgs.getArg( cmdLineArgs.getPos() );
    }

    std::string collCatalogRead = cmdLineArgs.getArg( "-ccread" );
    if( (collCatalogRead == "") ||
        (collCatalogRead == "-src") ||
        (collCatalogRead == "-query") ||
        (collCatalogRead == "-queryopt") ||
        (collCatalogRead == "-minevents") ||
        (collCatalogRead == "-srcconnect") ||
        (collCatalogRead == "-fcread") ||
        (collCatalogRead == "-ccread") )
    {
      collCatalogRead = COLLECTION_CATALOG_CONNECTION;
    }
    while( (collCatalogRead != "") &&
           (collCatalogRead != "-src") &&
           (collCatalogRead != "-query") &&
           (collCatalogRead != "-queryopt") &&
           (collCatalogRead != "-minevents") &&
           (collCatalogRead != "-srcconnect") &&
           (collCatalogRead != "-fcread") &&
           (collCatalogRead != "-ccread") )
    {
      collCatalogReadVec.push_back( collCatalogRead );
      collCatalogRead = cmdLineArgs.getArg( cmdLineArgs.getPos() );
    }

    int minEvents = 0;
    std::string minEventsString = "";
    if( cmdLineArgs.hasQual("-minevents") )
    {
       minEventsString = cmdLineArgs.getArg( "-minevents" );
       if( (minEventsString == "") ||
           (minEventsString == "-src") ||
           (minEventsString == "-query") ||
           (minEventsString == "-queryopt") ||
           (minEventsString == "-minevents") ||
           (minEventsString == "-srcconnect") ||
           (minEventsString == "-fcread") ||
           (minEventsString == "-ccread") )
       {
          minEvents = DEFAULT_MIN_EVENTS;
       }
       std::cout << "Minimum number of events: " << minEventsString << std::endl; 
       minEvents = atoi( minEventsString.c_str());
    }
 
    if(!::getenv( "POOL_OUTMSG_LEVEL" ))    {
       //pool::POOLContext::setMessageVerbosityLevel(seal::Msg::Warning);
    }
    else    {
       //pool::PoolMessageStream pms("get threshold");
       //pool::POOLContext::setMessageVerbosityLevel( pms.threshold() );
    }

    std::auto_ptr<pool::CollectionService> collectionService( new pool::CollectionService() );

    for ( unsigned int i=0; i<fileCatalogReadVec.size(); i++ )    {
      collectionService->addReadFileCatalog( fileCatalogReadVec[i] );
    }

    for ( unsigned int i=0; i<collCatalogReadVec.size(); i++ )    {
      collectionService->addReadCollectionCatalog( collCatalogReadVec[i] );
    }
    
    pool::IDataSvc* dataSvc = pool::DataSvcFactory::instance( collectionService->getFileCatalog() );
    pool::DatabaseConnectionPolicy policy;
    policy.setReadMode( pool::DatabaseConnectionPolicy::READ );
    dataSvc->session().setDefaultConnectionPolicy( policy );
    if( ! dataSvc->transaction().start( pool::ITransaction::READ ) )  {
       std::cerr << "CollListFileGUID: ERROR Cannot start database "
                 << "transaction." << std::endl;
       return 1;
    }

    std::map<int, pool::ICollection*> collMap;
    int collectionNumber = 1;

    // for each input collection, loop through events and write out sub-collections
    for ( unsigned int i=0; i<collNameVec.size(); i++ )
    {
       
      std::map< std::string, int > fileIdMap;
      std::map< std::string, pool::ICollection*> subCollMap;  // map to keep guid --> sub-collection
      std::multimap< std::string, std::string> invCollMap;  // map to keep sub-collection name --> guid

      bool readOnly( true );
      pool::ICollection* collection = collectionService->handle( collNameVec[i], collTypeVec[i], collConnectVec[i], readOnly );
	 
      collMap[i] = collection;

      const pool::ICollectionDescription &firstDesc = collMap[0]->description();
      const pool::ICollectionDescription &nextDesc = collection->description();



      if ( ((query != "") || (queryOptions != "")) && !nextDesc.equals(firstDesc) )
      {
        std::cout << "CollSplitByGUID: WARNING The schemas of one or more "
                  << "input collections are different and a query has been "
                  << "requested. This may lead to unexpected behavior."
                  << std::endl;
      }

      pool::ICollectionQuery *collQuery = collection->newQuery();
      pool::ICollectionQuery *collQuery2 = collection->newQuery();
      collQuery->setCondition( query );
      collQuery2->setCondition( query );
      if( queryOptions.size() ) {
         collQuery->addToOutputList( queryOptions );
         collQuery2->addToOutputList( queryOptions );
      } else {
         collQuery->selectAll();
         collQuery2->selectAll();
      }
      std::cout << ">> Executing query for the source collection" << std::endl;
      
      pool::ICollectionCursor& cursor = collQuery->execute();

      // set parameters for the sub-collections. 
      // currently caters only for writing local, Root-based collections.
      std::string subCollType = "RootCollection";
      std::string subCollConnect = "";

      // first loop: for each event, find the GUID / fileId and 
      // count how many events match that GUID
      int totalEvents = 0;
      int numFiles = 0;

      // loop currently fixed to take EVENT_REFERENCE
      while( cursor.next() ) {
         const pool::TokenList &tokens = cursor.currentRow().tokenList();
         for( pool::TokenList::const_iterator iter = tokens.begin(); iter != tokens.end(); ++iter ) {
            std::string name = iter.tokenName();
            if( name == collection->description().eventReferenceColumnName() ) {   
               std::string fileId = iter->dbID();
               std::map< std::string, int >::iterator mapIter;
               mapIter = fileIdMap.find( fileId );
               if( mapIter == fileIdMap.end() )        {
                  fileIdMap[ fileId ] = numFiles++;
               }
               else {
               fileIdMap[ fileId ] += 1;
               }
            }
         }
         totalEvents++;
      }


      // make suitable collections, each with nEvents >= minEvents
      int heldEvents = 0;
      int currentFile = 0;
      bool lastCollection = false;
      std::vector<std::string> heldFileIds;   //to hold GUIDs until minEvents reached 
      pool::ICollection* previousCollection=0;

      std::map< std::string, int >::iterator mapIter = fileIdMap.begin();
      while ( mapIter != fileIdMap.end()) 
      {
         std::string fileId = mapIter->first;
         int thisNumEvents = mapIter->second;
         currentFile += 1;
         if ( currentFile == numFiles )
            lastCollection = true;

         // test for various conditions in which a new sub-collection ought to be created
         if ( thisNumEvents + heldEvents >= minEvents ||  lastCollection == true ) 
         {
            if (!lastCollection || collectionNumber == 1 || (lastCollection && (thisNumEvents + heldEvents >=minEvents)) )
            {
	      //create sub-collection
              std::stringstream collectionNameStream; 
              collectionNameStream << "sub_collection_" << collectionNumber; 
              std::string subCollName = collectionNameStream.str();
      
              pool::CollectionDescription  destDescription( collNameVec[i], "type" );
              destDescription = nextDesc;
              pool::CollectionDescription newDestDesc( destDescription );
              newDestDesc.setName( subCollName );
              newDestDesc.setType( subCollType );               
              newDestDesc.setConnection( subCollConnect );
              pool::ICollection* subCollection = collectionService->create( newDestDesc, true );
              subCollection->dataEditor().setRowCacheSize( 0 );                                                       

              // map to appropriate GUID
              subCollMap[ fileId ] = subCollection;
              invCollMap.insert( std::pair<std::string, std::string>(subCollName, fileId));
              for (unsigned int i=0; i<heldFileIds.size(); i++)
              { 
                 std::string thisId = heldFileIds[i];
                 subCollMap[ thisId ] = subCollection; 
                 invCollMap.insert( std::pair<std::string, std::string>(subCollName, thisId));
              }
              heldEvents = 0;
              heldFileIds.clear();
              collectionNumber++;
              previousCollection = subCollection;
            }
            else
            {
              subCollMap[ fileId ] = previousCollection;
              invCollMap.insert( std::pair<std::string, std::string>(previousCollection->description().name(), fileId));   
              for (unsigned int i=0; i<heldFileIds.size(); i++)
              {
                 std::string thisId = heldFileIds[i];
                 subCollMap[ thisId ] = previousCollection;
                 invCollMap.insert( std::pair<std::string, std::string>(previousCollection->description().name(), fileId));
              }
            }
         }
         else
         {
            heldFileIds.push_back(fileId); 
            heldEvents += thisNumEvents;
         }
         ++mapIter;
      }

      std::string lastCollName = "";
      for( std::multimap<std::string, std::string>::const_iterator invCollIter = invCollMap.begin(); 
           invCollIter != invCollMap.end(); 
           ++invCollIter) 
      { 
         std::string thisCollName = invCollIter->first;
         if (thisCollName == lastCollName)
            std::cout << " " << invCollIter->second;
         else {   
            if (invCollIter != invCollMap.begin())
               std::cout << std::endl;
            std::cout << "Created new sub-collection " << thisCollName << " with files:\t" << invCollIter->second;
         }
         lastCollName = thisCollName;
      }
      std::cout << std::endl;
 
      // second loop: write events to correct sub-collection
      // loop currently fixed to use EVENT_REFERENCE
      pool::ICollectionCursor& cursor2 = collQuery2->execute();
      const std::string& srcRefName = collection->description().eventReferenceColumnName();
      while( cursor2.next() ) {
         // get file GUID from event
        const pool::TokenList &tokens = cursor2.currentRow().tokenList();
       for( pool::TokenList::const_iterator ti = tokens.begin(); ti != tokens.end(); ++ti ) {
           std::string name = ti.tokenName();
           if( name == srcRefName ) {
               std::string guid = ti->dbID();
               pool::CollectionRowBuffer &rowBuffer = subCollMap[guid]->dataEditor().rowBuffer();
               rowBuffer.attributeList() = cursor2.currentRow().attributeList();
               const std::string& dstRefName = subCollMap[guid]->description().eventReferenceColumnName();
               if( ti.tokenName() == srcRefName ) {
                  ti->setData( &rowBuffer.tokenList()[ dstRefName ] );
               } else {
                  ti->setData( &rowBuffer.tokenList()[ ti.tokenName() ] );
               }
               subCollMap[guid]->dataEditor().insertRow( rowBuffer );
           }
        }
      }

      std::cout << "Finished writing all events from input collection " << collection->description().name() << std::endl;

      // commit changes and close all sub-collections
      std::map< std::string, pool::ICollection* >::iterator closeIter = subCollMap.begin();
      while ( closeIter != subCollMap.end() )
      {
         std::string guid = closeIter->first;
         pool::ICollection* subCollection = closeIter->second;
         subCollection->commit();  
         subCollection->close();
         ++closeIter;
      }

    } // for each input collection

    for( unsigned int i=0; i<collNameVec.size(); i++ )
    {      
      collMap[i]->close();
    }

    return 0;
  }
  catch( pool::Exception& poolException )
  {
    std::cerr << "pool::Exception: " << poolException.what() << std::endl;
    return 1;
  }
  catch( std::exception& exception )
  {
    std::cerr << "std::Exception: " << exception.what() << std::endl;
    return 1;
  }
  catch( ... )
  {
    std::cerr << "Unknown exception. " << std::endl;
    return 1;
  }
}
